import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { utils } from "./utils.js";
import { CCSType, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCConstantContainer } from "./PCCContainer.js";
export class PCCCompileProgramVisitor extends PC.AbstractVisitor {
    compiler;
    compilingNodes;
    constructor(compiler, compilingNodes) {
        super();
        this.compiler = compiler;
        this.compilingNodes = compilingNodes;
    }
    compileNode(node, context) {
        this.compilingNodes.push(node);
        node.accept(this, context);
        this.compilingNodes.pop();
    }
    visitProgram(node) {
        node.declarations.forEach((n) => this.compileNode(n));
    }
    visitMainAgent(node) {
        this.compiler.beginMainAgent(node);
        this.compiler.statementsVisitor.compileNode(node.body);
        this.compiler.emitStop();
        this.compiler.endMainAgent();
    }
    visitProcedureDeclaration(node) {
        this.compiler.beginProcedure(node.name);
        const proc = this.compiler.getProcedureWithName(node.name);
        if (proc.isMonitorProcedure()) {
            const guard = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("guard", null, true));
            const agent = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("a", null, true));
            this.compiler.emitOutput("lock", guard.getContainer(this.compiler), agent.getContainer(this.compiler));
        }
        this.compiler.statementsVisitor.compileNode(node.body);
        proc.emitExit(this.compiler); // Ask the procedure to emit exit, because this depends on the procedure type: monitor procedures for example must return the mutex.
        this.compiler.endProcedure();
    }
    visitMonitor(node) {
        this.compiler.beginClass(node.name);
        this.compiler.setNeedsMutex();
        node.childNodes.forEach((n) => this.compileNode(n));
        this.compiler.endClass();
    }
    visitStruct(node) {
        this.compiler.beginClass(node.name);
        node.childNodes.forEach((n) => this.compileNode(n));
        this.compiler.endClass();
    }
    visitConditionDeclaration(node) {
        this.compiler.setNeedsWaitRoom();
        const variable = new PCCVariableInfo(node, node.name, PCT.SystemTypes.condition);
        this.compiler.handleNewVariable(variable);
    }
    visitVariableDeclarations(node) {
        // Check if a lock is declared
        // 1. Get the type from the type node
        let type = this.compiler.staticSemantics.typeForNode(node.typeNode).declaredType;
        // 2. If the type is an array type, unwrap the type to get the inner most type
        while (type instanceof PCT.ArrayType) {
            type = type.elementsType;
        }
        // 3. Check if this is the lock type
        if (type.isLockType) {
            this.compiler.setNeedsMutex();
        }
        // Compile the declarators
        node.declarators.forEach((n) => this.compileNode(n));
    }
    visitVariableDeclarator(node) {
        const type = this.compiler.staticSemantics.typeForNode(node).declaredType;
        const variable = new PCCVariableInfo(node, node.name, type);
        this.compiler.handleNewVariable(variable);
    }
    // The following nodes don't need a compile method
    visitProcedureParameter(node, context) {
        throw new Error("Method not implemented.");
    }
    visitVariableInitializer(node, context) {
        throw new Error("Method not implemented.");
    }
    visitArrayTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitSimpleTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitChannelTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitClassTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    // The remaining nodes are covered by the other visitors
    visitStartExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAssignExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAssignDestination(node, context) {
        throw new Error("Method not implemented.");
    }
    visitSendExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitConditionalExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitOrExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAndExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitEqualityExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitRelationalExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAdditiveExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitMultiplicativeExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitUnaryExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitPostfixExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitReceiveExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitProcedureCall(node, context) {
        throw new Error("Method not implemented.");
    }
    visitClassCall(node, context) {
        throw new Error("Method not implemented.");
    }
    visitArrayExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitLiteralExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitIdentifierExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitStatementWrapper(node, context) {
        throw new Error("Method not implemented.");
    }
    visitBreakStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitContinueStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitStatementBlock(node, context) {
        throw new Error("Method not implemented.");
    }
    visitStatementExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitSelectStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitCase(node, context) {
        throw new Error("Method not implemented.");
    }
    visitConditionStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitIfStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitWhileStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitDoStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitForStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitForInit(node, context) {
        throw new Error("Method not implemented.");
    }
    visitReturnStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitPrimitiveStatement(node, context) {
        throw new Error("Method not implemented.");
    }
    visitPrintStatement(node, context) {
        throw new Error("Method not implemented.");
    }
}
export class PCCVariableInitialisation {
    // constructor(readonly compiler: PCCCompiler) { }
    // private static sharedInstance: PCCVariableInitialisation | null = null; 
    // static shared(compiler: PCCCompiler): PCCVariableInitialisation {
    //     if (!this.sharedInstance) {
    //         this.sharedInstance = new PCCVariableInitialisation(compiler);
    //     } else {
    //         if (this.sharedInstance.compiler != compiler) {
    //             throw new Error("Cannot have a shared instance for different compilers");
    //         }
    //     }
    //     return this.sharedInstance;
    // }
    static createContainerForPrimitiveType(compiler, type) {
        if (type.concretePrimitiveType == PCT.ConcretePrimitiveType.BOOL) {
            return new PCCConstantContainer(false);
        }
        else if (type.concretePrimitiveType == PCT.ConcretePrimitiveType.INT) {
            return new PCCConstantContainer(0);
        }
        else if (type.concretePrimitiveType == PCT.ConcretePrimitiveType.STRING) {
            return new PCCConstantContainer("");
        }
        else {
            // Void type is not inhabited
            throw new Error("Cannot create container for type 'void'");
        }
    }
    static createContainerForSystemType(compiler, type) {
        if (type.concreteSystemType == PCT.ConcreteSystemType.LOCK) {
            // Create a new lock object and store the reference in `result`
            const result = compiler.getFreshContainer(CCSType.INT);
            // Query the lock constructor (formerly called mutex)
            compiler.emitInput("mutex_create", null, result);
            return result;
        }
        else if (type.concreteSystemType == PCT.ConcreteSystemType.CONDITION) {
            // Query the wait room process
            const result = compiler.getFreshContainer(CCSType.INT);
            compiler.emitInput("wait_create", null, result);
            return result;
        }
        else {
            // There is no construction mechanism for agents, except for starting a new agent (which is not done here)
            throw new Error("Cannot create container for type " + type.toString());
        }
    }
    static createContainerForArrayType(compiler, type, initialElements = []) {
        // We can only create a new array instance if we know the array's capacity
        if (type.size == PCT.UNKNOWN_SIZE) {
            throw new Error("Cannot create container for array of unknown size");
        }
        // Create container that will hold the array reference
        const result = compiler.getFreshContainer(CCSType.fromPseucoType(type));
        // Create a new array using the array constructor process
        compiler.emitOutput(`array${type.size}_prepare`, null, null);
        compiler.emitInput(`array${type.size}_create`, null, result);
        // The creation of an array needs a value with which every element in the array is initialised. If the array is an array of objects, then the elements are initialised as "null references", although such a thing does not exist in pseuCo. If necessary, the null references are replaced later in this method.
        compiler.emitOutput("array_setDefault", result, CCSType.fromPseucoType(type.elementsType).getDefaultContainer());
        // In case the array holds values of complex data type, we must create a new instance of this type for each position in the array (minus the elements we already have via `initialElements`)
        if (this.typeRequiresCustomDefaultContainer(compiler, type.elementsType)) {
            for (let i = initialElements.length; i < type.size; i++) {
                initialElements.push(this.createContainerForType(compiler, type.elementsType));
            }
        }
        // Every non-default value must be stored in the array; i.e., every initialElement and every custom default container created above
        for (const [i, c] of initialElements.entries()) {
            compiler.emitOutput("array_access", result, new PCCConstantContainer(i));
            compiler.emitOutput("array_set", result, c);
        }
        return result;
    }
    static createContainerForChannelType(compiler, type) {
        // Create a new channel object and store reference in `res`
        const res = compiler.getFreshContainer(CCSType.fromPseucoType(type));
        // The channel constructor depends on whether the channel is buffered or a synchronous channel
        if (type.capacity != PCT.UNKNOWN_CAPACITY && type.capacity != 0) {
            compiler.emitOutput(`channel${type.capacity}_prepare`);
            compiler.emitInput(`channel${type.capacity}_create`, null, res);
        }
        else {
            compiler.emitInput("channel_create", null, res);
        }
        return res;
    }
    static createContainerForClassType(compiler, type) {
        // Create a new class object and store the reference in `result`
        const result = compiler.getFreshContainer(CCSType.fromPseucoType(type));
        // Query the constructor to create a new reference
        compiler.emitInput(`class_${type.className}_create`, null, result);
        compiler.emitInput(`class_${type.className}_initialized`, result, null);
        return result;
    }
    static typeRequiresCustomDefaultContainer(compiler, type) {
        // Everything which is not a primitive type needs some kind of initialisation into a custom container
        return !(type instanceof PCT.PrimitiveType);
    }
    static createContainerForType(compiler, type) {
        if (type instanceof PCT.PrimitiveType) {
            return this.createContainerForPrimitiveType(compiler, type);
        }
        else if (type instanceof PCT.SystemType) {
            return this.createContainerForSystemType(compiler, type);
        }
        else if (type instanceof PCT.ArrayType) {
            return this.createContainerForArrayType(compiler, type);
        }
        else if (type instanceof PCT.ChannelType) {
            return this.createContainerForChannelType(compiler, type);
        }
        else if (type instanceof PCT.ClassType) {
            return this.createContainerForClassType(compiler, type);
        }
        else {
            // DeclarationType or ProcedureType
            throw new Error("Cannot create container for type " + type.toString());
        }
    }
    static compileDefaultValue(compiler, variable) {
        const node = variable.node;
        if (node instanceof PC.VariableDeclarator) {
            if (node.initializer) {
                return this.compileVariableInitializer(compiler, node.initializer, variable.type);
            }
            else {
                return this.createContainerForType(compiler, variable.type);
            }
        }
        // In every other case we fall back to the default construction method
        return this.createContainerForType(compiler, variable.type);
    }
    static compileVariableInitializer(compiler, initializer, type) {
        if (initializer.arrayInitializers) {
            if (!(type instanceof PCT.ArrayType)) {
                throw new Error("Expected array type for array.");
            }
            const cc = initializer.arrayInitializers.map((c) => this.compileVariableInitializer(compiler, c, type.elementsType));
            return this.createContainerForArrayType(compiler, type, cc);
        }
        else {
            return utils.assertNonNull(compiler.expressionsVisitor.compileNode(initializer.expression));
        }
    }
}
//# sourceMappingURL=PCCCompileProgramVisitor.js.map