import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import * as CCS from "@pseuco/ccs-interpreter";
import { CCSType, PCCClass, PCCProcedure, PCCVariable, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
import { PCCGroupable, PCCProcessFrame } from "./PCCProcessFrame.js";
import { PCCApplicationPlaceholderStackElement, PCCChoiceStackElement, PCCParallelStackElement, PCCProcessDefinitionStackElement, PCCSequenceStackElement } from "./PCCCompilerStack.js";
import { PCCCompileProgramVisitor } from "./PCCCompileProgramVisitor.js";
import { PCCCompileStatementsVisitor } from "./PCCCompileStatementsVisitor.js";
import { PCCCompileExpressionsVisitor } from "./PCCCompileExpressionsVisitor.js";
export declare class PCCCompiler {
    readonly program: PC.Program;
    readonly staticSemantics: PCT.StaticSemantics;
    private stack;
    private readonly groupElements;
    private readonly controller;
    private readonly processDefinitions;
    private readonly procedureProcessDefinitions;
    private readonly systemProcesses;
    private readonly compilingNodes;
    readonly programVisitor: PCCCompileProgramVisitor;
    readonly statementsVisitor: PCCCompileStatementsVisitor;
    readonly expressionsVisitor: PCCCompileExpressionsVisitor;
    private needsReturn;
    private needsMutex;
    private needsWaitRoom;
    private needsArrayManager;
    private needsChannelManager;
    private needsAgentManager;
    private needsAgentJoiner;
    /**
     * Creates a new compiler object to compile `program`, which is well-typed according to `staticSemantics`.
     * @param program The program that shall be compiled to a CCS process.
     * @param staticSemantics The static semantics objects that has been used to check whether `program` is well-typed.
     */
    constructor(program: PC.Program, staticSemantics: PCT.StaticSemantics);
    /**
     * Report to the compiler that the pseuCo program that is being compiled has functions (or methods) that return a value.
     * Adds system process `Return`.
     */
    setNeedsReturn(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled uses at least one lock (formerly called mutex).
     * Adds system processes `Mutex[i, c, a]` and `Mutex_cons[next_i]`.
     */
    setNeedsMutex(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled has at least one `waitForCondition` instruction.
     * Adds system processes `WaitRoom[i, c]`, `WaitDistributor[i, c]`, and `WaitRoom_cons[next_i]`.
     */
    setNeedsWaitRoom(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled uses arrays; the array manager allows different array constructors (for different array sizes) to share the same reference domain. That is, when a new array is created, the array manager provides the new array's "pointer reference".
     * Adds system process `ArrayManager[next_i]`.
     */
    setNeedsArrayManager(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled uses buffered channels; the array manager allows different channel constructors (for different channel capacities) to share the same reference domain.
     * Adds system process `ChannelManager[next_i]`.
     */
    setNeedsChannelManager(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled starts agents other than the main agent; the agent manages the unique agent identifiers.
     * Adds system process `AgentManager[next_i]`.
     */
    setNeedsAgentManager(): void;
    /**
     * Report to the compiler that the pseuCo program that is being compiled waits on other agents' termination using `join`.
     * Adds system processes `AgentJoiner[a, c]` and `JoinDistributor[a, c]`.
     */
    setNeedsAgentJoiner(): void;
    /**
     * Reports to the compiler that a particular procedure is started as a separate agent.
     * @param procedure The procedure that is started as an agent.
     */
    processProcedureAsAgent(procedure: PCCProcedure): void;
    /**
     * Compiles the program to CCS.
     * @returns Returns a CCS object that holds all process definitions and the main system process.
     */
    compileProgram(): CCS.CCS;
    private __dumpUncheckedCCS;
    private assertEmptyGroupStack;
    private pushStackElement;
    private getSystem;
    private getRestrictedChannels;
    getVariableWithName(name: string): PCCVariable | null;
    getVariableWithNameOfClass(name: string, className: string | null, isInternal: boolean): PCCVariable | null;
    getProcedureWithName(name: string): PCCProcedure;
    getProcedureWithNameOfClass(name: string, className: string | null): PCCProcedure;
    private getClassWithName;
    getCurrentClass(): PCCClass | null;
    getCurrentProcedure(): PCCProcedure | null;
    private getGlobal;
    private rememberPseucoCCSRelationship;
    getFreshContainer(ccsType: CCSType, desiredName?: string): PCCVariableContainer;
    handleNewVariable(variable: PCCVariableInfo): void;
    private getControlElement;
    private handleStackResult;
    private beginSystemProcess;
    private endSystemProcess;
    emitSystemProcessApplication(processName: string, argumentContainers: Array<PCCContainer>, weight: number): void;
    beginProcessGroup(groupable: PCCGroupable, variables: Array<PCCVariableInfo>): void;
    endProcessGroup(): void;
    getProcessFrame(): PCCProcessFrame | null;
    private addProcessGroupFrame;
    emitNewScope(derivationFrame?: PCCProcessFrame): PCCProcessFrame;
    emitNextProcessFrame(derivationFrames?: Array<PCCProcessFrame>): PCCProcessFrame;
    emitMergeOfProcessFramesOfPlaceholders(placeholders: Array<PCCApplicationPlaceholderStackElement>): PCCProcessFrame;
    protectContainer(container: PCCContainer): void;
    unprotectContainer(): PCCContainer;
    getProtectedContainer(): PCCContainer;
    silentlyAddProcessDefinition(processName: string, argumentContainers: Array<PCCVariableContainer>): PCCProcessDefinitionStackElement;
    beginProcessDefinition(processName: string, argumentContainers: Array<PCCVariableContainer>): void;
    isCurrentProcessCompleted(): boolean;
    endProcessDefinition(): void;
    beginClass(className: string): void;
    endClass(): void;
    beginMainAgent(node: PC.MainAgent): void;
    endMainAgent(): void;
    beginProcedure(procedureName: string): void;
    endProcedure(): void;
    reopenEnvironment(node: PC.Node): void;
    closeEnvironment(): void;
    private usingFrames;
    emitStop(): void;
    emitExit(): void;
    emitProcessApplication(processName: string, argumentContainers?: Array<PCCContainer>): void;
    emitSimplePrefix(channel: string, channelParameter?: PCCContainer | null): void;
    emitOutput(channel: string, channelParameter?: PCCContainer | null, valueContainer?: PCCContainer | null): void;
    emitInput(channel: string, channelParameter?: PCCContainer | null, container?: PCCVariableContainer | null): void;
    emitMatch(channel: string, channelParameter: PCCContainer | null, valueContainer: PCCContainer): void;
    emitCondition(condition: PCCContainer): void;
    emitChoice(): PCCChoiceStackElement;
    emitParallel(): PCCParallelStackElement;
    emitSequence(): PCCSequenceStackElement;
    emitRestriction(restrictedChannelNames: Array<string>): void;
    emitProcessApplicationPlaceholder(): PCCApplicationPlaceholderStackElement;
    private compileMainAgentApplication;
    private compileGlobalVariableInitialisation;
    compileSimpleMutex(): void;
    compileReentrantMutex(): void;
    compileMutexCons(reentrantMutex: boolean): void;
    compileWaitRoom(): void;
    compileThrowException(...comps: Array<string | number | boolean | PCCContainer>): void;
    compileArrayWithCapacity(size: number): void;
    compileArrayManager(): void;
    compileChannelWithCapacity(capacity: number): void;
    compileUnbufferedChannelCons(): void;
    compileChannelManager(): void;
    compileAgentTools(addAgentJoiner: boolean): void;
    compileReturn(): void;
    static SysAgent: number;
    static SysEnvironment: number;
    static SysChannel: number;
    static SysArray: number;
    static SysMutexCons: number;
    static SysWaitRoom: number;
    static SysInstanceManager: number;
    static SysUnknownWeight: number;
}
