import * as CCS from "@pseuco/ccs-interpreter";
import * as PC from "@pseuco/lang";
import { PCCCompiler } from "./PCCCompiler.js";
import { PCCProcessFrame } from "./PCCProcessFrame.js";
import { PCCClass, PCCGlobal, PCCProcedure, PCCVariable, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
type CCSNode = CCS.CCSProcess | CCS.CCSProcessDefinition | CCS.CCSAction | CCS.CCSExpression;
export declare class PCCStackResult {
    readonly type: PCCStackResultType;
    readonly data: CCS.CCSProcess | CCS.CCSProcessDefinition;
    constructor(type: PCCStackResultType, data: CCS.CCSProcess | CCS.CCSProcessDefinition);
    get ccsProcess(): CCS.CCSProcess;
    get ccsProcessDefinition(): CCS.CCSProcessDefinition;
}
export declare class PCCStackResultContainer {
    private results;
    get allResults(): ReadonlyArray<PCCStackResult>;
    addResult(type: PCCStackResultType, data: CCS.CCSProcess | CCS.CCSProcessDefinition): void;
    replaceResult(type: PCCStackResultType, data: CCS.CCSProcess | CCS.CCSProcessDefinition): void;
    getResult(): PCCStackResult;
    removeResult(): void;
    appendContainer(container: PCCStackResultContainer): void;
}
export type PCCControlElement = PCCGlobal | PCCProcedure | PCCClass;
export declare abstract class PCCStackElement {
    parent: PCCStackElement | null;
    pseucoNode: PC.Node | null;
    private myStack;
    get stack(): PCCCompilerStack;
    set stack(v: PCCCompilerStack);
    abstract getResults(): PCCStackResultContainer;
    abstract getNext(): PCCStackElement | null;
    abstract setNext(next: PCCStackElement): void;
    abstract removeNext(next: PCCStackElement): void;
    getTopElement(): PCCStackElement;
    getCurrentProcessFrame(): PCCProcessFrame | null;
    getCurrentControlElement(): PCCControlElement | null;
    removeFromStack(): PCCStackResultContainer;
    isCompletedProcess(): boolean;
    rememberPseucoCalculusRelationship(ccsNode: CCSNode): void;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
}
export declare class PCCUnaryStackElement extends PCCStackElement {
    private next;
    getResults(): PCCStackResultContainer;
    getNext(): PCCStackElement | null;
    setNext(next: PCCStackElement): void;
    removeNext(next: PCCStackElement): void;
}
export declare enum PCCBinaryTarget {
    LEFT = 0,
    RIGHT = 1,
    TOP = 2
}
export declare abstract class PCCBinaryStackElement extends PCCStackElement {
    protected leftStack: PCCStackElement | null;
    protected rightStack: PCCStackElement | null;
    protected topStack: PCCStackElement | null;
    private target;
    getNext(): PCCStackElement | null;
    private setOrRemoveNext;
    setNext(next: PCCStackElement): void;
    removeNext(next: PCCStackElement): void;
    setTarget(target: PCCBinaryTarget): void;
    setBranchFinished(): void;
    updateBinaryTargets(destination: PCCStackElement): void;
    isCompletedProcess(): boolean;
}
export declare enum PCCStackResultType {
    UNSPECIFIC = 0,
    CCSPROCESS = 1,
    CCSPROCESS_DEFINITION = 2,
    CCSPROCESS_DEFINITION_FOR_PROCEDURE = 3
}
export declare class PCCStopStackElement extends PCCUnaryStackElement {
    getResults(): PCCStackResultContainer;
    isCompletedProcess(): boolean;
}
export declare class PCCExitStackElement extends PCCUnaryStackElement {
    getResults(): PCCStackResultContainer;
    isCompletedProcess(): boolean;
}
export declare class PCCPrefixStackElement extends PCCUnaryStackElement {
    readonly channel: string;
    readonly channelParameter: PCCContainer | null;
    constructor(channel: string, channelParameter?: PCCContainer | null);
    protected getChannel(): CCS.CCSChannel;
    protected getAction(): CCS.CCSAction;
    getResults(): PCCStackResultContainer;
}
export declare class PCCInputStackElement extends PCCPrefixStackElement {
    readonly value: PCCVariableContainer | null;
    constructor(channel: string, channelParameter?: PCCContainer | null, value?: PCCVariableContainer | null);
    protected getAction(): CCS.CCSAction;
}
export declare class PCCOutputStackElement extends PCCPrefixStackElement {
    readonly value: PCCContainer | null;
    constructor(channel: string, channelParameter?: PCCContainer | null, value?: PCCContainer | null);
    protected getAction(): CCS.CCSAction;
}
export declare class PCCMatchStackElement extends PCCPrefixStackElement {
    readonly value: PCCContainer;
    constructor(channel: string, channelParameter: PCCContainer | null, value: PCCContainer);
    protected getAction(): CCS.CCSAction;
}
export declare class PCCConditionStackElement extends PCCUnaryStackElement {
    readonly conditionContainer: PCCContainer;
    constructor(conditionContainer: PCCContainer);
    getResults(): PCCStackResultContainer;
}
export declare class PCCRestrictionStackElement extends PCCUnaryStackElement {
    readonly restrictedChannels: ReadonlyArray<string>;
    constructor(restrictedChannels: ReadonlyArray<string>);
    getResults(): PCCStackResultContainer;
}
export declare class PCCApplicationStackElement extends PCCUnaryStackElement {
    readonly processName: string;
    readonly argContainers: ReadonlyArray<PCCContainer>;
    constructor(processName: string, argContainers?: ReadonlyArray<PCCContainer>);
    getResults(): PCCStackResultContainer;
    static computeResults(stackElement: PCCApplicationStackElement | PCCApplicationPlaceholderStackElement): PCCStackResultContainer;
    isCompletedProcess(): boolean;
}
export declare class PCCApplicationPlaceholderStackElement extends PCCUnaryStackElement {
    readonly frame: PCCProcessFrame;
    constructor(frame: PCCProcessFrame);
    private appStackElement;
    get applicationStackElement(): PCCApplicationStackElement | null;
    set(processName: string, argContainers?: ReadonlyArray<PCCContainer>): void;
    get processName(): string | undefined;
    get argContainers(): readonly PCCContainer[] | undefined;
    getResults(): PCCStackResultContainer;
    isCompletedProcess(): boolean;
}
export declare abstract class PCCBinaryCCSStackElement extends PCCBinaryStackElement {
    protected abstract createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess): CCS.CCSProcess;
    getResults(): PCCStackResultContainer;
    isCompletedProcess(): boolean;
}
export declare class PCCChoiceStackElement extends PCCBinaryCCSStackElement {
    protected createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess): CCS.CCSChoice;
}
export declare class PCCParallelStackElement extends PCCBinaryCCSStackElement {
    protected createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess): CCS.CCSParallel;
}
export declare class PCCSequenceStackElement extends PCCBinaryCCSStackElement {
    protected createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess): CCS.CCSSequence;
}
export declare class PCCSystemProcessStackElement extends PCCUnaryStackElement {
    readonly weight: number;
    constructor(weight: number);
    getResults(): PCCStackResultContainer;
}
export declare class PCCProcessDefinitionStackElement extends PCCUnaryStackElement {
    readonly processName: string;
    readonly argContainers: Array<PCCVariableContainer>;
    private isProcedure;
    get isMarkedAsProcedure(): boolean;
    constructor(processName: string, argContainers: Array<PCCVariableContainer>);
    markAsProcedure(): void;
    getResults(): PCCStackResultContainer;
}
export declare class PCCProcessFrameStackElement extends PCCUnaryStackElement {
    readonly frame: PCCProcessFrame;
    constructor(frame: PCCProcessFrame);
    getCurrentProcessFrame(): PCCProcessFrame | null;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
    isCompletedProcess(): boolean;
}
export declare class PCCClassStackElement extends PCCUnaryStackElement {
    readonly classInfo: PCCClass;
    constructor(classInfo: PCCClass);
    getCurrentControlElement(): PCCClass;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
}
export declare class PCCGlobalStackElement extends PCCUnaryStackElement {
    readonly global: PCCGlobal;
    constructor(global: PCCGlobal);
    getCurrentControlElement(): PCCGlobal;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
}
export declare class PCCProcedureStackElement extends PCCUnaryStackElement {
    readonly procedure: PCCProcedure;
    constructor(procedure: PCCProcedure);
    getCurrentControlElement(): PCCProcedure;
}
export declare class PCCCompilerStack {
    topElement: PCCStackElement | null;
    getCurrentProcessFrame(): PCCProcessFrame | null;
    getCurrentControlElement(): PCCControlElement | null;
    pushElement(e: PCCStackElement): void;
    isCurrentProcessCompleted(): boolean;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
}
export {};
