import * as CCS from "@pseuco/ccs-interpreter";
export class PCCStackResult {
    type;
    data;
    constructor(type, data) {
        this.type = type;
        this.data = data;
        if (this.type == PCCStackResultType.CCSPROCESS) {
            if (!(this.data instanceof CCS.CCSProcess)) {
                throw new Error("Type mismatch!");
            }
        }
        else if (this.type == PCCStackResultType.CCSPROCESS_DEFINITION || this.type == PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE) {
            if (!(this.data instanceof CCS.CCSProcessDefinition)) {
                throw new Error("Type mismatch!");
            }
        }
        else {
            throw new Error("Unknown result type!");
        }
    }
    get ccsProcess() {
        if (this.type == PCCStackResultType.CCSPROCESS) {
            return this.data;
        }
        else {
            throw new Error("Result is not a CCS process.");
        }
    }
    get ccsProcessDefinition() {
        if (this.type == PCCStackResultType.CCSPROCESS_DEFINITION || this.type == PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE) {
            return this.data;
        }
        else {
            throw new Error("Result is not a CCS process definition.");
        }
    }
}
export class PCCStackResultContainer {
    results = new Array();
    // ToDo: Do we really need an array here? Can't we just drop whatever we would at index >= 1?
    get allResults() {
        return this.results;
    }
    addResult(type, data) {
        this.results.unshift(new PCCStackResult(type, data));
    }
    replaceResult(type, data) {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        this.results[0] = new PCCStackResult(type, data);
    }
    getResult() {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        return this.results[0];
    }
    removeResult() {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        this.results.shift();
    }
    appendContainer(container) {
        this.results = this.results.concat(container.results);
    }
}
export class PCCStackElement {
    parent = null;
    pseucoNode = null;
    myStack = null;
    get stack() {
        if (!this.myStack) {
            throw new Error("Stack has not been set.");
        }
        return this.myStack;
    }
    set stack(v) {
        if (this.myStack && v != this.myStack) {
            throw new Error("Stack can be set only once!");
        }
        this.myStack = v;
    }
    getTopElement() {
        const next = this.getNext();
        if (next == null) {
            return this;
        }
        else {
            return next.getTopElement();
        }
    }
    getCurrentProcessFrame() {
        if (this.parent == null) {
            return null;
        }
        else {
            return this.parent.getCurrentProcessFrame();
        }
    }
    getCurrentControlElement() {
        if (this.parent == null) {
            return null;
        }
        else {
            return this.parent.getCurrentControlElement();
        }
    }
    removeFromStack() {
        this.stack.topElement = this.parent;
        if (this.parent != null) {
            this.parent.removeNext(this);
        }
        return this.getResults();
    }
    isCompletedProcess() {
        // ! This is a correction of the old implementation
        return this.getNext()?.isCompletedProcess() ?? false;
    }
    // createCalculusNode ???
    rememberPseucoCalculusRelationship(ccsNode) {
        // Todo
    }
    compilerGetVariable(compiler, identifier) {
        if (this.parent != null) {
            return this.parent.compilerGetVariable(compiler, identifier);
        }
        else {
            return null;
        }
    }
    compilerGetProcedure(compiler, identifier) {
        if (this.parent != null) {
            return this.parent.compilerGetProcedure(compiler, identifier);
        }
        else {
            return null;
        }
    }
    // Returns true if the new variable could be handled.
    compilerHandleNewVariable(compiler, variable) {
        if (this.parent != null) {
            this.parent.compilerHandleNewVariable(compiler, variable);
        }
        else {
            //return null;
        }
    }
}
export class PCCUnaryStackElement extends PCCStackElement {
    next = null;
    getResults() {
        if (this.next == null) {
            return new PCCStackResultContainer();
        }
        else {
            return this.next.getResults();
        }
    }
    getNext() {
        return this.next;
    }
    setNext(next) {
        if (this.next != null) {
            throw new Error("Can't set next twice!");
        }
        next.parent = this;
        this.next = next;
    }
    removeNext(next) {
        if (this.next != next) {
            throw new Error("Unknown next");
        }
        next.parent = null;
        this.next = null;
    }
}
export var PCCBinaryTarget;
(function (PCCBinaryTarget) {
    PCCBinaryTarget[PCCBinaryTarget["LEFT"] = 0] = "LEFT";
    PCCBinaryTarget[PCCBinaryTarget["RIGHT"] = 1] = "RIGHT";
    PCCBinaryTarget[PCCBinaryTarget["TOP"] = 2] = "TOP";
})(PCCBinaryTarget || (PCCBinaryTarget = {}));
export class PCCBinaryStackElement extends PCCStackElement {
    leftStack = null;
    rightStack = null;
    topStack = null;
    target = PCCBinaryTarget.LEFT;
    getNext() {
        if (this.target == PCCBinaryTarget.LEFT) {
            return this.leftStack;
        }
        else if (this.target == PCCBinaryTarget.RIGHT) {
            return this.rightStack;
        }
        else {
            return this.topStack;
        }
    }
    setOrRemoveNext(next) {
        if (next != null) {
            next.parent = this;
        }
        if (this.target == PCCBinaryTarget.LEFT) {
            if (this.leftStack != null) {
                throw new Error("Can't set next twice!");
            }
            this.leftStack = next;
        }
        else if (this.target == PCCBinaryTarget.RIGHT) {
            if (this.rightStack != null) {
                throw new Error("Can't set next twice!");
            }
            this.rightStack = next;
        }
        else {
            if (this.topStack != null) {
                throw new Error("Can't set next twice!");
            }
            this.topStack = next;
        }
    }
    setNext(next) {
        this.setOrRemoveNext(next);
    }
    removeNext(next) {
        if (next != this.getNext()) {
            throw new Error("Unknown next");
        }
        this.setOrRemoveNext(null);
    }
    setTarget(target) {
        if (target < 0 || target > 2) {
            throw new Error("Illegal target!");
        }
        if (target != PCCBinaryTarget.TOP && this.topStack != null) {
            throw new Error("Left and right target are not allowed once top was modified!");
        }
        this.target = target;
        this.stack.topElement = this.getTopElement();
        // ToDo: Unclear why the following should be necessary:
        // let mom = this.parent;
        // while (mom != null) {
        //     if (mom instanceof PCCBinaryStackElement) {
        //         mom.updateBinaryTargets(this);
        //         break;
        //     } else {
        //         mom = mom.parent;
        //     }
        // }
    }
    setBranchFinished() {
        if (this.target == PCCBinaryTarget.LEFT) {
            this.setTarget(PCCBinaryTarget.RIGHT);
        }
        else if (this.target == PCCBinaryTarget.RIGHT) {
            this.setTarget(PCCBinaryTarget.TOP);
        }
        else {
            throw new Error("All branches are already finished!");
        }
    }
    updateBinaryTargets(destination) {
        if (destination == this.leftStack) {
            this.target = PCCBinaryTarget.LEFT;
        }
        else if (destination == this.rightStack) {
            this.target = PCCBinaryTarget.RIGHT;
        }
        else if (destination == this.topStack) {
            this.target = PCCBinaryTarget.TOP;
        }
        else {
            throw new Error("Unknown destination!");
        }
    }
    isCompletedProcess() {
        // ! This override was missing in old implementation
        if (this.leftStack != null && this.rightStack != null && this.topStack != null) {
            return this.leftStack.isCompletedProcess() && this.rightStack.isCompletedProcess() && this.topStack.isCompletedProcess();
        }
        else {
            return false;
        }
    }
}
export var PCCStackResultType;
(function (PCCStackResultType) {
    PCCStackResultType[PCCStackResultType["UNSPECIFIC"] = 0] = "UNSPECIFIC";
    PCCStackResultType[PCCStackResultType["CCSPROCESS"] = 1] = "CCSPROCESS";
    PCCStackResultType[PCCStackResultType["CCSPROCESS_DEFINITION"] = 2] = "CCSPROCESS_DEFINITION";
    PCCStackResultType[PCCStackResultType["CCSPROCESS_DEFINITION_FOR_PROCEDURE"] = 3] = "CCSPROCESS_DEFINITION_FOR_PROCEDURE";
})(PCCStackResultType || (PCCStackResultType = {}));
export class PCCStopStackElement extends PCCUnaryStackElement {
    getResults() {
        const more = (this.getNext() == null) ? new PCCStackResultContainer() : this.getNext().getResults();
        const stopProcess = new CCS.CCSStop();
        this.rememberPseucoCalculusRelationship(stopProcess);
        more.addResult(PCCStackResultType.CCSPROCESS, stopProcess);
        return more;
    }
    isCompletedProcess() {
        return true;
    }
}
export class PCCExitStackElement extends PCCUnaryStackElement {
    getResults() {
        const more = (this.getNext() == null) ? new PCCStackResultContainer() : this.getNext().getResults();
        const exitProcess = new CCS.CCSExit();
        this.rememberPseucoCalculusRelationship(exitProcess);
        more.addResult(PCCStackResultType.CCSPROCESS, exitProcess);
        return more;
    }
    isCompletedProcess() {
        return true;
    }
}
export class PCCPrefixStackElement extends PCCUnaryStackElement {
    channel;
    channelParameter;
    // chan(i). 0 --> `chan` is channel, `i` is channel parameter
    constructor(channel, channelParameter = null) {
        super();
        this.channel = channel;
        this.channelParameter = channelParameter;
    }
    getChannel() {
        return new CCS.CCSChannel(this.channel, this.channelParameter?.ccsTree() ?? null);
    }
    getAction() {
        const res = new CCS.CCSSimpleAction(this.getChannel());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
    getResults() {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext().getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const prefixProcess = new CCS.CCSPrefix(this.getAction(), pRes.ccsProcess);
        this.rememberPseucoCalculusRelationship(prefixProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, prefixProcess);
        return container;
    }
}
export class PCCInputStackElement extends PCCPrefixStackElement {
    value;
    constructor(channel, channelParameter = null, value = null) {
        super(channel, channelParameter);
        this.value = value;
    }
    getAction() {
        let variable = null;
        if (this.value != null) {
            variable = this.value.identifier;
            // this.rememberPseucoCalculusRelationship(variable);
        }
        const res = new CCS.CCSInput(this.getChannel(), variable, new CCS.CCSUnrestrictedValueSet());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCOutputStackElement extends PCCPrefixStackElement {
    value;
    constructor(channel, channelParameter = null, value = null) {
        super(channel, channelParameter);
        this.value = value;
    }
    getAction() {
        const res = new CCS.CCSOutput(this.getChannel(), (this.value == null) ? null : this.value.ccsTree());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCMatchStackElement extends PCCPrefixStackElement {
    value;
    constructor(channel, channelParameter, value) {
        super(channel, channelParameter);
        this.value = value;
    }
    getAction() {
        const res = new CCS.CCSMatch(this.getChannel(), this.value.ccsTree());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCConditionStackElement extends PCCUnaryStackElement {
    conditionContainer;
    constructor(conditionContainer) {
        super();
        this.conditionContainer = conditionContainer;
    }
    getResults() {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext().getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const conditionProcess = new CCS.CCSCondition(this.conditionContainer.ccsTree(), pRes.ccsProcess);
        this.rememberPseucoCalculusRelationship(conditionProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, conditionProcess);
        return container;
    }
}
export class PCCRestrictionStackElement extends PCCUnaryStackElement {
    restrictedChannels;
    constructor(restrictedChannels) {
        super();
        this.restrictedChannels = restrictedChannels;
    }
    getResults() {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext().getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const restrictionProcess = new CCS.CCSRestriction(pRes.ccsProcess, this.restrictedChannels.slice());
        this.rememberPseucoCalculusRelationship(restrictionProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, restrictionProcess);
        return container;
    }
}
export class PCCApplicationStackElement extends PCCUnaryStackElement {
    processName;
    argContainers;
    constructor(processName, argContainers = []) {
        super();
        this.processName = processName;
        this.argContainers = argContainers.slice();
    }
    getResults() {
        return PCCApplicationStackElement.computeResults(this);
    }
    static computeResults(stackElement) {
        const more = stackElement.getNext()?.getResults() ?? new PCCStackResultContainer();
        const argumentValues = stackElement.argContainers.map((c) => c.ccsTree());
        const process = new CCS.CCSProcessApplication(stackElement.processName, argumentValues);
        stackElement.rememberPseucoCalculusRelationship(process);
        more.addResult(PCCStackResultType.CCSPROCESS, process);
        return more;
    }
    isCompletedProcess() {
        return true;
    }
}
export class PCCApplicationPlaceholderStackElement extends PCCUnaryStackElement {
    frame;
    constructor(frame) {
        super();
        this.frame = frame;
    }
    appStackElement = null;
    get applicationStackElement() {
        return this.appStackElement;
    }
    set(processName, argContainers = []) {
        if (this.appStackElement != null) {
            throw new Error("Placeholder values can't be set twice!");
        }
        this.appStackElement = new PCCApplicationStackElement(processName, argContainers);
    }
    get processName() {
        return this.appStackElement?.processName;
    }
    get argContainers() {
        return this.appStackElement?.argContainers;
    }
    getResults() {
        if (this.appStackElement == null) {
            throw new Error("Placeholder without values!");
        }
        else {
            return PCCApplicationStackElement.computeResults(this);
        }
    }
    isCompletedProcess() {
        return true;
    }
}
export class PCCBinaryCCSStackElement extends PCCBinaryStackElement {
    getResults() {
        const more = this.topStack?.getResults() ?? new PCCStackResultContainer();
        const leftContainer = this.leftStack?.getResults();
        if (!leftContainer) {
            throw new Error("Left process not available.");
        }
        const left = leftContainer.getResult();
        // if (left.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const rightContainer = this.rightStack?.getResults();
        if (!rightContainer) {
            throw new Error("Right process not available.");
        }
        const right = rightContainer.getResult();
        // if (right.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        rightContainer.removeResult();
        leftContainer.appendContainer(rightContainer); // unclear why this happens...
        leftContainer.appendContainer(more);
        leftContainer.replaceResult(PCCStackResultType.CCSPROCESS, this.createCCSProcess(left.ccsProcess, right.ccsProcess));
        return leftContainer;
    }
    isCompletedProcess() {
        if (this.leftStack != null && this.rightStack != null) {
            return this.leftStack.isCompletedProcess() && this.rightStack.isCompletedProcess();
        }
        else {
            return false;
        }
    }
}
export class PCCChoiceStackElement extends PCCBinaryCCSStackElement {
    createCCSProcess(left, right) {
        const res = new CCS.CCSChoice(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCParallelStackElement extends PCCBinaryCCSStackElement {
    createCCSProcess(left, right) {
        const res = new CCS.CCSParallel(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCSequenceStackElement extends PCCBinaryCCSStackElement {
    createCCSProcess(left, right) {
        const res = new CCS.CCSSequence(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}
export class PCCSystemProcessStackElement extends PCCUnaryStackElement {
    weight;
    constructor(weight) {
        super();
        this.weight = weight;
    }
    getResults() {
        const container = this.getNext()?.getResults();
        if (!container) {
            throw new Error("Expected a `next` element.");
        }
        const pRes = container.getResult();
        if (pRes.type != PCCStackResultType.CCSPROCESS) {
            throw new Error("Unexpected result type!");
        }
        // return pRes;
        return container;
    }
}
export class PCCProcessDefinitionStackElement extends PCCUnaryStackElement {
    processName;
    argContainers;
    isProcedure = false;
    get isMarkedAsProcedure() {
        return this.isProcedure;
    }
    constructor(processName, argContainers) {
        super();
        this.processName = processName;
        this.argContainers = argContainers.slice();
    }
    markAsProcedure() {
        this.isProcedure = true;
    }
    getResults() {
        const container = this.getNext()?.getResults();
        if (!container) {
            throw new Error("Expected a `next` element.");
        }
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        // const argCCSVariables = this.argContainers.map((c) => {
        //     const ccsNode = new CCS.Variable(c.identifier);
        //     this.rememberPseucoCalculusRelationship(ccsNode);
        //     return ccsNode;
        // });
        const argCCSVariables = this.argContainers.map((c) => c.identifier);
        const ccsDef = new CCS.CCSProcessDefinition(this.processName, pRes.ccsProcess, argCCSVariables);
        const resultType = (this.isMarkedAsProcedure) ? PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE : PCCStackResultType.CCSPROCESS_DEFINITION;
        container.replaceResult(resultType, ccsDef);
        return container;
    }
}
export class PCCProcessFrameStackElement extends PCCUnaryStackElement {
    frame;
    constructor(frame) {
        super();
        this.frame = frame;
    }
    getCurrentProcessFrame() {
        return this.frame;
    }
    compilerGetVariable(compiler, identifier) {
        return this.frame.compilerGetVariable(compiler, identifier) ?? super.compilerGetVariable(compiler, identifier);
    }
    compilerGetProcedure(compiler, identifier) {
        return this.frame.compilerGetProcedure(compiler, identifier) ?? super.compilerGetProcedure(compiler, identifier);
    }
    compilerHandleNewVariable(compiler, variable) {
        this.frame.compilerHandleNewVariable(compiler, variable);
    }
    isCompletedProcess() {
        return this.getNext()?.isCompletedProcess() ?? false;
    }
}
export class PCCClassStackElement extends PCCUnaryStackElement {
    classInfo;
    constructor(classInfo) {
        super();
        this.classInfo = classInfo;
    }
    getCurrentControlElement() {
        return this.classInfo;
    }
    compilerGetVariable(compiler, identifier) {
        return this.classInfo.compilerGetVariable(compiler, identifier) ?? super.compilerGetVariable(compiler, identifier);
    }
    compilerGetProcedure(compiler, identifier) {
        return this.classInfo.compilerGetProcedure(compiler, identifier) ?? super.compilerGetProcedure(compiler, identifier);
    }
    compilerHandleNewVariable(compiler, variable) {
        this.classInfo.compilerHandleNewVariable(compiler, variable);
    }
}
export class PCCGlobalStackElement extends PCCUnaryStackElement {
    global;
    constructor(global) {
        super();
        this.global = global;
    }
    getCurrentControlElement() {
        return this.global;
    }
    compilerGetVariable(compiler, identifier) {
        return this.global.compilerGetVariable(compiler, identifier); // ?? super.compilerGetVariable(compiler, identifier);
    }
    compilerGetProcedure(compiler, identifier) {
        return this.global.compilerGetProcedure(compiler, identifier);
    }
    compilerHandleNewVariable(compiler, variable) {
        this.global.compilerHandleNewVariable(compiler, variable);
    }
}
export class PCCProcedureStackElement extends PCCUnaryStackElement {
    procedure;
    constructor(procedure) {
        super();
        this.procedure = procedure;
    }
    getCurrentControlElement() {
        return this.procedure;
    }
}
export class PCCCompilerStack {
    topElement = null;
    getCurrentProcessFrame() {
        return this.topElement?.getCurrentProcessFrame() ?? null;
    }
    getCurrentControlElement() {
        return this.topElement?.getCurrentControlElement() ?? null;
    }
    pushElement(e) {
        e.stack = this;
        this.topElement?.setNext(e);
        this.topElement = e;
    }
    isCurrentProcessCompleted() {
        return this.topElement?.isCompletedProcess() ?? false;
    }
    compilerGetVariable(compiler, identifier) {
        return this.topElement?.compilerGetVariable(compiler, identifier) ?? null;
    }
    compilerGetProcedure(compiler, identifier) {
        return this.topElement?.compilerGetProcedure(compiler, identifier) ?? null;
    }
    compilerHandleNewVariable(compiler, variable) {
        // return this.topElement?.compilerHandleNewVariable(compiler, variable) ?? null;
        this.topElement?.compilerHandleNewVariable(compiler, variable);
    }
}
//# sourceMappingURL=PCCCompilerStack.js.map