import { CCSType } from "./PCCProgramController.js";
import * as CCS from "@pseuco/ccs-interpreter";
// Containers represent arithmetic expressions in CCS
export class PCCContainer {
    ccsType;
    constructor(ccsType) {
        this.ccsType = ccsType;
    }
}
export class PCCConstantContainer extends PCCContainer {
    value;
    constructor(value) {
        let inferredCCSType;
        if (typeof value == "boolean") {
            inferredCCSType = CCSType.BOOL;
        }
        else if (typeof value == "number") {
            inferredCCSType = CCSType.INT;
        }
        else if (typeof value == "string") {
            inferredCCSType = CCSType.STRING;
        }
        else {
            throw new Error("Unknown value type");
        }
        super(inferredCCSType);
        this.value = value;
    }
    isReadonly() {
        return true;
    }
    ccsTree() {
        return new CCS.CCSConstantExpression(this.value);
    }
    isEqual(container) {
        return container instanceof PCCConstantContainer && container.value === this.value;
    }
}
export class PCCVariableContainer extends PCCContainer {
    identifier;
    constructor(identifier, ccsType) {
        super(ccsType);
        this.identifier = identifier;
    }
    isReadonly() {
        return false;
    }
    ccsTree() {
        return new CCS.CCSVariableExpression(this.identifier);
    }
    isEqual(container) {
        return container instanceof PCCVariableContainer && container.identifier == this.identifier;
    }
}
export class PCCBinaryContainer extends PCCContainer {
    leftContainer;
    rightContainer;
    operator;
    _ccsTree;
    constructor(leftContainer, rightContainer, operator) {
        let inferredCCSType;
        let ccsTree;
        if (operator == "+" || operator == "^") {
            if (leftContainer.ccsType.isString() || rightContainer.ccsType.isString() || operator == "^") {
                ccsTree = new CCS.CCSConcatenatingExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
                inferredCCSType = CCSType.STRING;
            }
            else {
                ccsTree = new CCS.CCSAdditiveExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
                inferredCCSType = CCSType.INT;
            }
        }
        else if (operator == "-") {
            ccsTree = new CCS.CCSAdditiveExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.INT;
        }
        else if (operator == "*" || operator == "/" || operator == "%") {
            ccsTree = new CCS.CCSMultiplicativeExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.INT;
        }
        else if (operator == "<" || operator == "<=" || operator == ">" || operator == ">=") {
            ccsTree = new CCS.CCSRelationalExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.BOOL;
        }
        else if (operator == "==" || operator == "!=") {
            ccsTree = new CCS.CCSEqualityExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.BOOL;
        }
        else if (operator == "&&") {
            ccsTree = new CCS.CCSAndExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
            inferredCCSType = CCSType.BOOL;
        }
        else if (operator == "||") {
            ccsTree = new CCS.CCSOrExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
            inferredCCSType = CCSType.BOOL;
        }
        else {
            throw Error("Unknown operator");
        }
        super(inferredCCSType);
        this.leftContainer = leftContainer;
        this.rightContainer = rightContainer;
        this.operator = operator;
        this._ccsTree = ccsTree;
    }
    isReadonly() {
        return true;
    }
    ccsTree() {
        return this._ccsTree;
    }
    isEqual(container) {
        return container instanceof PCCBinaryContainer && container.operator == this.operator && container.leftContainer.isEqual(this.leftContainer) && container.rightContainer.isEqual(this.rightContainer);
    }
}
export class PCCUnaryContainer extends PCCContainer {
    operator;
    container;
    _ccsTree;
    constructor(operator, container) {
        let inferredCCSType;
        let ccsTree;
        if (operator == "!") {
            ccsTree = new CCS.CCSComplementExpression(container.ccsTree());
            inferredCCSType = CCSType.BOOL;
        }
        else if (operator == "-") {
            ccsTree = new CCS.CCSAdditiveExpression(new CCS.CCSConstantExpression(0), container.ccsTree(), "-");
            inferredCCSType = CCSType.INT;
        }
        else if (operator == "+") {
            ccsTree = container.ccsTree();
            inferredCCSType = CCSType.INT;
        }
        else {
            throw Error("Unknown operator");
        }
        super(inferredCCSType);
        this.operator = operator;
        this.container = container;
        this._ccsTree = ccsTree;
    }
    isReadonly() {
        return true;
    }
    ccsTree() {
        return this._ccsTree;
    }
    isEqual(container) {
        return container instanceof PCCUnaryContainer && container.operator == this.operator && container.container.isEqual(this.container);
    }
}
//# sourceMappingURL=PCCContainer.js.map