/***
 * Manages multiple versions of process frames. A process frame is the set of all arguments, local variables and temporary containers of a compile unit, also called a process groupable (e.g. a procedure) at a specific time.
 * Process frames may exist in multiple versions: One version for each CCS process. Because of restrictions of CCS it is not always possible to compile into a single process. This class helps to manage the fragmentation of code into multiple processes.
 * Versions are partially ordered.
 *
 * ***/
import { PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
import { CCSType, PCCLocalVariable, PCCProcedure, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCCompiler } from "./PCCCompiler.js";
import { PCCApplicationPlaceholderStackElement, PCCProcessDefinitionStackElement } from "./PCCCompilerStack.js";
export interface PCCGroupable {
    getProcessName(): string;
}
export declare class PCCSimpleGroupable implements PCCGroupable {
    private processName;
    constructor(processName: string);
    getProcessName(): string;
}
export declare class PCCProcessFrame {
    readonly groupable: PCCGroupable;
    readonly variables: Array<PCCVariableInfo>;
    readonly tempTypes: Array<CCSType>;
    private processID;
    private containerIndex;
    private usedContainers;
    private varTable;
    private protections;
    private initialVariablesVariableContainers?;
    private initialProtectionsVariableContainers?;
    private initialVariableCount;
    private isScope;
    private parentFrame?;
    /***
     * @param groupable: An object that implements the «Groupable» interface. All process frames with the same groupable object are in the same process group.
    * @param variables: An array of PCCVariableInfo objects, which represent the PseuCo variables currently covered by the process frame. It contains the incoming variables that will get passed on transition to the represented process definition.
    * @param tempTypes: An array of PCTType objects, which represents the values that will get passed as temporary values on transition to the represented process definition.
    * @param autoInit: By default true. If the created process frame will be followup process frame, the creator of the frame might want to perform its own initialization, which it can communicate through this parameter.
    ***/
    constructor(groupable: PCCGroupable, variables?: Array<PCCVariableInfo>, tempTypes?: Array<CCSType>, autoInit?: boolean);
    private relateGroupableToProcessFrame;
    private argumentsForProcessDefinition;
    emitProcessDefinition(compiler: PCCCompiler): PCCProcessDefinitionStackElement | null;
    private getProcessName;
    private getProcessArgCount;
    private didAlreadyUseContainerWithIdentifier;
    createContainer(ccsType: CCSType, desiredName?: string): PCCVariableContainer;
    getContainerForVariable(identifier: string): PCCContainer;
    assignContainerToVariable(identifier: string, container: PCCContainer): void;
    private addLocalVariable;
    /***
     * Container protection is a core feature of this compiler. To realise control flow it is sometimes necessary to transition from one CCS process definition to another (follow-up) CCS process definition. In such a case it must be known which temporary values are still needed. Such values can be `protected` and protected containers are always preserved during process definition transition.
     * Example:
     * A := a?x. B[x]
     * B[x] := b!x. 0
     * Here, x must be protected before the transition from A to B is realised. If x were not protected, the result would be erratic CCS process (because of the unbound `x` in B):
     * A := a?x. B
     * B := b!x. 0
     */
    protectContainer(container: PCCContainer): void;
    unprotectContainer(): PCCContainer;
    getProtectedContainer(): PCCContainer;
    isContainerLocalVariable(container: PCCContainer): boolean;
    createFollowupFrame(): PCCProcessFrame;
    protected createFollowupFrameAcceptingTempTypes(tempTypes: Array<CCSType>): PCCProcessFrame;
    protected getTypesForTemporaryValues(): Array<CCSType>;
    createScope(): PCCProcessFrame;
    protected configureVariablesInScope(frame: PCCProcessFrame): void;
    /***
     * Guarantees that it is allowed to transition from `frame` to `this`.
     */
    private checkTransitionConsistency;
    private variablesForTransitions;
    private argumentsToCallProcessFromFrame;
    emitCallProcessFromFrame(compiler: PCCCompiler, frame: PCCProcessFrame, appPlaceholder?: PCCApplicationPlaceholderStackElement): void;
    emitCallProcessWithArgumentContainers(compiler: PCCCompiler, containers: Array<PCCContainer>, appPlaceholder?: PCCApplicationPlaceholderStackElement): void;
    private emitScopeTransition;
    emitTransitionFromFrame(compiler: PCCCompiler, frame: PCCProcessFrame): void;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCLocalVariable | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null;
    static groupables: Map<PCCGroupable, Array<PCCProcessFrame>>;
    /***
     * The following code is used to create follow up frames from one or more process frames
     * @param frames: The array with the process frames that the new frame should be derived by.
     */
    static createFollowupFrameForFrames(frames: Array<PCCProcessFrame>): PCCProcessFrame;
    static checkTempTypesEquality(prot1: Array<PCCContainer>, prot2: Array<PCCContainer>): boolean;
    static checkFramesForConsistency(frames: Array<PCCProcessFrame>): void;
    static findClosesAncestorForFrames(frames: Array<PCCProcessFrame>): PCCProcessFrame;
}
export declare class PCCProcedureFrame extends PCCProcessFrame {
    readonly procedure: PCCProcedure;
    readonly tempTypes: Array<CCSType>;
    constructor(procedure: PCCProcedure, variables?: Array<PCCVariableInfo> | null, tempTypes?: Array<CCSType>, autoInit?: boolean);
    getProcedure(): PCCProcedure;
    private createFollowupProcedureFrameAcceptingTempTypes;
    protected createFollowupFrameAcceptingTempTypes(tempTypes: Array<CCSType>): PCCProcessFrame;
    createScope(): PCCProcedureFrame;
    emitProcessDefinition(compiler: PCCCompiler): PCCProcessDefinitionStackElement | null;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null;
}
