import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { PCCVariableInfo } from "./PCCProgramController.js";
export class PCCProgramAnalysisVisitor extends PC.AbstractVisitor {
    programController;
    constructor(programController) {
        super();
        this.programController = programController;
    }
    visitProgram(node, context) {
        node.declarations.forEach((decl) => decl.accept(this));
    }
    visitMainAgent(node, context) {
        this.programController.beginMainAgent(node);
        node.body.accept(this);
        this.programController.endMainAgent();
    }
    visitProcedureDeclaration(node, context) {
        const args = node.parameters.map((param) => {
            const declType = this.programController.staticSemantics.typeForNode(param);
            return new PCCVariableInfo(param, param.name, declType.declaredType);
        });
        const procType = this.programController.staticSemantics.typeForNode(node).declaredType;
        const returnType = procType.returnType;
        this.programController.beginNewProcedure(node, node.name, returnType, args);
        node.body.accept(this);
        this.programController.endProcedure();
    }
    visitMonitor(node, context) {
        this.programController.beginClass(node.name);
        node.declarations.forEach((decl) => decl.accept(this));
        this.programController.endClass();
    }
    visitStruct(node, context) {
        this.programController.beginClass(node.name);
        node.declarations.forEach((decl) => decl.accept(this));
        this.programController.endClass();
    }
    visitConditionDeclaration(node, context) {
        const v = new PCCVariableInfo(node, node.name, PCT.SystemTypes.condition);
        this.programController.processNewVariable(v);
    }
    visitVariableDeclarations(node, context) {
        node.declarators.forEach((decl) => decl.accept(this));
    }
    visitVariableDeclarator(node, context) {
        const type = this.programController.staticSemantics.typeForNode(node);
        const v = new PCCVariableInfo(node, node.name, type.declaredType);
        this.programController.processNewVariable(v);
    }
    visitStatementWrapper(node, context) {
        node.statement?.accept(this);
    }
    visitBreakStatement(node, context) {
        // nothing to do
    }
    visitContinueStatement(node, context) {
        // nothing to do
    }
    visitStatementBlock(node, context) {
        this.programController.openEnvironment(node);
        node.blockStatements.forEach((stmt) => stmt.accept(this));
        this.programController.closeEnvironment();
    }
    visitStatementExpression(node, context) {
        // nothing to do
    }
    visitSelectStatement(node, context) {
        this.programController.openEnvironment(node);
        node.cases.forEach((cse) => cse.accept(this));
        this.programController.closeEnvironment();
    }
    visitCase(node, context) {
        node.statement.accept(this);
    }
    visitIfStatement(node, context) {
        this.programController.openEnvironment(node);
        node.consequence.accept(this);
        node.alternative?.accept(this);
        this.programController.closeEnvironment();
    }
    visitWhileStatement(node, context) {
        this.programController.openEnvironment(node);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }
    visitDoStatement(node, context) {
        this.programController.openEnvironment(node);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }
    visitForStatement(node, context) {
        this.programController.openEnvironment(node);
        node.loopInit?.accept(this);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }
    visitForInit(node, context) {
        node.variableInitialisations.forEach((init) => init.accept(this));
    }
    visitReturnStatement(node, context) {
        // nothing to do
    }
    visitPrimitiveStatement(node, context) {
        // nothing to do
    }
    visitPrintStatement(node, context) {
        // nothing to do
    }
    visitConditionStatement(node, context) {
        // nothing to do
    }
    // The remaining visit methods will not be called
    visitProcedureParameter(node, context) {
        throw new Error("Method not implemented.");
    }
    visitVariableInitializer(node, context) {
        throw new Error("Method not implemented.");
    }
    visitArrayTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitSimpleTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitChannelTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitClassTypeNode(node, context) {
        throw new Error("Method not implemented.");
    }
    visitStartExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAssignExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAssignDestination(node, context) {
        throw new Error("Method not implemented.");
    }
    visitSendExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitConditionalExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitOrExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAndExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitEqualityExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitRelationalExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitAdditiveExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitMultiplicativeExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitUnaryExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitPostfixExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitReceiveExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitProcedureCall(node, context) {
        throw new Error("Method not implemented.");
    }
    visitClassCall(node, context) {
        throw new Error("Method not implemented.");
    }
    visitArrayExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitLiteralExpression(node, context) {
        throw new Error("Method not implemented.");
    }
    visitIdentifierExpression(node, context) {
        throw new Error("Method not implemented.");
    }
}
//# sourceMappingURL=PCCProgramAnalysisVisitor.js.map