import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { PCCContainer } from "./PCCContainer.js";
import { PCCCompiler } from "./PCCCompiler.js";
import { PCCGroupable } from "./PCCProcessFrame.js";
export declare class PCCEnvironmentNode {
    node: PC.Node;
    label: string;
    protected _parent?: PCCEnvironmentNode;
    get parent(): PCCEnvironmentNode;
    private children;
    protected _variables: Map<string, PCCVariable>;
    get variables(): ReadonlyMap<string, PCCVariable>;
    protected _procedures: Map<string, PCCProcedure>;
    get procedures(): ReadonlyMap<string, PCCProcedure>;
    protected _blocks: Array<PCCEnvironmentNode>;
    get blocks(): ReadonlyArray<PCCEnvironmentNode>;
    __id?: number;
    constructor(node: PC.Node, label: string);
    addChild(child: PCCEnvironmentNode | PCCVariable): void;
    getVariableWithName(name: string): PCCVariable;
    getProcedureWithName(name: string): PCCProcedure;
    constructVariableObject(variableNode: PC.Node | null, variableName: string, variableType: PCT.Type): PCCVariable;
    getBlockWithId(id: number): PCCEnvironmentNode;
    getComposedLabel(): string;
    getAllClasses(): Array<PCCClass>;
    isDescendantOfProcedure(): boolean;
}
export declare class PCCGlobal extends PCCEnvironmentNode {
    constructor(program: PC.Program);
    constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCGlobalVariable;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): PCCVariable;
}
export declare class PCCClass extends PCCEnvironmentNode {
    type: PCT.ClassType;
    protected usedClassTypes: Array<PCT.ClassType>;
    constructor(node: PC.Node, type: PCT.ClassType);
    addChild(child: PCCEnvironmentNode | PCCVariable): void;
    get fields(): ReadonlyMap<string, PCCField>;
    getName(): string;
    isMonitor(): boolean;
    addUseOfClassType(type: PCT.ClassType): void;
    getUsedClassTypes(): ReadonlyArray<PCT.Type>;
    hasEnvironment(): boolean;
    getAllConditions(): Array<PCCVariable>;
    constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCField;
    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable;
    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure;
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void;
    emitEnvironment(compiler: PCCCompiler): void;
    emitConstructor(compiler: PCCCompiler): void;
    getProcessName(): string;
    getEnvProcessName(): string;
}
export declare class PCCProcedure extends PCCEnvironmentNode implements PCCGroupable {
    readonly returnType: PCT.Type;
    readonly procedureArguments: Array<PCCVariableInfo>;
    constructor(node: PC.Node, name: string, returnType: PCT.Type, procedureArguments: Array<PCCVariableInfo>);
    getName(): string;
    isDescendantOfProcedure(): boolean;
    isClassProcedure(): boolean;
    isMonitorProcedure(): boolean;
    getProcessName(): string;
    getAgentStarterChannel(): string;
    getAgentProcessName(): string;
    constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCLocalVariable;
    getAllArgumentContainers(compiler: PCCCompiler, explicitArgumentContainers: Array<PCCContainer>, instanceContainer: PCCContainer | null): Array<PCCContainer>;
    getImplicitAndExplicitArgumentCount(): number;
    emitAgentConstructor(compiler: PCCCompiler, addJoiner?: boolean): void;
    emitExit(compiler: PCCCompiler): void;
}
export declare class PCCVariableInfo {
    readonly node: PC.Node | null;
    readonly label: string;
    readonly type: PCT.Type;
    readonly isInternal: boolean;
    constructor(node: PC.Node | null, label: string, type: PCT.Type, isInternal?: boolean);
    /**
     * Get the name of the variable
     * @returns The name of the variable
     */
    getName(): string;
    /**
     * Variables can be internal or regular. The identifier provides a unique name to distinguish internal and regular variables in the same name space.
     * @returns Returns an identifier for the variable
     */
    getIdentifier(): string;
    /**
     *
     * @deprecated
     */
    getComposedLabel(): string;
    /**
     * Provides a common way to transfer pseuCo variable identifiers into CCS variable identifiers.
     * @returns An identifier for a CCS variable
     */
    getSuggestedContainerName(): string;
    /**
     * Translates pseuCo value types into CCS value types.
     * @returns The type of CCS values encoding the pseuCo values that this variable represents.
     */
    getCCSType(): CCSType;
    /**
     * Compiles a variable's default value.
     * @param compiler The shared compiler object
     * @returns A container that holds the variable's default value
     */
    compileDefaultValue(compiler: PCCCompiler): PCCContainer;
    private static suggestCCSVariableNameForInternalVariable;
    private static suggestCCSVariableNameForPseucoVariable;
}
export declare abstract class PCCVariable extends PCCVariableInfo {
    /**
     * Retrieves the current value of this variable and returns it as Container object.
     * @param compiler The shared compiler object
     * @returns A container that holds the variable's current value.
     */
    abstract getContainer(compiler: PCCCompiler): PCCContainer;
    /**
     * Stores the value of `container` as the variable's current value.
     * @param compiler The shared compiler object
     * @param container The new value for this variable
     */
    abstract setContainer(compiler: PCCCompiler, container: PCCContainer): void;
}
export declare class PCCGlobalVariable extends PCCVariable {
    accessorChannel(set: boolean): string;
    getContainer(compiler: PCCCompiler): PCCContainer;
    setContainer(compiler: PCCCompiler, container: PCCContainer): void;
    /**
     * Emits actions to the environment process of a global variable, monitor variable, or struct variable to get or set the value of the variable.
     * @param compiler The shared compiler coordination object
     * @param variables A list of variables that is stored in the CCS environment process to which the accessors are added; always contains at least `this`
     * @param instance For accessors of monitor (or struct) variables, this parameter is a container that contains a reference to the monitor (struct) object.
     */
    emitAccessors(compiler: PCCCompiler, variables: Array<PCCVariable>, instance?: PCCContainer | null): void;
    emitConstructor(compiler: PCCCompiler): void;
    getProcessName(): string;
    getEnvProcessName(): string;
}
export declare class PCCField extends PCCGlobalVariable {
    readonly cls: PCCClass;
    constructor(node: PC.Node | null, label: string, type: PCT.Type, cls: PCCClass, isInternal?: boolean);
    accessorChannel(set: boolean): string;
    getContainer(compiler: PCCCompiler): PCCContainer;
    setContainer(compiler: PCCCompiler, container: PCCContainer): void;
    getEnvProcessName(): string;
}
export declare class PCCInternalReadOnlyField extends PCCField {
    constructor(node: PC.Node | null, label: string, type: PCT.Type, cls: PCCClass);
    emitAccessors(compiler: PCCCompiler, variables: Array<PCCVariable>, instance: PCCContainer): void;
    getContainer(compiler: PCCCompiler): PCCContainer;
    setContainer(compiler: PCCCompiler, container: PCCContainer): void;
}
export declare class PCCCondition extends PCCField {
    readonly expressionNode: PC.Node;
    constructor(node: PC.Node | null, label: string, cls: PCCClass, expressionNode: PC.Node);
}
export declare class PCCLocalVariable extends PCCVariable {
    getContainer(compiler: PCCCompiler): PCCContainer;
    setContainer(compiler: PCCCompiler, container: PCCContainer): void;
}
export declare class PCCProgramController {
    readonly program: PC.Program;
    readonly staticSemantics: PCT.StaticSemantics;
    root: PCCGlobal;
    classes: Map<string, PCCClass>;
    agents: Map<string, PCCProcedure>;
    protected _envStack: PCCEnvironmentNode;
    protected blockCounter: number;
    pseucoToEnvironmentNode: Map<PC.Node, PCCEnvironmentNode>;
    constructor(program: PC.Program, staticSemantics: PCT.StaticSemantics);
    analyseProgram(): void;
    static PCTEnvironmentControllerInternalMainAgentName: string;
    getGlobal(): PCCGlobal;
    getStackState(): PCCEnvironmentNode;
    private restoreStackState;
    openEnvironment(node: PC.Node): PCCEnvironmentNode;
    reopenEnvironment(node: PC.Node): PCCEnvironmentNode;
    closeEnvironment(): void;
    getClassWithName(name: string): PCCClass;
    getAllClasses(): PCCClass[];
    getVariableWithName(name: string): PCCVariable;
    getProcedureWithName(name: string): PCCProcedure;
    getMainAgent(): PCCProcedure;
    processNewClass(node: PC.Monitor | PC.Struct, classType: PCT.ClassType): void;
    _processNewClass(node: PCCClass): void;
    beginClass(className: string): PCCClass;
    endClass(): void;
    beginNewProcedure(node: PC.ProcedureDeclaration | PC.MainAgent, procedureName: string, returnType: PCT.Type, args: PCCVariableInfo[]): void;
    protected _beginNewProcedure(node: PCCProcedure): PCCProcedure;
    beginProcedure(procedureName: string): PCCProcedure;
    endProcedure(): void;
    beginMainAgent(node: PC.MainAgent): void | PCCProcedure;
    endMainAgent(): void;
    processNewVariable(variable: PCCVariableInfo): void;
    _processNewVariable(variable: PCCVariable): void;
    processProcedureAsAgent(procedure: PCCProcedure): void;
    getAgents(): Array<PCCProcedure>;
    getUsedTypes(): {
        arrays: Set<number>;
        channels: Set<number>;
    };
    private collectTypes;
}
export declare enum CCSTypeDescriptor {
    VOID = 0,
    BOOL = 1,
    INT = 2,
    STRING = 3
}
export declare class CCSType {
    readonly descriptor: CCSTypeDescriptor;
    constructor(descriptor: CCSTypeDescriptor);
    isVoid(): boolean;
    isInt(): boolean;
    isString(): boolean;
    isBool(): boolean;
    isArrayLike(): boolean;
    isClass(): boolean;
    isEqual(other: CCSType): boolean;
    getDefaultContainer(): PCCContainer;
    static readonly VOID: CCSType;
    static readonly BOOL: CCSType;
    static readonly INT: CCSType;
    static readonly STRING: CCSType;
    static fromPseucoType(pType: PCT.Type): CCSType;
}
export declare class CCSArrayLikeType extends CCSType {
    readonly elementsType: CCSType;
    constructor(elementsType: CCSType);
    isArrayLike(): boolean;
    isEqual(other: CCSType): boolean;
}
export declare class CCSClassType extends CCSType {
    readonly className: string;
    constructor(className: string);
    isClass(): boolean;
    isEqual(other: CCSType): boolean;
}
