import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { PCCCompiler } from "./PCCCompiler.js";
import { utils } from "./utils.js";
import { PCCBinaryContainer, PCCBinaryContainerOperator, PCCConstantContainer, PCCContainer, PCCUnaryContainer } from "./PCCContainer.js";
import { CCSArrayLikeType, CCSClassType, CCSType } from "./PCCProgramController.js";
import { CasePreparationData } from "./PCCCompileStatementsVisitor.js";


export class PCCCompileExpressionsVisitor extends PC.AbstractVisitor<PCCContainer | null> {
    constructor(readonly compiler: PCCCompiler, readonly compilingNodes: Array<PC.Node>) {
        super();
    }

    compileNode(node: PC.Expression | PC.AssignDestination, context?: unknown): PCCContainer | null {
        this.compilingNodes.push(node);
        const res = node.accept(this, context);
        this.compilingNodes.pop();
        return res;
    }

    
    visitStartExpression(node: PC.StartExpression): PCCContainer | null {
        this.compiler.setNeedsAgentManager();
        const call = node.callExpression;
        if (call instanceof PC.ProcedureCall) {
            return this.compileStart(call, null, null);
        } else {
            const classType = this.compiler.staticSemantics.typeForNode(call.instanceExpression) as PCT.ClassType;
            const instanceContainer = this.compileNode(call.instanceExpression);
            return this.compileStart(call.callExpression, instanceContainer, classType.className);
        }
    }

    private compileStart(callNode: PC.ProcedureCall, instanceContainer: PCCContainer | null, className: string | null): PCCContainer {
        // Get the procedure
        const proc = this.compiler.getProcedureWithNameOfClass(callNode.procedureName, className);
        // Tell the compiler that this procedure is used as an agent
        this.compiler.processProcedureAsAgent(proc);
        if (instanceContainer) {
            this.compiler.protectContainer(instanceContainer);
        }
        callNode.argumentExpressions.forEach((e) => {
            const c = utils.assertNonNull(this.compileNode(e));
            this.compiler.protectContainer(c);
        });
        const args: Array<PCCContainer> = [];
        // Recover the arguments (note that the recovering happens in reversed argument order, i.e., the instanceContainer is recovered last)
        callNode.argumentExpressions.forEach((e) => args.unshift(this.compiler.unprotectContainer()));
        const i = (instanceContainer) ? this.compiler.unprotectContainer() : null;
        const implicitAndExplicitArgs = proc.getAllArgumentContainers(this.compiler, args, i);
        const result = this.compiler.getFreshContainer(CCSType.INT);
        this.compiler.emitInput(proc.getAgentStarterChannel(), null, result);
        implicitAndExplicitArgs.forEach((c) => this.compiler.emitOutput("start_set_arg", result, c));
        return result;
    }

    visitAssignExpression(node: PC.AssignExpression, context: CasePreparationData): PCCContainer | null {
        // If context is not undefined, then the assignment is used in a case expression. All we have to do is to forward the preparation data to the receive expression.

        let c = utils.assertNonNull(this.compileNode(node.valueExpression, context));
        if (node.operator == "=") {
            this.assignContainerToAssignDestination(node.assignDestination, c);
            return c;
        }

        const currentDestinationValue = utils.assertNonNull(this.compileNode(node.assignDestination));
        if (node.operator == "+=") {
            c = new PCCBinaryContainer(currentDestinationValue, c, "+");
        } else if (node.operator == "-=") {
            c = new PCCBinaryContainer(currentDestinationValue, c, "-");
        } else if (node.operator == "*=") {
            c = new PCCBinaryContainer(currentDestinationValue, c, "*");
        } else if (node.operator == "/=") {
            c = new PCCBinaryContainer(currentDestinationValue, c, "/");
        } else {
            throw new Error("Unknown assign operator");
        }
        this.assignContainerToAssignDestination(node.assignDestination, c);
        return c;
    }

    private assignContainerToAssignDestination(node: PC.AssignDestination, value: PCCContainer) {
        const v = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
        if (node.arrayIndexExpressions.length == 0) {
            v.setContainer(this.compiler, value);
        } else {
            let ai = v.getContainer(this.compiler);
            for (let i = 0; i < node.arrayIndexExpressions.length-1; i++) {
                const arrayIndex = utils.assertNonNull(this.compileNode(node.arrayIndexExpressions[i]));
                ai = this.getValueOfArrayAtIndexOfAssignDestination(ai, arrayIndex);
                
            }
            const arrayIndex = utils.assertNonNull(this.compileNode(node.arrayIndexExpressions[node.arrayIndexExpressions.length - 1]));
            this.setValueOfArrayAtIndexOfAssignDestination(ai, arrayIndex, value);
        }
    }

    private getValueOfArrayAtIndexOfAssignDestination(arrayInstance: PCCContainer, arrayIndex: PCCContainer): PCCContainer {
        this.compiler.emitOutput("array_access", arrayInstance, arrayIndex);
        const arrayType = arrayInstance.ccsType as CCSArrayLikeType;
        const result = this.compiler.getFreshContainer(arrayType.elementsType);
        this.compiler.emitInput("array_get", arrayInstance, result);
        return result
    }

    private setValueOfArrayAtIndexOfAssignDestination(arrayInstance: PCCContainer, arrayIndex: PCCContainer, value: PCCContainer) {
        this.compiler.emitOutput("array_access", arrayInstance, arrayIndex);
	    this.compiler.emitOutput("array_set", arrayInstance, value);
    }

    visitAssignDestination(node: PC.AssignDestination): PCCContainer | null {
        // For identifiers, it does the same as visitIdentifierExpression, for arrays the same as visitArrayExpression. I.e., it handles the the assign destination as right-hand-side expression.
        const v = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
        let res = v.getContainer(this.compiler);
        for (let i = 0; i < node.arrayIndexExpressions.length; i++) {
            const arrayIndex = utils.assertNonNull(this.compileNode(node.arrayIndexExpressions[i]));
            res = this.getValueOfArrayAtIndexOfAssignDestination(res, arrayIndex);
        }
        return res;
    }


    visitSendExpression(node: PC.SendExpression, context?: CasePreparationData): PCCContainer | null {
        let chanContainer = context?.channel;
        let valContainer = context?.valueToSend;
        // If do not have prepared data (because the expression is not used in a case condition), we must compile it now.
        if (!chanContainer) {
            chanContainer = utils.assertNonNull(this.compileNode(node.channelExpression));
        }
        if (!valContainer) {
            // Protect the channel container
            this.compiler.protectContainer(chanContainer);
            // Compile the value
            valContainer = utils.assertNonNull(this.compileNode(node.valueExpression));
            // Restore the channel container
            chanContainer = this.compiler.unprotectContainer();
        }

        // Protect value, because it is the result of this expression
        this.compiler.protectContainer(valContainer);
        // We must determine at runtime if the channel is buffered or unbuffered -> we need the choice operator
        const control = this.compiler.emitChoice();
        // Case 1: Channel is buffered; in this case the channel reference is non-negative...
        this.compiler.emitCondition(new PCCBinaryContainer(chanContainer, new PCCConstantContainer(0), ">="));
        // ... and we use the CCS channel `put` to send the value
        this.compiler.emitOutput("put", chanContainer, valContainer);
        // We merge with the other branch in a yet unknown new process definition; we place an application placeholder here
        const p1 = this.compiler.emitProcessApplicationPlaceholder();
        control.setBranchFinished();
        // Case 2: Channel is unbuffered; in this case the channel reference is negative...
        this.compiler.emitCondition(new PCCBinaryContainer(chanContainer, new PCCConstantContainer(0), "<"));
        // ... and we use the CCS channel `receive` to send the value
        this.compiler.emitOutput("receive", chanContainer, valContainer);
        // Again we place an application placeholder here for the merge
        const p2 = this.compiler.emitProcessApplicationPlaceholder();
        control.setBranchFinished();

        // We are ready to merge
        this.compiler.emitMergeOfProcessFramesOfPlaceholders([p1, p2]);
        // Restore the sent value
        valContainer = this.compiler.unprotectContainer();

        // We return the sent value as the result of this expression
        return valContainer;
    }

    visitReceiveExpression(node: PC.ReceiveExpression, context?: CasePreparationData): PCCContainer | null {
        let chanContainer = context?.channel;
        // If do not have prepared data (because the expression is not used in a case condition), we must compile it now.
        if (!chanContainer) {
            chanContainer = utils.assertNonNull(this.compileNode(node.channelExpression));
        }
        // Create a fresh container that will hold the received value
        const res = this.compiler.getFreshContainer((chanContainer.ccsType as CCSArrayLikeType).elementsType);
        // Receive the value over CCS channel `receive` (note that for receiving we don't have to distinguish between buffered and unbuffered)
        this.compiler.emitInput("receive", chanContainer, res);
        // Return the received value as the result of this expression
        return res;
    }


    visitConditionalExpression(node: PC.ConditionalExpression, context?: unknown): PCCContainer | null {
        // First evaluate the predicate
        const b = utils.assertNonNull(this.compileNode(node.predicate));
        // To allow for both alternatives, we use the choice operator
        const control = this.compiler.emitChoice();
        // Case 1: the predicate holds
        this.compiler.emitCondition(b);
        // Compile the consequence
        const c1 = utils.assertNonNull(this.compileNode(node.consequence));
        // Make sure it will be passed to the merging process definition
        this.compiler.protectContainer(c1);
        // Remember this position for moving to the (yet unknown) merging process definition
        const lp = this.compiler.emitProcessApplicationPlaceholder();

        // Alternative
        control.setBranchFinished();
        // Negate the predicate value
        this.compiler.emitCondition(new PCCUnaryContainer("!", b));
        // Compile the alternative
        const c2 = utils.assertNonNull(this.compileNode(node.alternative));
        // Make sure it will be passed to the merging process definition
        this.compiler.protectContainer(c2);
        // Remember this position for moving to the (yet unknown) merging process definition
        const rp = this.compiler.emitProcessApplicationPlaceholder();
        
        // Merge the two branches
        this.compiler.emitMergeOfProcessFramesOfPlaceholders([lp, rp]);
        // Every branch from which we merged has one protected container. After merging, we have exactly one protected container (which is either the left or the right branch value)
        return this.compiler.unprotectContainer();
    }


    private visitSimpleExpression(leftNode: PC.Expression, rightNode: PC.Expression, operator: PCCBinaryContainerOperator) {
        // Compile the left operand
        let left = utils.assertNonNull(this.compileNode(leftNode));
        // Make sure the value still exists after the right operand is compiled
        this.compiler.protectContainer(left);
        // COmpile the right operand
        const right = utils.assertNonNull(this.compileNode(rightNode));
        // Restore the left operand value
        left = this.compiler.unprotectContainer();
        // The result is a CCS VP expression
        return new PCCBinaryContainer(left, right, operator);
    }

    visitOrExpression(node: PC.OrExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, "||");
    }

    visitAndExpression(node: PC.AndExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, "&&");
    }

    visitEqualityExpression(node: PC.EqualityExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, node.operator);
    }

    visitRelationalExpression(node: PC.RelationalExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, node.operator);
    }

    visitAdditiveExpression(node: PC.AdditiveExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, node.operator);
    }

    visitMultiplicativeExpression(node: PC.MultiplicativeExpression, context?: unknown): PCCContainer | null {
        return this.visitSimpleExpression(node.left, node.right, node.operator);
    }

    
    visitUnaryExpression(node: PC.UnaryExpression, context?: unknown): PCCContainer | null {
        const v = utils.assertNonNull(this.compileNode(node.valueExpression));
        return new PCCUnaryContainer(node.operator, v);
    }

    visitPostfixExpression(node: PC.PostfixExpression, context?: unknown): PCCContainer | null {
        // First, compute the new value
        const op = (node.operator == "++") ? "+" : "-";
        const v = utils.assertNonNull(this.compileNode(node.assignDestination));
        const newValue = new PCCBinaryContainer(v, new PCCConstantContainer(1), op);
        // Then, store the value; we already have method is is able to compile assignments to assign destinations
        this.assignContainerToAssignDestination(node.assignDestination, newValue);
        // Return the new value (however, pseuCo currently does not allow to use the new value in well-typed programs.)
        return newValue;
    }


    private compileProcedureCall(node: PC.ProcedureCall, instanceContainer?: PCCContainer, className?: string): PCCContainer | null {
        // Get the procedure object
        const proc = this.compiler.getProcedureWithNameOfClass(node.procedureName, className ?? null);
        // Prepare the arguments with which the procuedure shall be called
        // For class calls we have an instanceContainer that must be protected
        if (instanceContainer) {
            this.compiler.protectContainer(instanceContainer);
        }
        // Compile and protect the remaining arguments
        node.argumentExpressions.forEach((arg) => {
            const v = utils.assertNonNull(this.compileNode(arg));
            this.compiler.protectContainer(v);
        });
        // The procedure call is realised using the CCS sequence operator
        const control = this.compiler.emitSequence();
        // Recover arguments (recovering happens in REVERSE order; don't get confused by the usage of `map` here)
        const args = node.argumentExpressions.map((_) => this.compiler.unprotectContainer());
        // Reverse the order of args to get the original argument order
        args.reverse();
        // Recover the instanceContainer if it was provided initially
        if (instanceContainer) {
            instanceContainer = this.compiler.unprotectContainer();
        }
        // Give the procedure object the chance to add more internal arguments
        const allArgs = proc.getAllArgumentContainers(this.compiler, args, instanceContainer ?? null);
        // Perform the procedure call by applying the corresponding CCS process definition
        this.compiler.emitProcessApplication(proc.getProcessName(), allArgs);
        // The left operand of the sequence operator is finished
        control.setBranchFinished();

        // Next is about retrieving the return value of non-void procedures
        if (proc.returnType.isVoidType) {
            // Nothing to return
            return null;
        } else {
            // Create a fresh container to hold the return value
            const res = this.compiler.getFreshContainer(CCSType.fromPseucoType(proc.returnType));
            // Return values are transferred using an auxiliary process that provides us the value via CCS channel `rreturn`
            this.compiler.emitInput("rreturn", null, res);
            return res;
        }
        
        
        

    }

    visitProcedureCall(node: PC.ProcedureCall, context?: unknown): PCCContainer | null {
        return this.compileProcedureCall(node);
    }

    visitClassCall(node: PC.ClassCall, context?: unknown): PCCContainer | null {
        // Compile the instance of the class
        const instance = utils.assertNonNull(this.compileNode(node.instanceExpression));
        // Retrieve the class name from the CCSType of the instance container
        const ccsType = instance.ccsType as CCSClassType;
        const className = ccsType.className;
        // Use the existing method to compile procedure calls
        return this.compileProcedureCall(node.callExpression, instance, className);
    }


    visitArrayExpression(node: PC.ArrayExpression, context?: unknown): PCCContainer | null {
        // Get array instance value and protect it from being forgotten
        let arrayInstance = utils.assertNonNull(this.compileNode(node.instanceExpression));
        this.compiler.protectContainer(arrayInstance);
        // Get the index which we want to access
        const indexContainer = utils.assertNonNull(this.compileNode(node.indexExpression));
        // Recover the array index
        arrayInstance = this.compiler.unprotectContainer();
        // Notify the array that we want to access the given index
        this.compiler.emitOutput("array_access", arrayInstance, indexContainer);
        // Create a fresh container that will hold the accessed value
        const arrayType = arrayInstance.ccsType as CCSArrayLikeType;
        const res = this.compiler.getFreshContainer(arrayType.elementsType);
        // Retrieve the actual value and return it
        this.compiler.emitInput("array_get", arrayInstance, res);
        return res;
    }

    visitLiteralExpression(node: PC.LiteralExpression, context?: unknown): PCCContainer | null {
        return new PCCConstantContainer(node.value);
    }

    visitIdentifierExpression(node: PC.IdentifierExpression, context?: unknown): PCCContainer | null {
        const v = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
        return v.getContainer(this.compiler);
    }




    // The remaining nodes are not expressions
    visitProgram(node: PC.Program, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitMainAgent(node: PC.MainAgent, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitProcedureDeclaration(node: PC.ProcedureDeclaration, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitProcedureParameter(node: PC.ProcedureParameter, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitMonitor(node: PC.Monitor, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitStruct(node: PC.Struct, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitConditionDeclaration(node: PC.ConditionDeclaration, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitVariableDeclarations(node: PC.VariableDeclarations, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitVariableDeclarator(node: PC.VariableDeclarator, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitVariableInitializer(node: PC.VariableInitializer, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitArrayTypeNode(node: PC.ArrayTypeNode, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitSimpleTypeNode(node: PC.SimpleTypeNode, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitChannelTypeNode(node: PC.ChannelTypeNode, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitClassTypeNode(node: PC.ClassTypeNode, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitStatementWrapper(node: PC.StatementWrapper, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitBreakStatement(node: PC.BreakStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitContinueStatement(node: PC.ContinueStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitStatementBlock(node: PC.StatementBlock, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitStatementExpression(node: PC.StatementExpression, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitSelectStatement(node: PC.SelectStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitCase(node: PC.Case, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitConditionStatement(node: PC.ConditionStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitIfStatement(node: PC.IfStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitWhileStatement(node: PC.WhileStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitDoStatement(node: PC.DoStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitForStatement(node: PC.ForStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitForInit(node: PC.ForInit, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitReturnStatement(node: PC.ReturnStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitPrimitiveStatement(node: PC.PrimitiveStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }
    visitPrintStatement(node: PC.PrintStatement, context?: unknown): PCCContainer | null {
        throw new Error("Method not implemented.");
    }

    
    

}