import * as PC from "@pseuco/lang";
import * as CCS from "@pseuco/ccs-interpreter";
import { PCCCompiler } from "./PCCCompiler.js";
import { utils } from "./utils.js";
import { PCCApplicationPlaceholderStackElement, PCCChoiceStackElement } from "./PCCCompilerStack.js";
import { PCCBinaryContainer, PCCConstantContainer, PCCContainer, PCCUnaryContainer } from "./PCCContainer.js";
import { PCCProcessFrame } from "./PCCProcessFrame.js";
import { CCSType, PCCVariable } from "./PCCProgramController.js";


export class SelectPreparationData {
    private casePrepData?: Map<PC.Case, CasePreparationData>;
    get casePreparationData(): ReadonlyMap<PC.Case, CasePreparationData> {
        if (!this.casePrepData) {
            throw new Error("Select statement not yet prepared.");
        }
        return this.casePrepData;
    }
    get isPrepared(): boolean {
        return this.casePrepData != undefined;
    }
    // private protections?: Array<CasePreparationData>;

    constructor(readonly compiler: PCCCompiler, readonly selectStatement: PC.SelectStatement) { }

    prepare() {
        if (this.isPrepared) {
            return;
        }
        // Initialise the map from case nodes to case preparation data
        this.casePrepData = new Map();
        // In the process of preparing, we will have to protect the values we already prepared. We use `protections` to remember which and in which order the prep data was protected
        const protections = new Array<CasePreparationData>();
        // Iterate over all cases
        this.selectStatement.cases.forEach((caseNode) => {
            // The default case does not have a condition
            if (caseNode.condition) {
                // Create the CasePreparationData object; the constructor performs the necessary compilation
                const prepData = new CasePreparationData(this.compiler, caseNode);
                // Remember the connection between case node and the prep data
                this.casePrepData!.set(caseNode, prepData);
                // Protect the prepared values
                prepData.protectContainers();
                protections!.push(prepData);
            } // else: default case; nothing to be prepared
        });
        // Unprotect all prepared values (unprotecting must happen in reverse order)
        protections.reverse();
        protections.forEach((prepData) => prepData.unprotectContainers());
    }


}
export class CasePreparationData {
    channel: PCCContainer;
    valueToSend?: PCCContainer;

    constructor(readonly compiler: PCCCompiler, readonly caseNode: PC.Case) {
        // Compiles the channel and value expression to actual values
        // 1. Find the top-level send or receive expression (this is easy, because of a restricted syntax for case conditions)
        let chanExp: PC.SendExpression | PC.ReceiveExpression;
        if (caseNode.condition instanceof PC.SendExpression || caseNode.condition instanceof PC.ReceiveExpression) {
            chanExp = caseNode.condition;
        } else if (caseNode.condition instanceof PC.AssignExpression) {
            if (!(caseNode.condition.valueExpression instanceof PC.ReceiveExpression)) {
                throw new Error("Unexpected AST");
            }
            chanExp = caseNode.condition.valueExpression;
        } else {
            if (caseNode.condition != null) {
                throw new Error("Unexpected AST");
            } else {
                throw new Error("Cannot create CasePreparationData for default case");
            }
        }
        
        // Compile the channel expression
        this.channel = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(chanExp.channelExpression));
        // Protect it in case we must compile a value expression
        this.compiler.protectContainer(this.channel);
        
        if (chanExp instanceof PC.SendExpression) {
            // For send expressions, we also need the value to send
            this.valueToSend = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(chanExp.valueExpression));
        }
        // Unprotect the channel expression
        this.channel = this.compiler.unprotectContainer();
    }

    protectContainers() {
        this.compiler.protectContainer(this.channel);
        if (this.valueToSend) {
            this.compiler.protectContainer(this.valueToSend);
        }
    }

    unprotectContainers() {
        if (this.valueToSend) {
            this.valueToSend = this.compiler.unprotectContainer();
        }
        this.channel = this.compiler.unprotectContainer();
    }
}


enum LoopControlFlowEventType {
    CONTINUE, 
    BREAK
};

class LoopControlFlowEvent {
    constructor(readonly type: LoopControlFlowEventType, readonly appPlaceholder: PCCApplicationPlaceholderStackElement) { }

    get isBreakEvent(): boolean {
        return this.type == LoopControlFlowEventType.BREAK;
    }

    get isContinueEvent(): boolean {
        return this.type == LoopControlFlowEventType.CONTINUE;
    }

    static newBreakEvent(appPlaceHolder: PCCApplicationPlaceholderStackElement) {
        return new LoopControlFlowEvent(LoopControlFlowEventType.BREAK, appPlaceHolder);
    }

    static newContinueEvent(appPlaceHolder: PCCApplicationPlaceholderStackElement) {
        return new LoopControlFlowEvent(LoopControlFlowEventType.CONTINUE, appPlaceHolder);
    }
}


export class PCCCompileStatementsVisitor extends PC.AbstractVisitor<Array<LoopControlFlowEvent>> {
    // private loopEntry: PCCProcessFrame | null = null;

    constructor(readonly compiler: PCCCompiler, readonly compilingNodes: Array<PC.Node>) {
        super();
    }

    compileNode(node: PC.Statement | PC.StatementWrapper | PC.Case | PC.ForInit, context?: unknown): Array<LoopControlFlowEvent> {
        this.compilingNodes.push(node);
        const res = node.accept(this, context);
        this.compilingNodes.pop();
        return res;
    }

    

    visitSelectStatement(node: PC.SelectStatement): LoopControlFlowEvent[] {
        if (node.cases.length == 0) {
            // Nothing to do; no breaks to report about
            return [];
        }
        this.compiler.reopenEnvironment(node);
        
        // prepare the send and receive expressions, such that in CCS we can have a choice operator where the top level actions are the send or receive actions
        const prepData = new SelectPreparationData(this.compiler, node);
        prepData.prepare();
                
        const placeholders: Array<PCCApplicationPlaceholderStackElement> = [];
        let loopControls: Array<LoopControlFlowEvent> = [];

        // iterate over all except the last case to construct a proper choice term
        for (let i = 0; i < node.cases.length - 1; i++) {
            // Emit the choice operator
            const control = this.compiler.emitChoice();
            // this.compiler.emitNewScope();
            // Get the prepared channel and value containers and pass it into the case's compile method
            const casePreparationData = prepData.casePreparationData.get(node.cases[i]);
            // (Remember potential break instructions and forward them at the end; the select statement cannot be left using break)
            loopControls = loopControls.concat(this.compileNode(node.cases[i], casePreparationData));
            // Place an app placeholder to allow merging all cases later
            placeholders.push(this.compiler.emitProcessApplicationPlaceholder());
            // Switch to the right branch of the choice
            control.setBranchFinished();
        }
        // Do the last case as well
        // this.compiler.emitNewScope();
        const caseNode = node.cases[node.cases.length - 1];
        const casePreparationData = prepData.casePreparationData.get(caseNode);
        loopControls = loopControls.concat(this.compileNode(caseNode, casePreparationData));
        placeholders.push(this.compiler.emitProcessApplicationPlaceholder());
        // Merge the branches of all cases to a unified process
        this.compiler.emitMergeOfProcessFramesOfPlaceholders(placeholders);
        this.compiler.closeEnvironment();
        return loopControls;
    }

    visitCase(node: PC.Case, context: CasePreparationData | null): LoopControlFlowEvent[] {
        if (node.condition) {
            this.compiler.expressionsVisitor.compileNode(node.condition, context);
        } else {
            this.compiler.emitSimplePrefix(CCS.CCSInternalChannel.name);
        }
        return this.compileNode(node.statement);
    }

    

    visitStatementWrapper(node: PC.StatementWrapper, context?: unknown): LoopControlFlowEvent[] {
        if (node.isNoOp) {
            return [];
        } else {
            return this.compileNode(node.statement!);
        }
    }

    visitBreakStatement(node: PC.BreakStatement, context?: unknown): LoopControlFlowEvent[] {
        return [LoopControlFlowEvent.newBreakEvent(this.compiler.emitProcessApplicationPlaceholder())];
    }

    visitContinueStatement(node: PC.ContinueStatement, context?: unknown): LoopControlFlowEvent[] {
        // const processFrame = utils.assertNonNull(this.compiler.getProcessFrame());
        // utils.assertNonNull(this.loopEntry).emitCallProcessFromFrame(this.compiler, processFrame);
        // return [];
        return [LoopControlFlowEvent.newContinueEvent(this.compiler.emitProcessApplicationPlaceholder())];
    }

    visitStatementBlock(node: PC.StatementBlock, context?: unknown): LoopControlFlowEvent[] {
        this.compiler.reopenEnvironment(node);
        const statusQuo = utils.assertNonNull(this.compiler.getProcessFrame());
        this.compiler.emitNewScope();
        const individualBreaks = node.blockStatements.map((bs) => {
            if (bs instanceof PC.StatementWrapper) {
                return this.compileNode(bs);
            } else {
                // bs is VariableDeclarations or ConditionDeclaration
                this.compiler.programVisitor.compileNode(bs);
                return [];
            }
        });
        const breaks = individualBreaks.flat();
        this.compiler.emitNewScope(statusQuo);
        this.compiler.closeEnvironment();
        return breaks;
    }
    

    visitStatementExpression(node: PC.StatementExpression, context?: unknown): LoopControlFlowEvent[] {
        this.compiler.expressionsVisitor.compileNode(node.expression);
        return [];
    }


    visitIfStatement(node: PC.IfStatement, context?: unknown): LoopControlFlowEvent[] {
        // Keep the pseuCo scopes in sync
        this.compiler.reopenEnvironment(node);
        // Create an array to collect app placeholders at the end of each branch
        const placeholders: Array<PCCApplicationPlaceholderStackElement> = new Array();
        // COmpile the predicate
        const predicate = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.predicate));
        // Use a choice to distinguish between the two alternatives
        const control = this.compiler.emitChoice();
        // Case 1: Predicate holds
        this.compiler.emitCondition(predicate);
        // Compile the consequence and remember the breaks
        let breaks = this.compileNode(node.consequence);
        // If the process is incomplete, emit a process application to the process name in which we will merge the other branch of the if-statement
        if (!this.compiler.isCurrentProcessCompleted()) {
            // Remember the placeholder to continue with yet unknown process definition
            placeholders.push(this.compiler.emitProcessApplicationPlaceholder());
        }
        // The left branch of the choice is finished, continue with the right branch
        control.setBranchFinished();
        // Case 2: Predicate does not hold
        this.compiler.emitCondition(new PCCUnaryContainer("!", predicate));
        // If the node defines an alternative, compile it (otherwise nothing happens)
        if (node.alternative) {
            breaks = breaks.concat(this.compileNode(node.alternative));
        }
        // Again, possibly add an application to yet unknown merged process frame
        if (!this.compiler.isCurrentProcessCompleted()) {
            placeholders.push(this.compiler.emitProcessApplicationPlaceholder());
        }
        // Right branch is also finished
        control.setBranchFinished();
        // Now do the merge
        this.compiler.emitMergeOfProcessFramesOfPlaceholders(placeholders);
        // Keep the pseuCo scopes in sync
        this.compiler.closeEnvironment();
        // Return the collected breaks
        return breaks;
    }


    visitWhileStatement(node: PC.WhileStatement, context?: unknown): LoopControlFlowEvent[] {
        // Keep the pseuCo scopes in sync
        this.compiler.reopenEnvironment(node);
        // To get a "jump destination" to continue with the next loop iteration, we must start a new process frame (i.e., a new CCS process definition)
        const entry = this.compiler.emitNextProcessFrame();
        // We evaluate the predicate
        const predicate = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.predicate));
        // Use a choice to distinguish between satisfied and unsatisfied predicate
        const control = this.compiler.emitChoice();
        // Case 1: Predicate holds
        this.compiler.emitCondition(predicate);
        // We want to at least one CCS step, so we emit a tau action
        this.compiler.emitSimplePrefix(CCS.CCSInternalChannel.name);
        // We compile the loop body and receive a list of process application placeholders that were emitted at locations where a break occurs
        const loopControls = this.compileNode(node.loopBody);
        // Jump back to the loop entry after that
        entry.emitCallProcessFromFrame(this.compiler, utils.assertNonNull(this.compiler.getProcessFrame()));
        // Case 1 is finished
        control.setBranchFinished();
        // Case 2: Predicate does not hold
        this.compiler.emitCondition(new PCCUnaryContainer("!", predicate));
        // Create a new process frame for the loop exit, i.e., for the first statement below the loop; the compiler automatically transitions to this new process frame
        const loopExit = this.compiler.emitNextProcessFrame();
        // Instantiate the break and continue application placeholder with the loop exit and entry
        loopControls.forEach((event) => {
            if (event.isBreakEvent) {
                loopExit.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            } else if (event.isContinueEvent) {
                entry.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            }
        });
        // Keep the pseuCo scopes in sync
        this.compiler.closeEnvironment();
        // No new breaks for outer loops have been seen here
        return [];
    }


    visitDoStatement(node: PC.DoStatement, context?: unknown): LoopControlFlowEvent[] {
        // Keep the pseuCo scopes in sync
        this.compiler.reopenEnvironment(node);
        // Remember the status quo of available variables
        const statusQuo = utils.assertNonNull(this.compiler.getProcessFrame());
        // To get a "jump destination" to continue with the next loop iteration, we must start a new process frame (i.e., a new CCS process definition)
        const entry = this.compiler.emitNextProcessFrame();
        // We compile the loop body and receive a list of process application placeholders that were emitted at locations where a break occurs
        const loopControls = this.compileNode(node.loopBody);
        // We also need a "jump destination" for the predicate check for the continue statement
        const predicateCheckFrame = this.compiler.emitNextProcessFrame();
        // We evaluate the predicate
        const predicate = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.predicate));
        // Use a choice to distinguish between satisfied and unsatisfied predicate
        const control = this.compiler.emitChoice();
        // Case 1: Predicate holds
        this.compiler.emitCondition(predicate);
        // We want to at least one CCS step, so we emit a tau action
        this.compiler.emitSimplePrefix(CCS.CCSInternalChannel.name);
        // Jump back to the loop entry after that
        entry.emitCallProcessFromFrame(this.compiler, utils.assertNonNull(this.compiler.getProcessFrame()));
        // Case 1 is finished
        control.setBranchFinished();
        // Case 2: Predicate does not hold
        this.compiler.emitCondition(new PCCUnaryContainer("!", predicate));
        // Go to the first instruction after the loop; we don't know yet where to go, so we simply add a placeholder to the existing break placeholders
        loopControls.push(LoopControlFlowEvent.newBreakEvent(this.compiler.emitProcessApplicationPlaceholder()));
        // Case 2 is finished
        control.setBranchFinished();
        // Create a new process frame for the loop exit, i.e., for the first statement below the loop; the compiler automatically transitions to this new process frame
        const loopExit = this.compiler.emitNextProcessFrame([statusQuo]);
        // Instantiate the break and continue application placeholder with the loop exit and entry
        loopControls.forEach((event) => {
            if (event.isBreakEvent) {
                loopExit.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            } else if (event.isContinueEvent) {
                predicateCheckFrame.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            }
        });
        // Keep the pseuCo scopes in sync
        this.compiler.closeEnvironment();
        // No new breaks for outer loops have been seen here
        return [];
    }

    visitForStatement(node: PC.ForStatement, context?: unknown): LoopControlFlowEvent[] {
        // Keep the pseuCo scopes in sync
        this.compiler.reopenEnvironment(node);
        // Remember the status quo of available variables
        const statusQuo = utils.assertNonNull(this.compiler.getProcessFrame());
        // loop initialisation
        if (node.loopInit) {
            // Do the initialisation in a new scope; introduced variables aren't supposed to be accessible after the loop
            this.compiler.emitNewScope();
            this.compileNode(node.loopInit);
        }
        // At this point we want to continue with the loop body. However, in every following loop iteration, we will have to execute the update instruction. For the first iteration we want to skip these instructions. Thus, we emit a process application placeholder (because we don't know yet where to jump to)
        const toLoopBody = this.compiler.emitProcessApplicationPlaceholder();

        // For the next iterations, we define this point as the loop entry (note that `emitNextProcessFrame` emits a process application by construction. This application is, however, not reachable (because of the placeholder emitted in the line above). Still, we use `emitNextProcessFrame`, because it is easier to use than manually installing a new process frame)
        const entry = utils.assertNonNull(this.compiler.emitNextProcessFrame());
        // Compile the update statements
        node.loopUpdates.forEach((u) => this.compileNode(u));
        // Here begins the loop body
        const loopBody = this.compiler.emitNextProcessFrame();
        // Transition from the above loop body placeholder to the process
        loopBody.emitCallProcessFromFrame(this.compiler, toLoopBody.frame, toLoopBody);

        // Initialise the array to collect the break app placeholders
        let loopControls: Array<LoopControlFlowEvent> = [];
        // Create a variable for a choice control element; we only need it if the loop defines a condition
        let control: PCCChoiceStackElement | null = null;
        // Compile loop condition if available
        if (node.predicate) {
            // Compile the predicate
            const predicate = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.predicate));
            // Setup the choice operator
            control = this.compiler.emitChoice();
            // Case 1: predicate does not hold
            this.compiler.emitCondition(new PCCUnaryContainer("!", predicate));
            // We move beyond the loop; we so this by simply adding placeholder to breaks
            loopControls.push(LoopControlFlowEvent.newBreakEvent(this.compiler.emitProcessApplicationPlaceholder()));
            // Case 1 finished
            control.setBranchFinished();
            // Case 2: Predciate does hold
            this.compiler.emitCondition(predicate);
            // Continue as if no predicate would have existed...
        }
        // We want to at least one CCS step, so we emit a tau action
        this.compiler.emitSimplePrefix(CCS.CCSInternalChannel.name);
        // Compile the loop body and remember the break placeholders
        loopControls = loopControls.concat(this.compileNode(node.loopBody));
        // Jump back to the loop entry
        entry.emitCallProcessFromFrame(this.compiler, utils.assertNonNull(this.compiler.getProcessFrame()));
        // If we have a loop predicate, Case 2 is finished now
        control?.setBranchFinished();

        // Create a new process frame for the loop exit, i.e., for the first statement below the loop; the compiler automatically transitions to this new process frame
        const loopExit = this.compiler.emitNextProcessFrame([statusQuo]);
        // Instantiate the break and continue application placeholder with the loop exit and entry
        loopControls.forEach((event) => {
            if (event.isBreakEvent) {
                loopExit.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            } else if (event.isContinueEvent) {
                entry.emitCallProcessFromFrame(this.compiler, event.appPlaceholder.frame, event.appPlaceholder);
            }
        });
        // Keep the pseuCo scopes in sync
        this.compiler.closeEnvironment();
        // No new breaks for outer loops have been seen here
        return [];
    }

    visitForInit(node: PC.ForInit, context?: unknown): LoopControlFlowEvent[] {
        // COmpile the children depending on the concrete class type of the children
        node.variableInitialisations.forEach((c) => {
            if (c instanceof PC.StatementExpression) {
                this.compileNode(c);
            } else {
                this.compiler.programVisitor.compileNode(c);
            }
        });
        // No break statements are allowed to occur here
        return [];
    }

    visitReturnStatement(node: PC.ReturnStatement, context?: unknown): LoopControlFlowEvent[] {
        if (node.expression) {
            // Tell compiler that we need the return auxiliary process
            this.compiler.setNeedsReturn();
            // COmpile the expression
            const returnValue = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.expression));
            // Send the value to the auxiliary return process
            this.compiler.emitOutput("return", null, returnValue);
        }
        // Obtain the procedure in which the return is used
        const proc = this.compiler.getCurrentProcedure();
        if (proc) {
            // Let the procedure compile its exit (this is for example relevant for monitors, which must release a lock etc.)
            proc.emitExit(this.compiler);
        } else {
            // This can only happen if we are in the main agent. In this case we emit the CCS exit process
            this.compiler.emitExit();
        }
        // No break statements to be expected here
        return [];
    }

    visitPrimitiveStatement(node: PC.PrimitiveStatement, context?: unknown): LoopControlFlowEvent[] {
        // Distinguish between the different primitives
        if (node.primitive == "join") {
            // Tell compiler that we need the join auxiliary process
            this.compiler.setNeedsAgentJoiner();
            // Obtain the agent id
            const c = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.expression));
            // Register as a waiting agent
            this.compiler.emitOutput("join_register", c, null);
            // Wait for c's termination
			this.compiler.emitOutput("join", c, null);
        } else if (node.primitive == "lock") {
            // obtain the lock reference
            const c = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.expression));
            // obtain the current agent
            const agentVar = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("a", null, true));
            const a = agentVar.getContainer(this.compiler);
            // get the lock by synchronising with the lock auxiliary process
            this.compiler.emitOutput("lock", c, a);
        } else if (node.primitive == "unlock") {
            // obtain the lock reference
            const c = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.expression));
            // obtain the current agent
            const agentVar = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("a", null, true));
            const a = agentVar.getContainer(this.compiler);
            // release the lock by synchronising with the lock auxiliary process
            this.compiler.emitOutput("unlock", c, a);
        } else if (node.primitive == "assert") {
            // obtain the predicate value
            const c = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(node.expression));
            // use a choice to account for predciate's two values
            const control = this.compiler.emitChoice();
            // Case 1: assertion is violated
            this.compiler.emitCondition(new PCCUnaryContainer("!", c));
            // Output exception action
            this.compiler.emitOutput("exception", null, new PCCConstantContainer(`Assertion failed in line ${node.line}: \`${node.expression.toString()}\`!`));
            // Stop the agent
            this.compiler.emitStop();
            // End of case 1
			control.setBranchFinished();
            // Case 2: Assertion is satisfied
			this.compiler.emitCondition(c);
            // Continue as if nothing ever happened...
        }
        // No break statements in sight here
        return [];
    }

    visitConditionStatement(node: PC.ConditionStatement, context?: unknown): LoopControlFlowEvent[] {
        // Distinguish between the three keywords
        if (node.modifier == "waitForCondition") {
            // An identifier identifying the condition is necessary here
            if (!node.identifier) {
                throw new Error("Expected identifier for `waitForCondition`");
            }
            // get the condition variable
            const condition = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
            // we must implement a while loop
            // while(!condition) {
            //   addMyselfToWaitingRoom();
            //   int reentrancyLevel = fullUnlock(implicitLock);
            //   waitForASignal();
            //   /* Try a regular lock with my agent ID first */
            //   lock(implicitLock, myAgentID);
            //   /* When succeeded, restore the same reentrancy level as before the wait. Note that this needs two steps, because in CCS we can have at most two 'arguments' */
            //   multilock(implicitLock, reentrancyLevel - 1);
            //}

            // To get a "jump destination" to continue with the next loop iteration, we must start a new process frame (i.e., a new CCS process definition)
            const entry = this.compiler.emitNextProcessFrame();
            // the loop predicate is the expression encoded in the condition
            const conditionDecl = utils.assertNonNull(condition.node);
            if (!(conditionDecl instanceof PC.ConditionDeclaration)) {
                throw new Error("Expected COnditionDeclaration");
            }
            const predicate = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(conditionDecl.expression));
            // Use the choice operator to distinguish between pedicate's truth value
            const control = this.compiler.emitChoice();
            // Case 1: predicate does not hold
            this.compiler.emitCondition(new PCCUnaryContainer("!", predicate));
            // obtain the condition reference
            const condRef = condition.getContainer(this.compiler);
            // add ourselves to the waiting room
            this.compiler.emitOutput("add", condRef, null);
            // obtain the implicit lock
            const lockVar = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("guard", null, true));
            const lock = lockVar.getContainer(this.compiler);
            // get current agent ID
            const agentVar = utils.assertNonNull(this.compiler.getVariableWithNameOfClass("a", null, true));
            const a = agentVar.getContainer(this.compiler);
            // create a fresh container to store the reentrancy level
            const reentrancyLevel = this.compiler.getFreshContainer(CCSType.INT);
            // perform to full unlock by syncing with the lock auxiliary process; we receive the reentrancy level
            this.compiler.emitInput("fullunlock", lock, reentrancyLevel);
            // wait for a signal by syncing with the waiting room aux process
            this.compiler.emitOutput("wait", condRef, null);
            // try to regularly obtain the lock by syncing with the lock aux process (note that we cannot at the same time pass the agent id and the reentrancy level to the process due to CCS restrictions)
            this.compiler.emitOutput("lock", lock, a);
            // Restore the reentrancy level from before the waiting (note that the level is already 1 due to regular lock)
            const newReentrencyLevel = new PCCBinaryContainer(reentrancyLevel, new PCCConstantContainer(1), "-");
            this.compiler.emitOutput("multilock", lock, newReentrencyLevel);
            // Jump back to the beginning of the loop
            entry.emitCallProcessFromFrame(this.compiler, utils.assertNonNull(this.compiler.getProcessFrame()));
            // End of case 1
            control.setBranchFinished();
            // Case 2: The predicate holds
			this.compiler.emitCondition(predicate);
            // Continue as nothing ever happened...
        } else if (node.modifier == "signal") {
            // An identifier identifying the condition is necessary here
            if (!node.identifier) {
                throw new Error("Expected identifier for `waitForCondition`");
            }
            // get the condition variable
            const condition = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
            // and condition reference
            const condRef = condition.getContainer(this.compiler);
            // send the signal to the waiting room aux process
            this.compiler.emitOutput("signal", condRef, null);
        } else if (node.modifier == "signalAll") {
            // The identifier may be omitted in this case; if omitted, all of the monitor's conditions are signalled
            let vars: Array<PCCVariable> = [];
            if (node.identifier) {
                // If not omitted, get the variable for the condition and assign to `vars`
                const condition = utils.assertNonNull(this.compiler.getVariableWithName(node.identifier));
                vars = [condition];
            } else {
                // if omitted, assign all condition variables to `vars`
                const currentClass = utils.assertNonNull(this.compiler.getCurrentClass());
                vars = currentClass.getAllConditions();
            }
            // Now forward the signal_all to the waiting room aux processes of all conditions in vars
            vars.forEach((v) => this.compiler.emitOutput("signal_all", v.getContainer(this.compiler), null));
        }
        // No break statements in sight here
        return [];
    }

    
    visitPrintStatement(node: PC.PrintStatement, context?: unknown): LoopControlFlowEvent[] {
        if (node.argumentExpressions.length == 0) {
            // Nothing to do if println doesn't have arguments
            return [];
        }
        // Compile every argument and protect it
        node.argumentExpressions.forEach((exp) => {
            const container = utils.assertNonNull(this.compiler.expressionsVisitor.compileNode(exp));
            this.compiler.protectContainer(container);
        });
        // Recover the values -- in reversed order, don't be confused by the map
        const args = node.argumentExpressions.map((_) => this.compiler.unprotectContainer());
        // Reverse the order to get the original order again
        args.reverse();
        // String-concatenate all arguments
        let final = args[0];
        for (let i = 1; i < args.length; i++) {
            final = new PCCBinaryContainer(final, args[i], "^");
        }
        // print the result
        this.compiler.emitOutput("println", null, final);
        // No break statements in sight here
        return [];
    }





    // The remaining nodes are not statements
    visitProgram(node: PC.Program, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitMainAgent(node: PC.MainAgent, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitProcedureDeclaration(node: PC.ProcedureDeclaration, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitProcedureParameter(node: PC.ProcedureParameter, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitMonitor(node: PC.Monitor, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitStruct(node: PC.Struct, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitConditionDeclaration(node: PC.ConditionDeclaration, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitVariableDeclarations(node: PC.VariableDeclarations, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitVariableDeclarator(node: PC.VariableDeclarator, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitVariableInitializer(node: PC.VariableInitializer, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitArrayTypeNode(node: PC.ArrayTypeNode, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitSimpleTypeNode(node: PC.SimpleTypeNode, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitChannelTypeNode(node: PC.ChannelTypeNode, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitClassTypeNode(node: PC.ClassTypeNode, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitStartExpression(node: PC.StartExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitAssignExpression(node: PC.AssignExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitAssignDestination(node: PC.AssignDestination, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitSendExpression(node: PC.SendExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitConditionalExpression(node: PC.ConditionalExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitOrExpression(node: PC.OrExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitAndExpression(node: PC.AndExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitEqualityExpression(node: PC.EqualityExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitRelationalExpression(node: PC.RelationalExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitAdditiveExpression(node: PC.AdditiveExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitMultiplicativeExpression(node: PC.MultiplicativeExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitUnaryExpression(node: PC.UnaryExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitPostfixExpression(node: PC.PostfixExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitReceiveExpression(node: PC.ReceiveExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitProcedureCall(node: PC.ProcedureCall, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitClassCall(node: PC.ClassCall, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitArrayExpression(node: PC.ArrayExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitLiteralExpression(node: PC.LiteralExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }
    visitIdentifierExpression(node: PC.IdentifierExpression, context?: unknown): LoopControlFlowEvent[] {
        throw new Error("Method not implemented.");
    }

}