import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import * as CCS from "@pseuco/ccs-interpreter";
import { CCSType, PCCClass, PCCGlobal, PCCGlobalVariable, PCCProcedure, PCCProgramController, PCCVariable, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCBinaryContainer, PCCConstantContainer, PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
import { PCCGroupable, PCCProcedureFrame, PCCProcessFrame, PCCSimpleGroupable } from "./PCCProcessFrame.js";
import { PCCApplicationPlaceholderStackElement, PCCApplicationStackElement, PCCChoiceStackElement, PCCClassStackElement, PCCCompilerStack, PCCConditionStackElement, PCCControlElement, PCCExitStackElement, PCCGlobalStackElement, PCCInputStackElement, PCCMatchStackElement, PCCOutputStackElement, PCCParallelStackElement, PCCPrefixStackElement, PCCProcedureStackElement, PCCProcessDefinitionStackElement, PCCProcessFrameStackElement, PCCRestrictionStackElement, PCCSequenceStackElement, PCCStackElement, PCCStackResultContainer, PCCStackResultType, PCCStopStackElement, PCCSystemProcessStackElement } from "./PCCCompilerStack.js";
import { utils } from "./utils.js";
import { PCCCompileProgramVisitor, PCCVariableInitialisation } from "./PCCCompileProgramVisitor.js";
import { PCCCompileStatementsVisitor } from "./PCCCompileStatementsVisitor.js";
import { PCCCompileExpressionsVisitor } from "./PCCCompileExpressionsVisitor.js";

type PCCGroupStackElement = PCCClassStackElement | PCCProcedureStackElement | PCCSystemProcessStackElement | PCCProcessFrameStackElement | PCCProcessDefinitionStackElement;

export class PCCCompiler {
    // The `stack` is used to construct the CCS processes and process definitions. Intuitively, the compiler pushes CCS fragments on this stack and the CCS process is constructed by destructing the stack and by thus building the CCS process from the end (i.e., the stop process) to the beginning (i.e., by adding prefixes before the existing process)
    private stack: PCCCompilerStack;
    // The `groupElements` are special elements on the stack, for example binary operators
    private readonly groupElements: Array<PCCGroupStackElement> = new Array();
    // The `controller` performs some additional analysis of the program and constructs objects that represent variables, classes, and procedures. These objects provide further instructions for compilation; the concrete instructions differ depending on the type of the variable and where it appears (e.g., local variables are treated differently than global variables)
    private readonly controller: PCCProgramController;

    // Here we collect completely built CCS objects
    private readonly processDefinitions: Array<CCS.CCSProcessDefinition> = new Array();
    private readonly procedureProcessDefinitions: Array<CCS.CCSProcessDefinition> = new Array(); // subset of process definitions that represent procedures
    private readonly systemProcesses: Map<number, Array<CCS.CCSProcess>> = new Map();
    // `compilingNodes` is a stack that remembers which pseuCo node is being compiled
    private readonly compilingNodes: Array<PC.Node> = new Array();

    // For the actual compilation we use the visitor pattern; we have separate visitors for expressions, statements and declarations (because they must have different return types)
    readonly programVisitor = new PCCCompileProgramVisitor(this, this.compilingNodes);
    readonly statementsVisitor = new PCCCompileStatementsVisitor(this, this.compilingNodes);
    readonly expressionsVisitor = new PCCCompileExpressionsVisitor(this, this.compilingNodes);

    // To avoid unnecessary CCS processes, we remember which constructs are actually necessary for a pseuCo program
    private needsReturn = false;
    private needsMutex = false;
    private needsWaitRoom = false
	private needsArrayManager = false
	private needsChannelManager = false
	private needsAgentManager = false
	private needsAgentJoiner = false

    // Remember if there are global variables that need custom initialisation

    /**
     * Creates a new compiler object to compile `program`, which is well-typed according to `staticSemantics`.
     * @param program The program that shall be compiled to a CCS process.
     * @param staticSemantics The static semantics objects that has been used to check whether `program` is well-typed.
     */
    constructor(readonly program: PC.Program, readonly staticSemantics: PCT.StaticSemantics) {
        this.controller = new PCCProgramController(program, staticSemantics);
        this.stack = new PCCCompilerStack();
        this.stack.pushElement(new PCCGlobalStackElement(this.controller.getGlobal()));
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled has functions (or methods) that return a value.
     * Adds system process `Return`.
     */
    setNeedsReturn() {
        this.needsReturn = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled uses at least one lock (formerly called mutex).
     * Adds system processes `Mutex[i, c, a]` and `Mutex_cons[next_i]`.
     */
    setNeedsMutex() {
        this.needsMutex = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled has at least one `waitForCondition` instruction.
     * Adds system processes `WaitRoom[i, c]`, `WaitDistributor[i, c]`, and `WaitRoom_cons[next_i]`.
     */
    setNeedsWaitRoom() {
        this.needsWaitRoom = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled uses arrays; the array manager allows different array constructors (for different array sizes) to share the same reference domain. That is, when a new array is created, the array manager provides the new array's "pointer reference".
     * Adds system process `ArrayManager[next_i]`.
     */
    setNeedsArrayManager() {
        this.needsArrayManager = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled uses buffered channels; the array manager allows different channel constructors (for different channel capacities) to share the same reference domain.
     * Adds system process `ChannelManager[next_i]`.
     */
    setNeedsChannelManager() {
        this.needsChannelManager = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled starts agents other than the main agent; the agent manages the unique agent identifiers.
     * Adds system process `AgentManager[next_i]`.
     */
    setNeedsAgentManager() {
        this.needsAgentManager = true;
    }

    /**
     * Report to the compiler that the pseuCo program that is being compiled waits on other agents' termination using `join`. 
     * Adds system processes `AgentJoiner[a, c]` and `JoinDistributor[a, c]`.
     */
    setNeedsAgentJoiner() {
        this.needsAgentJoiner = true;
    }

    /**
     * Reports to the compiler that a particular procedure is started as a separate agent. 
     * @param procedure The procedure that is started as an agent. 
     */
    processProcedureAsAgent(procedure: PCCProcedure) {
        this.controller.processProcedureAsAgent(procedure);
    }

    /**
     * Compiles the program to CCS.
     * @returns Returns a CCS object that holds all process definitions and the main system process.
     */
    compileProgram(): CCS.CCS {
        // Obtain additional information about the program that is not available from the staticSemantics.
        this.controller.analyseProgram();
        const usedTypes = this.controller.getUsedTypes();

        // If there is a buffered channel used in the program, we need a channel manager
        for (const capacity of usedTypes.channels) {
            if (capacity > 0) {
                this.setNeedsChannelManager();
                break;
            }
        }

        // If there is at least one array, we need an array manager
        this.needsArrayManager = (usedTypes.arrays.size > 0);

        // Use the declarations visitor on the root node of the AST to start compiling the program.
        this.programVisitor.compileNode(this.program);

        // Add a constructor for every monitor and struct
        this.assertEmptyGroupStack();
        for (const cls of this.controller.getAllClasses()) {
            cls.emitConstructor(this);
        }

        // Compile system processes that are required for the program being compiled
        this.assertEmptyGroupStack();
        if (this.needsReturn) {
            this.compileReturn();
            this.assertEmptyGroupStack();
        }
        if (this.needsMutex) {
            this.compileReentrantMutex();
            this.assertEmptyGroupStack();
        }
        if (this.needsWaitRoom) {
            this.compileWaitRoom();
            this.assertEmptyGroupStack();
        }
        if (this.needsArrayManager) {
            this.compileArrayManager();
            this.assertEmptyGroupStack();
        }
        // Add a constructor for every array size that appears in the program
        for (const size of usedTypes.arrays) {
            this.compileArrayWithCapacity(size);
        }
        this.assertEmptyGroupStack();
        if (this.needsChannelManager) {
            this.compileChannelManager();
            this.assertEmptyGroupStack();
        }
        // Add a constructor for every channel capacity that appears in the program
        for (const capacity of usedTypes.channels) {
            this.compileChannelWithCapacity(capacity);
        }
        this.assertEmptyGroupStack();
        if (this.needsAgentManager) {
            this.compileAgentTools(this.needsAgentJoiner);
            this.assertEmptyGroupStack();
        }

        // For every procedure that gets started as an agent, add a processes that starts this agent. 
        for (const p of this.controller.getAgents()) {
            p.emitAgentConstructor(this, this.needsAgentJoiner);
            this.beginSystemProcess(PCCCompiler.SysAgent);
            // The arguments to the procedure are sent via channels instead of passing them to the process as arguments (because the latter is not possible given that that the agent starting mechanism is a top-level system process, while the start instruction occurs inside an existing agent)
            this.emitProcessApplication(p.getAgentProcessName(), []);
            this.endSystemProcess();
        }
        this.assertEmptyGroupStack();

        // We initialise the global variables before the main agent starts. To do so, we build a separate process, if necessary.
        // Remember if there are global variables that need custom initialisation
        const globalVariablesNeedInitialisation = this.compileGlobalVariableInitialisation();

        // We now have all processes and process definitions. We will now reorder and structure them a bit to make the resulting CCS better readable.
        // In this array we'll collect CCS process definitions that define system meta functionality (the CCSEnvironmentNonSemanticLine objects will be empty lines to structure the CCS code):
        const sysDefs = new Array<CCS.CCSProcessDefinition | CCS.CCSEnvironmentNonSemanticLine>();
        // CCS process definitions that define procedures have already been separated:
        const procDefs: Array<CCS.CCSProcessDefinition | CCS.CCSEnvironmentNonSemanticLine> = this.procedureProcessDefinitions;
        // In this array we'll collect CCS process definitions that define the main agent (the CCSEnvironmentNonSemanticLine objects will be empty lines to structure the CCS code):
        const mainAgentDefs = new Array<CCS.CCSProcessDefinition | CCS.CCSEnvironmentNonSemanticLine>();
        // Split the system processes and the main agent processes:
        for (const def of this.processDefinitions) {
            if (def.name.substring(0,9) == "MainAgent") {
                mainAgentDefs.push(def);
            } else {
                sysDefs.push(def);
            }
        }

        // Improve CCS code readability
        const findFirstProcessDefinitionIndex = function (pdefs: Array<CCS.CCSProcessDefinition | CCS.CCSEnvironmentNonSemanticLine>, predicate: ((pdef: CCS.CCSProcessDefinition) => boolean)) {
            return pdefs.findIndex((pdef) => {
                if (pdef instanceof CCS.CCSProcessDefinition) {
                    return predicate(pdef);
                } else {
                    return false;
                }
            })
        };

        // Insert some helpful comments
        // Global variables
        const firstGlobalEnvProcessIndex = findFirstProcessDefinitionIndex(sysDefs, (pdef) => pdef.name.indexOf("Env_global") === 0);
        if (firstGlobalEnvProcessIndex >= 0) {
            sysDefs.splice(firstGlobalEnvProcessIndex, 0, new CCS.CCSEnvironmentCommentLine("Global variables"));
        }

        // Classes
        let firstClassEnvProcessIndex = findFirstProcessDefinitionIndex(sysDefs, (pdef) => pdef.name.indexOf("Env_class") === 0);
        if (firstClassEnvProcessIndex >= 0) {
            if (firstGlobalEnvProcessIndex >= 0) {
                // Separate global variables and classes by an empty line
                sysDefs.splice(firstClassEnvProcessIndex, 0, new CCS.CCSEnvironmentBlankLine());
                // Everything is shifted one line further down
                firstClassEnvProcessIndex++;
            }
            sysDefs.splice(firstClassEnvProcessIndex, 0, new CCS.CCSEnvironmentCommentLine("Structs and monitors"));
        }

        // System processes
        let firstNonEnvProcessIndex = findFirstProcessDefinitionIndex(sysDefs, (pdef) => pdef.name.indexOf("Env_") !== 0);
        if (firstNonEnvProcessIndex > 0) {
            // If non environment process definition exists and if it's not the first one, insert an empty line to separate the environment process definitions from the rest
            sysDefs.splice(firstNonEnvProcessIndex, 0, new CCS.CCSEnvironmentBlankLine());
            // Everything is shifted one line further down
            firstNonEnvProcessIndex++;
        }
        if (firstNonEnvProcessIndex >= 0) {
            sysDefs.splice(firstNonEnvProcessIndex, 0, new CCS.CCSEnvironmentCommentLine("System processes"));
        }

        // Procedures
        if (procDefs.length > 0) {
            procDefs.unshift(new CCS.CCSEnvironmentCommentLine("Procedures"));
        }

        // Main agent
        if (mainAgentDefs.length > 0) {
            mainAgentDefs.unshift(new CCS.CCSEnvironmentCommentLine("Main agent"));
        }



        // Make the CCS more readable by inserting empty lines between the process definitions       
        // Insert empty line between process definitions and remaining system environments
        if (procDefs.length > 0 && sysDefs.length > 0) {
            sysDefs.push(new CCS.CCSEnvironmentBlankLine())
        }
        if (mainAgentDefs.length > 0 && (procDefs.length > 0 || sysDefs.length > 0)) {
            mainAgentDefs.unshift(new CCS.CCSEnvironmentBlankLine());
        }

        


        // Concatenate all process definitions in the new order (including the empty lines)
        const allPdefs = sysDefs.concat(procDefs, mainAgentDefs);

        // Get the main system that uses the process definitions obtained above
        const system = this.getSystem(globalVariablesNeedInitialisation);

        // For debugging purposes, we may dump the generated CCS before it is checked by the CCS interpreter package
        // this.__dumpUncheckedCCS(allPdefs, system);

        // Create the CCS object and return it
        const ccsEnv = new CCS.CCSEnvironment(allPdefs, []);
        return new CCS.CCS(ccsEnv, system);
    }

    private __dumpUncheckedCCS(pdefs: Array<CCS.CCSProcessDefinition>, system: CCS.CCSProcess) {
        console.log('Dumping unchecked CCS.')
		let res = "";
        for (const pd of pdefs) {
            res += pd.toString() + "\n";
        }
		res += `\n${system.toString()}`;
		console.log(res);
    }

    private assertEmptyGroupStack() {
        if (this.groupElements.length > 0) {
            throw new Error("Assertion violation: @groupElements must be empty!");
        }
    }



    private pushStackElement(element: PCCStackElement) {
        element.pseucoNode = this.compilingNodes[this.compilingNodes.length - 1]; // TODO: Replace this line with `rememberPseucoCCSRelationship`?
        this.stack.pushElement(element);
    }

    private getSystem(globalVariablesNeedInitialisation: boolean): CCS.CCSProcess {
        this.compileMainAgentApplication(globalVariablesNeedInitialisation);

        // Now the system processes are ordered by weight in a weird way. But I don't change it now, because we want to compare if the old and new compiler produce equal CCS --> TODO for later
        const weights = Array.from(this.systemProcesses.keys()).filter((n) => !isNaN(n));
        weights.sort((a,b) => a - b);

        let system: CCS.CCSProcess | null = null;
        for (const w of weights) {
            const mainProcesses = this.systemProcesses.get(w)!;
            for (const p of mainProcesses) {
                if (system) {
                    system = new CCS.CCSParallel(system, p);
                } else {
                    system = p;
                }
            }
        }
        
        return new CCS.CCSRestriction(system!, this.getRestrictedChannels());
    }

    private getRestrictedChannels(): Array<string> {
        return ["*", "println", "exception"];
        // Removed from previous CoffeeScript version: tracking channels
    }

    getVariableWithName(name:string): PCCVariable | null {
        return this.getVariableWithNameOfClass(name, null, false);
    }

    getVariableWithNameOfClass(name: string, className: string | null, isInternal: boolean): PCCVariable | null {
        const internalName = (isInternal) ? `#${name}` : name;
        if (className != null) {
            return this.controller.getClassWithName(className).compilerGetVariable(this, internalName);
        } else {
            return this.stack.compilerGetVariable(this, internalName);
        }
    }

    getProcedureWithName(name: string): PCCProcedure {
        return this.getProcedureWithNameOfClass(name, null);
    }

    getProcedureWithNameOfClass(name: string, className: string | null): PCCProcedure {
        if (className) {
            return this.controller.getClassWithName(className).getProcedureWithName(name);
        } else {
            return this.controller.getProcedureWithName(name);
        }
    }

    private getClassWithName(name: string): PCCClass {
        return this.controller.getClassWithName(name);
    }

    getCurrentClass(): PCCClass | null {
        for(const e of this.groupElements) {
            if (e instanceof PCCClassStackElement) {
                return e.classInfo;
            }
        }
        return null;
    }

    getCurrentProcedure(): PCCProcedure | null {
        for (const e of this.groupElements) {
            if (e instanceof PCCProcedureStackElement) {
                return e.procedure;
            }
        }
        return null;
    }

    private getGlobal(): PCCGlobal {
        return this.controller.getGlobal();
    }

    private rememberPseucoCCSRelationship(pseucoNode: PC.Node, ccsObject: CCS.CCSProcess | CCS.CCSExpression) {
        // Do nothing for now. Fill this method if eventually you want to support relationship tracking.
    }

    getFreshContainer(ccsType: CCSType, desiredName?: string): PCCVariableContainer {
        const res = utils.assertNonNull(this.getProcessFrame()).createContainer(ccsType, desiredName);
        // if (this.compilingNodes.length > 0) {
        //     this.rememberPseucoCCSRelationship(this.compilingNodes[this.compilingNodes.length - 1], res);
        // }
        return res;
    }

    handleNewVariable(variable: PCCVariableInfo) {
        this.stack.compilerHandleNewVariable(this, variable);
    }

    private getControlElement(): PCCControlElement | null {
        return this.stack.getCurrentControlElement();
    }

    private handleStackResult(resultContainer: PCCStackResultContainer, controlElement: PCCControlElement) {
        for (const result of resultContainer.allResults) {
            if (result.type == PCCStackResultType.CCSPROCESS_DEFINITION) {
                // controlElement.compilerPushPDef(result.data);
                this.processDefinitions.push(result.ccsProcessDefinition);
            } else if (result.type == PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE) {
                this.procedureProcessDefinitions.push(result.ccsProcessDefinition);
            }
        }
    }

    private beginSystemProcess(weight: number = PCCCompiler.SysUnknownWeight) {
        const element = new PCCSystemProcessStackElement(weight);
        this.groupElements.push(element);
        this.pushStackElement(element);
    }

    private endSystemProcess() {
        const element = this.groupElements.pop();
        if (!(element instanceof PCCSystemProcessStackElement)) {
            throw new Error("Unexpected stack element!");
        }
        const weight = element.weight;
        const res = element.removeFromStack().getResult();
        if(!this.systemProcesses.has(weight)) {
            this.systemProcesses.set(weight, new Array(res.ccsProcess));
        } else {
            this.systemProcesses.get(weight)!.push(res.ccsProcess);
        }
    }

    emitSystemProcessApplication(processName: string, argumentContainers: Array<PCCContainer>, weight: number) {
        this.beginSystemProcess(weight);
        this.emitProcessApplication(processName, argumentContainers);
        this.endSystemProcess();
    }

    beginProcessGroup(groupable: PCCGroupable, variables: Array<PCCVariableInfo>) {
        const frame = new PCCProcessFrame(groupable, variables);
        const element = new PCCProcessFrameStackElement(frame);
        this.groupElements.push(element);
        this.pushStackElement(element);
        frame.emitProcessDefinition(this);
    }

    endProcessGroup() {
        const frame = this.groupElements.pop();
        if (!(frame instanceof PCCProcessFrameStackElement)) {
            throw new Error("Unexpected stack element!")
        }
        const controlElement = utils.assertNonNull(this.getControlElement());
        this.handleStackResult(frame.removeFromStack(), controlElement);
    }

    getProcessFrame(): PCCProcessFrame | null {
        return this.stack.getCurrentProcessFrame();
    }

    private addProcessGroupFrame(nextFrame: PCCProcessFrame) {
        this.pushStackElement(new PCCProcessFrameStackElement(nextFrame));
        nextFrame.emitProcessDefinition(this);
    }

    emitNewScope(derivationFrame?: PCCProcessFrame): PCCProcessFrame {
        const frame = utils.assertNonNull(this.getProcessFrame());
        if (!derivationFrame) {
            derivationFrame = frame;
        }
        const scope = derivationFrame.createScope();
        scope.emitTransitionFromFrame(this, frame);
        this.addProcessGroupFrame(scope);
        return scope;
    }

    emitNextProcessFrame(derivationFrames?: Array<PCCProcessFrame>): PCCProcessFrame {
        const frame = utils.assertNonNull(this.getProcessFrame());
        if (!derivationFrames || derivationFrames.length == 0) {
            derivationFrames = [frame];
        }
        const next = PCCProcessFrame.createFollowupFrameForFrames(derivationFrames);
        next.emitTransitionFromFrame(this, frame);
        this.addProcessGroupFrame(next);
        return next;
    }

    emitMergeOfProcessFramesOfPlaceholders(placeholders: Array<PCCApplicationPlaceholderStackElement>): PCCProcessFrame {
        if (placeholders.length == 0) {
            throw Error("Placeholders array must not be empty.");
        }
        const frames = placeholders.map((p) => p.frame);
        const followup = PCCProcessFrame.createFollowupFrameForFrames(frames);
        for (const p of placeholders) {
            followup.emitCallProcessFromFrame(this, p.frame, p);
        }
        this.addProcessGroupFrame(followup);
        return followup;
    }

    protectContainer(container: PCCContainer) {
        utils.assertNonNull(this.getProcessFrame()).protectContainer(container);
    }

    unprotectContainer(): PCCContainer {
        return utils.assertNonNull(this.getProcessFrame()).unprotectContainer();
    }

    getProtectedContainer(): PCCContainer {
        return utils.assertNonNull(this.getProcessFrame()).getProtectedContainer();
    }


    silentlyAddProcessDefinition(processName: string, argumentContainers: Array<PCCVariableContainer>): PCCProcessDefinitionStackElement {
        const element = new PCCProcessDefinitionStackElement(processName, argumentContainers);
        this.pushStackElement(element);
        return element;
    }

    beginProcessDefinition(processName: string, argumentContainers: Array<PCCVariableContainer>) {
        const element = this.silentlyAddProcessDefinition(processName, argumentContainers);
        this.groupElements.push(element);
    }

    isCurrentProcessCompleted(): boolean {
        return this.stack.isCurrentProcessCompleted();
    }

    endProcessDefinition() {
        const def = this.groupElements.pop();
        if (!(def instanceof PCCProcessDefinitionStackElement)) {
            throw new Error("Unexpected stack element!");
        }
        const controlElement = utils.assertNonNull(this.getControlElement());
        this.handleStackResult(def.removeFromStack(), controlElement);
    }


    beginClass(className: string) {
        this.controller.beginClass(className);
        const curClass = this.controller.getClassWithName(className);
        // according to types, curClass != undefined or null
        const element = new PCCClassStackElement(curClass);
		this.pushStackElement(element);
		this.groupElements.push(element);
    }

    endClass() {
        this.controller.endClass();
        const cls = this.groupElements.pop();
        if (!(cls instanceof PCCClassStackElement)) {
            throw new Error("Unexpected stack element!")
        }
        cls.removeFromStack();
    }

    beginMainAgent(node: PC.MainAgent) {
        this.controller.beginMainAgent(node);
        this.beginProcessGroup(new PCCSimpleGroupable("MainAgent"), [new PCCVariableInfo(null, "a", PCT.SystemTypes.agent, true)]);
    }

    endMainAgent() {
        this.controller.endMainAgent();
        this.endProcessGroup();
    }

    beginProcedure(procedureName: string) {
        this.controller.beginProcedure(procedureName);
        const procedure = this.getProcedureWithName(procedureName);
        const frame = new PCCProcedureFrame(procedure);
        const element = new PCCProcedureStackElement(procedure);
        this.pushStackElement(element);
        this.groupElements.push(element);
        this.addProcessGroupFrame(frame);
    }

    endProcedure() {
        this.controller.endProcedure();
        const proc = this.groupElements.pop();
        if (!(proc instanceof PCCProcedureStackElement)) {
            throw new Error("Unexpected stack element!");
        }
        const controlElement = utils.assertNonNull(this.getControlElement());
        this.handleStackResult(proc.removeFromStack(), controlElement);
    }

    reopenEnvironment(node: PC.Node) {
        this.controller.reopenEnvironment(node);
    }

    closeEnvironment() {
        this.controller.closeEnvironment();
    }


    /* Ignored the following (CoffeeScript):
     * beginStatement: (statement) ->
	*  endStatement: ->
	*
	*  beginExpression: (expression) ->
	*  endExpression: ->
    */

    private usingFrames(): boolean {
        return this.getProcessFrame() != null;
    }





    emitStop() {
        this.pushStackElement(new PCCStopStackElement());
    }

    emitExit() {
        this.pushStackElement(new PCCExitStackElement());
    }

    emitProcessApplication(processName: string, argumentContainers: Array<PCCContainer> = []) {
        this.pushStackElement(new PCCApplicationStackElement(processName, argumentContainers));
    }

    emitSimplePrefix(channel: string, channelParameter: PCCContainer | null = null) {
        this.pushStackElement(new PCCPrefixStackElement(channel, channelParameter));
    }

    emitOutput(channel: string, channelParameter: PCCContainer | null = null, valueContainer: PCCContainer | null = null) {
        this.pushStackElement(new PCCOutputStackElement(channel, channelParameter, valueContainer));
    }

    emitInput(channel: string, channelParameter: PCCContainer | null = null, container: PCCVariableContainer | null = null) {
        this.pushStackElement(new PCCInputStackElement(channel, channelParameter, container));
    }

    emitMatch(channel: string, channelParameter: PCCContainer | null, valueContainer: PCCContainer) {
        this.pushStackElement(new PCCMatchStackElement(channel, channelParameter, valueContainer));
    }

    emitCondition(condition: PCCContainer) {
        this.pushStackElement(new PCCConditionStackElement(condition));
    }

    emitChoice(): PCCChoiceStackElement {
        const res = new PCCChoiceStackElement();
        this.pushStackElement(res);
        if (this.usingFrames()) {
            this.emitNewScope();
        }
        return res;
    }

    emitParallel(): PCCParallelStackElement {
        const res = new PCCParallelStackElement();
        this.pushStackElement(res);
        if (this.usingFrames()) {
            this.emitNewScope();
        }
        return res;
    }

    emitSequence(): PCCSequenceStackElement {
        const res = new PCCSequenceStackElement();
        this.pushStackElement(res);
        // ToDo: Understand why we don't emit a new scope here
        return res;
    }

    emitRestriction(restrictedChannelNames: Array<string>) {
        this.pushStackElement(new PCCRestrictionStackElement(restrictedChannelNames));
    }

    emitProcessApplicationPlaceholder(): PCCApplicationPlaceholderStackElement {
        const processFrame = utils.assertNonNull(this.getProcessFrame());
        const res = new PCCApplicationPlaceholderStackElement(processFrame);
        this.pushStackElement(res);
        return res;
    }





    private compileMainAgentApplication(initGlobalVars: boolean) {
        this.beginSystemProcess(PCCCompiler.SysAgent);
        if (initGlobalVars) {
            // We must initialise global variables and run the main agent sequentially
            const control = this.emitSequence();
            // Initialise all global variables
            this.emitProcessApplication("GlobalVariablesInitialisations", []);
            control.setBranchFinished();
            // Start the main agent (which has agent ID 1)
            this.emitProcessApplication("MainAgent", [new PCCConstantContainer(1)]);
            control.setBranchFinished();
        } else {
            // Start the main agent (which has agent ID 1)
            this.emitProcessApplication("MainAgent", [new PCCConstantContainer(1)]);
        }
        this.endSystemProcess();
    }


    // Returns true if global initialisations are necessary, otherwise false
    private compileGlobalVariableInitialisation(): boolean {
        // Collect all variable declarations
        const allDecls = new Array<PC.VariableDeclarator>();
        for (const decls of this.program.declarations) {
            if (decls instanceof PC.VariableDeclarations) {
                allDecls.push(...decls.declarators);
            }
        }

        // Initialise all declarators
        const declsRequiringFurtherInitialisation = new Array<PC.VariableDeclarator>();
        for (const decl of allDecls) {
            const v = this.getVariableWithName(decl.identifier);
            if (!(v instanceof PCCGlobalVariable)) {
                throw new Error("Unexpected variable type or variable unknown.");
            }
            // Some types always require a custom initialisation (e.g., arrays). Primitive types require further initialisation only if there is a custom initialiser
            // const declRequiresFurtherInitialisation = PCCVariableInitialisation.typeRequiresCustomDefaultContainer(this, v.type) || decl.initializer != null;
            // if (declRequiresFurtherInitialisation) {
            //     const newContainer = PCCVariableInitialisation.compileDefaultValue(this, v);
            //     v.setContainer(this, newContainer);
            // }
            if (PCCVariableInitialisation.typeRequiresCustomDefaultContainer(this, v.type) || decl.initializer != null) {
                declsRequiringFurtherInitialisation.push(decl);
            }
        }

        if (declsRequiringFurtherInitialisation.length > 0) {
            // The initialisation goes into its own process definition(s)
            this.beginProcessGroup(new PCCSimpleGroupable("GlobalVariablesInitialisations"), []);

            // Compile initialisations
            declsRequiringFurtherInitialisation.forEach((decl) => {
                const v = this.getVariableWithName(decl.identifier)!;
                const newContainer = PCCVariableInitialisation.compileDefaultValue(this, v);
                v.setContainer(this, newContainer);
            });

            this.emitExit(); // initialisation will be executed sequentially before the main agent
            this.endProcessGroup();

            return true;
        }
        
        return false;
    }

    compileSimpleMutex() {
        const i = new PCCVariableContainer("i", CCSType.INT);
        this.beginProcessDefinition("Mutex", [i]);
		this.emitInput("lock", i, null);
		this.emitInput("unlock", i, null);
		this.emitProcessApplication("Mutex", [i]);
		this.endProcessDefinition();
		this.compileMutexCons(false);
    }

    compileReentrantMutex() {
        const i = new PCCVariableContainer("i", CCSType.INT);
		const c = new PCCVariableContainer("c", CCSType.INT);
		const a = new PCCVariableContainer("a", CCSType.INT);
		const a2 = new PCCVariableContainer("a2", CCSType.INT);
		
        this.beginProcessDefinition("Mutex", [i,c,a]);
		let control = this.emitChoice();
		this.emitCondition(new PCCBinaryContainer(c, new PCCConstantContainer(0), "=="));
		this.emitInput("lock", i, a);
		this.emitProcessApplication("Mutex", [i, new PCCConstantContainer(1), a]);
		control.setBranchFinished();
		this.emitCondition(new PCCBinaryContainer(c, new PCCConstantContainer(0), ">"));
		control = this.emitChoice();
		this.emitMatch("lock", i, a);
		this.emitProcessApplication("Mutex", [i, new PCCBinaryContainer(c, new PCCConstantContainer(1), "+"), a]);
		control.setBranchFinished();

		control = this.emitChoice();
		const r = new PCCVariableContainer("r", CCSType.INT);
		this.emitInput("multilock", i, r);
		this.emitProcessApplication("Mutex", [i, new PCCBinaryContainer(c, r, "+"), a]);	// add r reentrances to current counter c
		control.setBranchFinished();
		control = this.emitChoice();
		this.emitOutput("fullunlock", i, c);
		this.emitProcessApplication("Mutex", [i, new PCCConstantContainer(0), a]);
		control.setBranchFinished();


		this.emitInput("unlock", i, a2);
		control = this.emitChoice();
		this.emitCondition(new PCCBinaryContainer(a, a2, "=="));
		this.emitProcessApplication("Mutex", [i, new PCCBinaryContainer(c, new PCCConstantContainer(1), "-"), a]);
		control.setBranchFinished();
		this.emitCondition(new PCCBinaryContainer(a, a2, "!="));
		this.compileThrowException("Exception: Agent tried to unlock a mutex he did not lock!");
		this.endProcessDefinition();
		this.compileMutexCons(true);
    }

    compileMutexCons(reentrantMutex: boolean) {
        const i = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition("Mutex_cons", [i]);
		this.emitOutput("mutex_create", null, i);
		const control = this.emitParallel();
		this.emitProcessApplication("Mutex_cons", [new PCCBinaryContainer(i, new PCCConstantContainer(1), "+")]);
		control.setBranchFinished();
		if (reentrantMutex) {
			this.emitProcessApplication("Mutex", [i, new PCCConstantContainer(0), new PCCConstantContainer(0)]);
        } else {
			this.emitProcessApplication("Mutex", [i]);
        }
		control.setBranchFinished();
		this.endProcessDefinition();
		this.emitSystemProcessApplication("Mutex_cons", [new PCCConstantContainer(1)], PCCCompiler.SysMutexCons);
    }

    compileWaitRoom() {
        const i1 = new PCCVariableContainer("i", CCSType.INT);
		const c1 = new PCCVariableContainer("c", CCSType.INT);
		this.beginProcessDefinition("WaitRoom", [i1, c1]);
		const control1 = this.emitChoice();
		this.emitInput("signal", i1, null);
		const inner = this.emitChoice();
		this.emitCondition(new PCCBinaryContainer(c1, new PCCConstantContainer(0), "=="));
		this.emitProcessApplication("WaitRoom", [i1, c1]);
		inner.setBranchFinished();
		this.emitCondition(new PCCBinaryContainer(c1, new PCCConstantContainer(0), ">"));
		this.emitInput("wait", i1, null);
		this.emitProcessApplication("WaitRoom", [i1, new PCCBinaryContainer(c1, new PCCConstantContainer(1), "-")]);
		inner.setBranchFinished();
		control1.setBranchFinished();
		const control2 = this.emitChoice();
		this.emitInput("add", i1, null);
		this.emitProcessApplication("WaitRoom", [i1, new PCCBinaryContainer(c1, new PCCConstantContainer(1), "+")]);
		control2.setBranchFinished();
		const control3 = this.emitSequence();
		this.emitInput("signal_all", i1, null);
		this.emitProcessApplication("WaitDistributor", [i1, c1]);
		control3.setBranchFinished();
		this.emitProcessApplication("WaitRoom", [i1, new PCCConstantContainer(0)]);
		control3.setBranchFinished();
		control2.setBranchFinished();
		control1.setBranchFinished();
		this.endProcessDefinition();
		
		const i2 = new PCCVariableContainer("i", CCSType.INT);
		const c2 = new PCCVariableContainer("c", CCSType.INT);
		this.beginProcessDefinition("WaitDistributor", [i2, c2]);
		const controlA = this.emitChoice();
		this.emitCondition(new PCCBinaryContainer(c2, new PCCConstantContainer(0), "<="));
		this.emitExit();
		controlA.setBranchFinished();
		this.emitCondition(new PCCBinaryContainer(c2, new PCCConstantContainer(0), ">"));
		this.emitInput("wait", i2, null);
		this.emitProcessApplication("WaitDistributor", [i2, new PCCBinaryContainer(c2, new PCCConstantContainer(1), "-")]);
		controlA.setBranchFinished();
		this.endProcessDefinition();
		
		const i3 = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition("WaitRoom_cons", [i3]);
		this.emitOutput("wait_create", null, i3);
		const controlB = this.emitParallel();
		this.emitProcessApplication("WaitRoom_cons", [new PCCBinaryContainer(i3, new PCCConstantContainer(1), "+")]);
		controlB.setBranchFinished();
		this.emitProcessApplication("WaitRoom", [i3, new PCCConstantContainer(0)]);
		controlB.setBranchFinished();
		this.endProcessDefinition();
		this.emitSystemProcessApplication("WaitRoom_cons", [new PCCConstantContainer(1)], PCCCompiler.SysWaitRoom);
    }

    compileThrowException(...comps: Array<string | number | boolean | PCCContainer>) {
        const containers = comps.map((comp) => {
            if (comp instanceof PCCContainer) {
                return comp;
            } else {
                return new PCCConstantContainer(comp);
            }
        });

        let mainContainer: PCCContainer = containers.shift() ?? new PCCConstantContainer("Exception");
        
        while(containers.length > 0) {
            mainContainer = new PCCBinaryContainer(mainContainer, containers.shift()!, "^");
        }
        this.emitOutput("exception", null, mainContainer);
        this.emitStop();
    }

    compileArrayWithCapacity(size: number) {
        const i1 = new PCCVariableContainer("i", CCSType.INT);
		const args1 = [i1];
        for (let j = 0; j < size; j++) {
            args1.push(new PCCVariableContainer(`v${j}`, CCSType.INT));
        }
		this.beginProcessDefinition(`Array${size}`, args1);
		const index = new PCCVariableContainer("index", CCSType.INT);
		this.emitInput("array_access", i1, index);
        for (let j = 0; j < size; j++) {
            const control = this.emitChoice();
            this.emitCondition(new PCCBinaryContainer(index, new PCCConstantContainer(j), "=="));
			const inner = this.emitChoice();
			this.emitOutput("array_get", i1, args1[j+1]);
			this.emitProcessApplication(`Array${size}`, args1);
			inner.setBranchFinished();
			this.emitInput("array_set", i1, args1[j+1]);
			this.emitProcessApplication(`Array${size}`, args1);
			inner.setBranchFinished();
            control.setBranchFinished();
        }
		
		this.emitCondition(new PCCBinaryContainer(new PCCBinaryContainer(index, new PCCConstantContainer(size), ">="), new PCCBinaryContainer(index, new PCCConstantContainer(0), "<"), "||"));  // index >= size || index < 0
		this.compileThrowException("Index ", index, ` is out of array bounds ([0..${size-1}])!`);
		this.endProcessDefinition();
		
		const i2 = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition(`Array${size}_cons`, []);
		this.emitInput(`array${size}_prepare`, null, null);
		this.emitInput("array_new", null, i2);
		const def = new PCCVariableContainer("d", CCSType.VOID);
		this.emitOutput(`array${size}_create`, null, i2);
		this.emitInput("array_setDefault", i2, def);
		const control = this.emitParallel();
		this.emitProcessApplication(`Array${size}_cons`, []);
		control.setBranchFinished();
		const args2 = [i2];
        for (let j = 0; j < size; j++) {
            args2.push(def);
        }
		this.emitProcessApplication(`Array${size}`, args2);
		control.setBranchFinished();
		this.endProcessDefinition();
		this.emitSystemProcessApplication(`Array${size}_cons`, [], PCCCompiler.SysArray);
    }

    compileArrayManager() {
        const i = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition("ArrayManager", [i]);
		this.emitOutput("array_new", null, i);
		this.emitProcessApplication("ArrayManager", [new PCCBinaryContainer(i, new PCCConstantContainer(1), "+")]);
		this.endProcessDefinition();
		this.emitSystemProcessApplication("ArrayManager", [new PCCConstantContainer(1)], PCCCompiler.SysInstanceManager);
    }

    compileChannelWithCapacity(capacity: number) {
        if (capacity <= 0) {
            this.compileUnbufferedChannelCons();
            return;
        }
        const i1 = new PCCVariableContainer("i", CCSType.INT);
		const c = new PCCVariableContainer("c", CCSType.INT);
		const args: Array<PCCVariableContainer> = [i1, c];
        for (let j = 0; j < capacity; j++) {
            args.push(new PCCVariableContainer(`v${j}`, CCSType.INT));
        }
		this.beginProcessDefinition(`Channel${capacity}`, args);
        const params: Array<PCCContainer> = args; // weaken type for parameters
		params[1] = new PCCBinaryContainer(c, new PCCConstantContainer(1), "+");
		for (let j = 0; j < capacity; j++) {
            const control = this.emitChoice();
			this.emitCondition(new PCCBinaryContainer(c, new PCCConstantContainer(j), "=="));
			const v = new PCCVariableContainer(`v${j}`, CCSType.INT);
			this.emitInput("put", i1, v);
			this.emitProcessApplication(`Channel${capacity}`, params);
			control.setBranchFinished();
        }
		this.emitCondition(new PCCBinaryContainer(c, new PCCConstantContainer(0), ">"));
		this.emitOutput("receive", i1, new PCCVariableContainer("v0", CCSType.INT));
		params.splice(2, 1);
		params.push(new PCCConstantContainer(0));
		params[1] = new PCCBinaryContainer(c, new PCCConstantContainer(1), "-");
		this.emitProcessApplication(`Channel${capacity}`, params);
		this.endProcessDefinition();
		
		const i2 = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition(`Channel${capacity}_cons`, []);
		this.emitInput(`channel${capacity}_prepare`, null, null);
		this.emitInput("channel_new", null, i2);
		//def = new PCCVariableContainer("d", PCCType.VOID)
		this.emitOutput(`channel${capacity}_create`, null, i2);
		//this.emitInput("channel_setDefault", i, def)
		const control = this.emitParallel();
		this.emitProcessApplication(`Channel${capacity}_cons`, []);
		control.setBranchFinished();
		const params2 = [i2, new PCCConstantContainer(0)];
		const def = new PCCConstantContainer(0);
        for (let j = 0; j < capacity; j++) {
            params2.push(def);
        }
		this.emitProcessApplication(`Channel${capacity}`, params2);
		control.setBranchFinished();
		this.endProcessDefinition();
		this.emitSystemProcessApplication(`Channel${capacity}_cons`, [], PCCCompiler.SysChannel);
    }

    compileUnbufferedChannelCons() {
        const i = new PCCVariableContainer("i", CCSType.INT);
		this.beginProcessDefinition("Channel_cons", [i]);
		this.emitOutput("channel_create", null, i);
		this.emitProcessApplication("Channel_cons", [new PCCBinaryContainer(i, new PCCConstantContainer(1), "-")]);
		this.endProcessDefinition();
		this.emitSystemProcessApplication("Channel_cons", [new PCCConstantContainer(-1)], PCCCompiler.SysChannel);
    }

    compileChannelManager() {
        const i = new PCCVariableContainer("next_i", CCSType.INT)
		this.beginProcessDefinition("ChannelManager", [i])
		this.emitOutput("channel_new", null, i)
		this.emitProcessApplication("ChannelManager", [new PCCBinaryContainer(i, new PCCConstantContainer(1), "+")])
		this.endProcessDefinition()
		this.emitSystemProcessApplication("ChannelManager", [new PCCConstantContainer(1)], PCCCompiler.SysInstanceManager)
    }

    compileAgentTools(addAgentJoiner: boolean) {
        const i = new PCCVariableContainer("next_i", CCSType.INT);
		this.beginProcessDefinition("AgentManager", [i]);
		this.emitOutput("agent_new", null, i);
		this.emitProcessApplication("AgentManager", [new PCCBinaryContainer(i, new PCCConstantContainer(1), "+")]);
		this.endProcessDefinition();
		this.emitSystemProcessApplication("AgentManager", [new PCCConstantContainer(2)], PCCCompiler.SysInstanceManager);
		
        if (!addAgentJoiner) {
            return;
        }

		const a1 = new PCCVariableContainer("a", CCSType.INT);
		const c1 = new PCCVariableContainer("c", CCSType.INT);
		this.beginProcessDefinition("AgentJoiner", [a1, c1]);
		const control1 = this.emitChoice();
		this.emitInput("join_register", a1, null);
		this.emitProcessApplication("AgentJoiner", [a1, new PCCBinaryContainer(c1, new PCCConstantContainer(1), "+")]);
		control1.setBranchFinished();
		this.emitInput("agent_terminate", a1, null);
		this.emitProcessApplication("JoinDistributor", [a1, c1]);
		control1.setBranchFinished();
		this.endProcessDefinition();
		
		const a2 = new PCCVariableContainer("a", CCSType.INT);
		const c2 = new PCCVariableContainer("c", CCSType.INT);
		this.beginProcessDefinition("JoinDistributor", [a2, c2]);
		const control2 = this.emitChoice();
		this.emitCondition(new PCCBinaryContainer(c2, new PCCConstantContainer(0), "<="));
		this.emitInput("join_register", a2, null);
		this.emitInput("join", a2, null);
		this.emitProcessApplication("JoinDistributor", [a2, new PCCConstantContainer(0)]);
		control2.setBranchFinished();
		this.emitCondition(new PCCBinaryContainer(c2, new PCCConstantContainer(0), ">"));
		this.emitInput("join", a2, null);
		this.emitProcessApplication("JoinDistributor", [a2, new PCCBinaryContainer(c2, new PCCConstantContainer(1), "-")]);
		control2.setBranchFinished();
		this.endProcessDefinition();
    }

    compileReturn() {
        this.beginProcessDefinition("Return", []);
		const v = new PCCVariableContainer("v", CCSType.INT);
		this.emitInput("return", null, v);
		this.emitOutput("rreturn", null, v);
		this.emitProcessApplication("Return", []);
		this.endProcessDefinition();
		this.emitSystemProcessApplication("Return", [], PCCCompiler.SysUnknownWeight); // ToDo: Should this be SysUnknownWeight?
    }


    



    static SysAgent = 50;
    static SysEnvironment = 150
    static SysChannel = 250
    static SysArray = 450
    static SysMutexCons = 550
    static SysWaitRoom = 650
    static SysInstanceManager = 750
    static SysUnknownWeight = 5000
}



// PCCConstructor was only used by PCClass and the code has essentially moved there

// export interface PCCConstructorDelegate {
//     constructorGetName(constructor: PCCConstructor, context: any): string;
//     constructorGetArguments(constructor: PCCConstructor, compiler: PCCCompiler, context: any): Array<PCCVariableInfo>;
//     constructorHasEnvironment(constructor: PCCConstructor, context: any): boolean;
//     constructorProtectEnvironmentArguments(constructor: PCCConstructor, compiler: PCCCompiler, variables: Array<PCCVariable>, context: any): number;
// }

// export class PCCConstructor {
//     constructor(readonly compiler: PCCCompiler, readonly delegate: PCCConstructorDelegate, readonly context: any) { }

//     emit() {
//         const envName = this.delegate.constructorGetName(this, this.context)
// 		const variables = this.delegate.constructorGetArguments(this, this.compiler, this.context)
// 		const hasEnvironment = this.delegate.constructorHasEnvironment(this, this.context);
// 		this.compiler.beginProcessGroup(new PCCSimpleGroupable(envName+"_cons"), variables)
// 		const entry = this.compiler.getProcessFrame()
// 		const localVariables = variables.map((v) => this.compiler.getVariableWithNameOfClass(v.getName(), null, v.isInternal))
// 		const envArgCount = this.delegate.constructorProtectEnvironmentArguments(this, this.compiler, localVariables, this.context)
// 		const vars = new Array<PCCContainer>();
//         for (let i = 0; i < envArgCount; i++) {
//             vars.unshift(this.compiler.unprotectContainer());
//         }
// 		this.compiler.emitOutput(`class_${this.delegate.getName()}_initialized`, vars[0], null)  // vars[0] is the instance number of the object being created
// 		const recursion = this.delegate.constructorShouldCallRecursively?(this, this.context);
// 		const control = (recursion && hasEnvironment) ? this.compiler.emitParallel() : null;
// 		if(hasEnvironment) {
// 			this.compiler.emitProcessApplication(envName, vars);
//         }
// 		if(recursion) {
//             control?.setBranchFinished()
// 			this.delegate.constructorUpdateVariablesForRecursiveCall(this, this.compiler, entry, variables, this.context);
// 			entry.emitCallProcessFromFrame(this.compiler, this.compiler.getProcessFrame())
// 			control?.setBranchFinished()
//         }
// 		if (!hasEnvironment && !recursion) {
// 			this.compiler.emitStop()
//         }
// 		this.compiler.endProcessGroup()
//     }
// }



