import * as CCS from "@pseuco/ccs-interpreter";
import * as PC from "@pseuco/lang";

import { PCCCompiler } from "./PCCCompiler.js";
import { PCCProcessFrame } from "./PCCProcessFrame.js";
import { PCCClass, PCCGlobal, PCCProcedure, PCCVariable, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCContainer, PCCVariableContainer } from "./PCCContainer.js";

type CCSNode = CCS.CCSProcess | CCS.CCSProcessDefinition | CCS.CCSAction | CCS.CCSExpression;

export class PCCStackResult {
    constructor(readonly type: PCCStackResultType, readonly data: CCS.CCSProcess | CCS.CCSProcessDefinition) {
        if (this.type == PCCStackResultType.CCSPROCESS) {
            if (!(this.data instanceof CCS.CCSProcess)) {
                throw new Error("Type mismatch!");
            }
        } else if (this.type == PCCStackResultType.CCSPROCESS_DEFINITION || this.type == PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE) {
            if (!(this.data instanceof CCS.CCSProcessDefinition)) {
                throw new Error("Type mismatch!");
            }
        } else {
            throw new Error("Unknown result type!");
        }
    }

    get ccsProcess(): CCS.CCSProcess {
        if (this.type == PCCStackResultType.CCSPROCESS) {
            return this.data as CCS.CCSProcess;
        } else {
            throw new Error("Result is not a CCS process.");
        }
    }

    get ccsProcessDefinition(): CCS.CCSProcessDefinition {
        if (this.type == PCCStackResultType.CCSPROCESS_DEFINITION || this.type == PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE) {
            return this.data as CCS.CCSProcessDefinition;
        } else {
            throw new Error("Result is not a CCS process definition.");
        }
    }
}

export class PCCStackResultContainer {
    private results: Array<PCCStackResult> = new Array();
    // ToDo: Do we really need an array here? Can't we just drop whatever we would at index >= 1?

    get allResults(): ReadonlyArray<PCCStackResult> {
        return this.results;
    }

    addResult(type: PCCStackResultType, data: CCS.CCSProcess | CCS.CCSProcessDefinition) {
        this.results.unshift(new PCCStackResult(type, data));
    }

    replaceResult(type: PCCStackResultType, data: CCS.CCSProcess | CCS.CCSProcessDefinition) {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        this.results[0] = new PCCStackResult(type, data);
    }

    getResult(): PCCStackResult {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        return this.results[0];
    }

    removeResult() {
        if (this.results.length == 0) {
            throw new Error("No results available in this container.");
        }
        this.results.shift();
    }

    appendContainer(container: PCCStackResultContainer) {
        this.results = this.results.concat(container.results);
    }
}

export type PCCControlElement = PCCGlobal | PCCProcedure | PCCClass;

export abstract class PCCStackElement {
    parent: PCCStackElement | null = null;
    pseucoNode: PC.Node | null = null;
    private myStack: PCCCompilerStack | null = null;
    get stack(): PCCCompilerStack {
        if (!this.myStack) {
            throw new Error("Stack has not been set.");
        }
        return this.myStack;
    }
    set stack(v: PCCCompilerStack) {
        if (this.myStack && v != this.myStack) {
            throw new Error("Stack can be set only once!");
        }
        this.myStack = v;
    }
    // get parentStackElement(): PCCStackElement | null {
    //     return this.parent;
    // }
    // set parentStackElement(v: PCCStackElement) {
    //     if (this.parent != null) {
    //         throw new Error("parentStackElement can be set only once.");
    //     }
    //     this.parent = v;
    // }

    abstract getResults(): PCCStackResultContainer;
    abstract getNext(): PCCStackElement | null;
    abstract setNext(next: PCCStackElement): void;
    abstract removeNext(next: PCCStackElement): void;

    getTopElement(): PCCStackElement {
        const next = this.getNext();
        if (next == null) {
            return this;
        } else {
            return next.getTopElement();
        }
    }

    getCurrentProcessFrame(): PCCProcessFrame | null {
        if (this.parent == null) {
            return null;
        } else {
            return this.parent.getCurrentProcessFrame();
        }
    }

    getCurrentControlElement(): PCCControlElement | null {
        if (this.parent == null) {
            return null;
        } else {
            return this.parent.getCurrentControlElement();
        }
    }

    removeFromStack(): PCCStackResultContainer {
        this.stack.topElement = this.parent;
        if (this.parent != null) {
            this.parent.removeNext(this);
        }
        return this.getResults();
    }

    isCompletedProcess(): boolean {
        // ! This is a correction of the old implementation
        return this.getNext()?.isCompletedProcess() ?? false;
    }

    // createCalculusNode ???

    rememberPseucoCalculusRelationship(ccsNode: CCSNode) {
        // Todo
    }


    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null {
        if (this.parent != null) {
            return this.parent.compilerGetVariable(compiler, identifier);
        } else {
            return null;
        }
    }

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null {
        if (this.parent != null) {
            return this.parent.compilerGetProcedure(compiler, identifier);
        } else {
            return null;
        }
    }

    // Returns true if the new variable could be handled.
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo) {
        if (this.parent != null) {
            this.parent.compilerHandleNewVariable(compiler, variable);
        } else {
            //return null;
        }
    }
}



export class PCCUnaryStackElement extends PCCStackElement {
    private next: PCCStackElement | null = null;

    override getResults(): PCCStackResultContainer {
        if (this.next == null) {
            return new PCCStackResultContainer();
        } else {
            return this.next.getResults();
        }
    }

    override getNext(): PCCStackElement | null {
        return this.next;
    }

    override setNext(next: PCCStackElement) {
        if (this.next != null) {
            throw new Error("Can't set next twice!");
        }
        next.parent = this;
        this.next = next;
    }

    override removeNext(next: PCCStackElement) {
        if (this.next != next) {
            throw new Error("Unknown next");
        }
        next.parent = null;
        this.next = null;
    }
}


export enum PCCBinaryTarget {
    LEFT = 0,
    RIGHT = 1,
    TOP = 2
}

export abstract class PCCBinaryStackElement extends PCCStackElement {
    protected leftStack: PCCStackElement | null = null;
    protected rightStack: PCCStackElement | null = null;
    protected topStack: PCCStackElement | null = null;
    private target = PCCBinaryTarget.LEFT;

    override getNext(): PCCStackElement | null {
        if (this.target == PCCBinaryTarget.LEFT) {
            return this.leftStack;
        } else if (this.target == PCCBinaryTarget.RIGHT) {
            return this.rightStack;
        } else {
            return this.topStack;
        }
    }

    private setOrRemoveNext(next: PCCStackElement | null) {
        if (next != null) {
            next.parent = this;
        }
        
        if (this.target == PCCBinaryTarget.LEFT) {
            if (this.leftStack != null) {
                throw new Error("Can't set next twice!")
            }
            this.leftStack = next;
        } else if (this.target == PCCBinaryTarget.RIGHT) {
            if (this.rightStack != null) {
                throw new Error("Can't set next twice!")
            }
            this.rightStack = next;
        } else {
            if (this.topStack != null) {
                throw new Error("Can't set next twice!")
            }
            this.topStack = next;
        }
    }

    override setNext(next: PCCStackElement) {
        this.setOrRemoveNext(next);
    }

    override removeNext(next: PCCStackElement) {
        if (next != this.getNext()) {
            throw new Error("Unknown next");
        }
        this.setOrRemoveNext(null);
    }


    setTarget(target: PCCBinaryTarget) {
        if (target < 0 || target > 2) {
            throw new Error("Illegal target!");
        }
        if (target != PCCBinaryTarget.TOP && this.topStack != null) {
            throw new Error("Left and right target are not allowed once top was modified!");
        }
        this.target = target;
        this.stack.topElement = this.getTopElement();
        // ToDo: Unclear why the following should be necessary:
        // let mom = this.parent;
        // while (mom != null) {
        //     if (mom instanceof PCCBinaryStackElement) {
        //         mom.updateBinaryTargets(this);
        //         break;
        //     } else {
        //         mom = mom.parent;
        //     }
        // }
    }

    setBranchFinished() {
        if (this.target == PCCBinaryTarget.LEFT) {
            this.setTarget(PCCBinaryTarget.RIGHT);
        } else if (this.target == PCCBinaryTarget.RIGHT) {
            this.setTarget(PCCBinaryTarget.TOP);
        } else {
            throw new Error("All branches are already finished!");
        }
    }

    updateBinaryTargets(destination: PCCStackElement) {
        if (destination == this.leftStack) {
            this.target = PCCBinaryTarget.LEFT;
        } else if (destination == this.rightStack) {
            this.target = PCCBinaryTarget.RIGHT;
        } else if (destination == this.topStack) {
            this.target = PCCBinaryTarget.TOP;
        } else {
            throw new Error("Unknown destination!");
        }
    }

    override isCompletedProcess(): boolean {
        // ! This override was missing in old implementation
        if (this.leftStack != null && this.rightStack != null && this.topStack != null) {
            return this.leftStack.isCompletedProcess() && this.rightStack.isCompletedProcess() && this.topStack.isCompletedProcess();
        } else {
            return false;
        }
    }
}



export enum PCCStackResultType {
    UNSPECIFIC = 0,
    CCSPROCESS = 1,
    CCSPROCESS_DEFINITION = 2,
    CCSPROCESS_DEFINITION_FOR_PROCEDURE = 3
}

export class PCCStopStackElement extends PCCUnaryStackElement {
    override getResults(): PCCStackResultContainer {
        const more = (this.getNext() == null) ? new PCCStackResultContainer() : this.getNext()!.getResults();
        const stopProcess = new CCS.CCSStop();
        this.rememberPseucoCalculusRelationship(stopProcess);
        more.addResult(PCCStackResultType.CCSPROCESS, stopProcess);
        return more;
    }

    override isCompletedProcess(): boolean {
        return true;
    }
}

export class PCCExitStackElement extends PCCUnaryStackElement {
    override getResults(): PCCStackResultContainer {
        const more = (this.getNext() == null) ? new PCCStackResultContainer() : this.getNext()!.getResults();
        const exitProcess = new CCS.CCSExit();
        this.rememberPseucoCalculusRelationship(exitProcess);
        more.addResult(PCCStackResultType.CCSPROCESS, exitProcess);
        return more;
    }

    override isCompletedProcess(): boolean {
        return true;
    }
}

export class PCCPrefixStackElement extends PCCUnaryStackElement {
    // chan(i). 0 --> `chan` is channel, `i` is channel parameter
    constructor(readonly channel: string, readonly channelParameter: PCCContainer | null = null) {
        super();
    }

    protected getChannel(): CCS.CCSChannel {
        return new CCS.CCSChannel(this.channel, this.channelParameter?.ccsTree() ?? null);
    }

    protected getAction(): CCS.CCSAction {
        const res = new CCS.CCSSimpleAction(this.getChannel());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }

    override getResults(): PCCStackResultContainer {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext()!.getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const prefixProcess = new CCS.CCSPrefix(this.getAction(), pRes.ccsProcess);
        this.rememberPseucoCalculusRelationship(prefixProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, prefixProcess);
        return container;
    }
}


export class PCCInputStackElement extends PCCPrefixStackElement {
    constructor(channel: string, channelParameter: PCCContainer | null = null, readonly value: PCCVariableContainer | null = null) {
        super(channel, channelParameter);
    }

    protected override getAction(): CCS.CCSAction {
        let variable: string | null = null;
        if (this.value != null) {
            variable = this.value.identifier;
            // this.rememberPseucoCalculusRelationship(variable);
        }
        const res = new CCS.CCSInput(this.getChannel(), variable, new CCS.CCSUnrestrictedValueSet());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    } 
    
}

export class PCCOutputStackElement extends PCCPrefixStackElement {
    constructor(channel: string, channelParameter: PCCContainer | null = null, readonly value: PCCContainer | null = null) {
        super(channel, channelParameter);
    }

    protected override getAction(): CCS.CCSAction {
        const res = new CCS.CCSOutput(this.getChannel(), (this.value == null) ? null : this.value.ccsTree());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}

export class PCCMatchStackElement extends PCCPrefixStackElement {
    constructor(channel: string, channelParameter: PCCContainer | null, readonly value: PCCContainer) {
        super(channel, channelParameter);
    }

    protected override getAction(): CCS.CCSAction {
        const res = new CCS.CCSMatch(this.getChannel(), this.value.ccsTree());
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}


export class PCCConditionStackElement extends PCCUnaryStackElement {
    constructor(readonly conditionContainer: PCCContainer) {
        super();
    }

    override getResults(): PCCStackResultContainer {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext()!.getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const conditionProcess = new CCS.CCSCondition(this.conditionContainer.ccsTree(), pRes.ccsProcess);
        this.rememberPseucoCalculusRelationship(conditionProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, conditionProcess);
        return container;
    }
}


export class PCCRestrictionStackElement extends PCCUnaryStackElement {
    constructor(readonly restrictedChannels: ReadonlyArray<string>) {
        super();
    }

    override getResults(): PCCStackResultContainer {
        if (this.getNext() == null) {
            throw new Error("Expected `next` element.");
        }
        const container = this.getNext()!.getResults();
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }
        const restrictionProcess = new CCS.CCSRestriction(pRes.ccsProcess, this.restrictedChannels.slice());
        this.rememberPseucoCalculusRelationship(restrictionProcess);
        container.replaceResult(PCCStackResultType.CCSPROCESS, restrictionProcess);
        return container;
    }
}


export class PCCApplicationStackElement extends PCCUnaryStackElement {
    readonly argContainers: ReadonlyArray<PCCContainer>;
    constructor(readonly processName: string, argContainers: ReadonlyArray<PCCContainer> = []) {
        super();
        this.argContainers = argContainers.slice();
    }

    override getResults(): PCCStackResultContainer {
        return PCCApplicationStackElement.computeResults(this);
    }

    static computeResults(stackElement: PCCApplicationStackElement | PCCApplicationPlaceholderStackElement) {
        const more = stackElement.getNext()?.getResults() ?? new PCCStackResultContainer();
        const argumentValues = stackElement.argContainers!.map((c) => c.ccsTree());
        const process = new CCS.CCSProcessApplication(stackElement.processName!, argumentValues);
        stackElement.rememberPseucoCalculusRelationship(process);
        more.addResult(PCCStackResultType.CCSPROCESS, process);
        return more;
    }

    override isCompletedProcess(): boolean {
        return true;
    }
}


export class PCCApplicationPlaceholderStackElement extends PCCUnaryStackElement {
    constructor(readonly frame: PCCProcessFrame) {
        super();
    }
    
    private appStackElement: PCCApplicationStackElement | null = null;
    get applicationStackElement(): PCCApplicationStackElement | null {
        return this.appStackElement;
    }

    set(processName: string, argContainers: ReadonlyArray<PCCContainer> = []) {
        if (this.appStackElement != null) {
            throw new Error("Placeholder values can't be set twice!");
        }
        this.appStackElement = new PCCApplicationStackElement(processName, argContainers);
    }

    get processName(): string | undefined {
        return this.appStackElement?.processName;
    }

    get argContainers(): readonly PCCContainer[] | undefined {
        return this.appStackElement?.argContainers;
    }

    override getResults(): PCCStackResultContainer {
        if (this.appStackElement == null) {
            throw new Error("Placeholder without values!");
        } else {
            return PCCApplicationStackElement.computeResults(this);
        }
    }

    override isCompletedProcess(): boolean {
        return true;
    }
}


export abstract class PCCBinaryCCSStackElement extends PCCBinaryStackElement {

    protected abstract createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess): CCS.CCSProcess;

    override getResults(): PCCStackResultContainer {
        const more = this.topStack?.getResults() ?? new PCCStackResultContainer();
        const leftContainer = this.leftStack?.getResults();
        if (!leftContainer) {
            throw new Error("Left process not available.");
        }
        const left = leftContainer.getResult();
        // if (left.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }

        const rightContainer = this.rightStack?.getResults();
        if (!rightContainer) {
            throw new Error("Right process not available.");
        }
        const right = rightContainer.getResult();
        // if (right.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }

        rightContainer.removeResult();
		leftContainer.appendContainer(rightContainer); // unclear why this happens...
		leftContainer.appendContainer(more);

        leftContainer.replaceResult(PCCStackResultType.CCSPROCESS, this.createCCSProcess(left.ccsProcess, right.ccsProcess));

        return leftContainer;
    }

    override isCompletedProcess(): boolean {
        if (this.leftStack != null && this.rightStack != null) {
            return this.leftStack.isCompletedProcess() && this.rightStack.isCompletedProcess();
        } else {
            return false;
        }
    }
}


export class PCCChoiceStackElement extends PCCBinaryCCSStackElement {
    protected override createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess) {
        const res = new CCS.CCSChoice(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}

export class PCCParallelStackElement extends PCCBinaryCCSStackElement {
    protected override createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess) {
        const res = new CCS.CCSParallel(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}

export class PCCSequenceStackElement extends PCCBinaryCCSStackElement {
    protected override createCCSProcess(left: CCS.CCSProcess, right: CCS.CCSProcess) {
        const res = new CCS.CCSSequence(left, right);
        this.rememberPseucoCalculusRelationship(res);
        return res;
    }
}



export class PCCSystemProcessStackElement extends PCCUnaryStackElement {
    constructor(readonly weight: number) {
        super();
    }

    override getResults(): PCCStackResultContainer {
        const container = this.getNext()?.getResults();
        if (!container) {
            throw new Error("Expected a `next` element.");
        }
        const pRes = container.getResult();
        if (pRes.type != PCCStackResultType.CCSPROCESS) {
            throw new Error("Unexpected result type!");
        }
        // return pRes;
        return container;
    }
}

export class PCCProcessDefinitionStackElement extends PCCUnaryStackElement {
    readonly argContainers: Array<PCCVariableContainer>;
    private isProcedure: boolean = false;
    get isMarkedAsProcedure(): boolean {
        return this.isProcedure;
    }
    constructor(readonly processName: string, argContainers: Array<PCCVariableContainer>) {
        super();
        this.argContainers = argContainers.slice();
    }

    markAsProcedure() {
        this.isProcedure = true;
    }

    override getResults(): PCCStackResultContainer {
        const container = this.getNext()?.getResults();
        if (!container) {
            throw new Error("Expected a `next` element.");
        }
        const pRes = container.getResult();
        // if (pRes.type != PCCStackResultType.CCSPROCESS) {
        //     throw new Error("Unexpected result type!");
        // }

        // const argCCSVariables = this.argContainers.map((c) => {
        //     const ccsNode = new CCS.Variable(c.identifier);
        //     this.rememberPseucoCalculusRelationship(ccsNode);
        //     return ccsNode;
        // });
        const argCCSVariables = this.argContainers.map((c) => c.identifier);
        const ccsDef = new CCS.CCSProcessDefinition(this.processName, pRes.ccsProcess, argCCSVariables);
        const resultType = (this.isMarkedAsProcedure) ? PCCStackResultType.CCSPROCESS_DEFINITION_FOR_PROCEDURE : PCCStackResultType.CCSPROCESS_DEFINITION;
        container.replaceResult(resultType, ccsDef);
        return container;
    }
}

export class PCCProcessFrameStackElement extends PCCUnaryStackElement {
    constructor(readonly frame: PCCProcessFrame) {
        super();
    }
    override getCurrentProcessFrame(): PCCProcessFrame | null {
        return this.frame;
    }
    override compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null {
        return this.frame.compilerGetVariable(compiler, identifier) ?? super.compilerGetVariable(compiler, identifier);
    }
    override compilerGetProcedure(compiler: PCCCompiler, identifier: string) {
        return this.frame.compilerGetProcedure(compiler, identifier) ?? super.compilerGetProcedure(compiler, identifier);
    }
    override compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo) {
        this.frame.compilerHandleNewVariable(compiler, variable);
    }
    override isCompletedProcess(): boolean {
        return this.getNext()?.isCompletedProcess() ?? false;
    }
}

export class PCCClassStackElement extends PCCUnaryStackElement {
    constructor(readonly classInfo: PCCClass) {
        super();
    }

    override getCurrentControlElement(): PCCClass {
        return this.classInfo;
    }
    override compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null {
        return this.classInfo.compilerGetVariable(compiler, identifier) ?? super.compilerGetVariable(compiler, identifier);
    }
    override compilerGetProcedure(compiler: PCCCompiler, identifier: string) {
        return this.classInfo.compilerGetProcedure(compiler, identifier) ?? super.compilerGetProcedure(compiler, identifier);
    }
    override compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo) {
        this.classInfo.compilerHandleNewVariable(compiler, variable);
    }
}

export class PCCGlobalStackElement extends PCCUnaryStackElement {
    constructor(readonly global: PCCGlobal) {
        super();
    }
    override getCurrentControlElement(): PCCGlobal {
        return this.global;
    }
    override compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable {
        return this.global.compilerGetVariable(compiler, identifier); // ?? super.compilerGetVariable(compiler, identifier);
    }
    override compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure {
        return this.global.compilerGetProcedure(compiler, identifier);
    }
    override compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo) {
        this.global.compilerHandleNewVariable(compiler, variable);
    }
}

export class PCCProcedureStackElement extends PCCUnaryStackElement {
    constructor(readonly procedure: PCCProcedure) {
        super();
    }

    override getCurrentControlElement(): PCCProcedure {
        return this.procedure;
    }
}






export class PCCCompilerStack {
    topElement: PCCStackElement | null = null;

    getCurrentProcessFrame(): PCCProcessFrame | null {
        return this.topElement?.getCurrentProcessFrame() ?? null;
    }

    getCurrentControlElement(): PCCControlElement | null {
        return this.topElement?.getCurrentControlElement() ?? null;
    }

    pushElement(e: PCCStackElement) {
        e.stack = this;
        this.topElement?.setNext(e);
        this.topElement = e;
    }

    isCurrentProcessCompleted(): boolean {
        return this.topElement?.isCompletedProcess() ?? false;
    }


    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCVariable | null {
        return this.topElement?.compilerGetVariable(compiler, identifier) ?? null;
    }

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null {
        return this.topElement?.compilerGetProcedure(compiler, identifier) ?? null;
    }

    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): /*PCCVariable | null*/ void {
        // return this.topElement?.compilerHandleNewVariable(compiler, variable) ?? null;
        this.topElement?.compilerHandleNewVariable(compiler, variable);
    }
}

