import { CCSType } from "./PCCProgramController.js";
import * as CCS from "@pseuco/ccs-interpreter";

// Containers represent arithmetic expressions in CCS

export abstract class PCCContainer {
    constructor (readonly ccsType: CCSType) { }

    abstract isReadonly(): boolean;
    abstract ccsTree(): CCS.CCSExpression;
    abstract isEqual(container: PCCContainer): boolean;
}


export class PCCConstantContainer extends PCCContainer {
    constructor(readonly value: string | boolean | number) {
        let inferredCCSType: CCSType;
        if (typeof value == "boolean") {
            inferredCCSType = CCSType.BOOL;
        } else if (typeof value == "number") {
            inferredCCSType = CCSType.INT;
        } else if (typeof value == "string") {
            inferredCCSType = CCSType.STRING;
        } else {
            throw new Error("Unknown value type");
        }
        super(inferredCCSType);
    }

    isReadonly(): boolean {
        return true;
    }

    ccsTree(): CCS.CCSConstantExpression {
        return new CCS.CCSConstantExpression(this.value);
    }

    isEqual(container: PCCContainer): boolean {
        return container instanceof PCCConstantContainer && container.value === this.value;
    }
}


export class PCCVariableContainer extends PCCContainer {
    constructor(readonly identifier: string, ccsType: CCSType) {
        super(ccsType);
    }

    isReadonly(): boolean {
        return false;
    }

    ccsTree(): CCS.CCSVariableExpression  {
        return new CCS.CCSVariableExpression(this.identifier);
    }

    isEqual(container: PCCContainer): boolean {
        return container instanceof PCCVariableContainer && container.identifier == this.identifier;
    }
}

export type PCCBinaryContainerOperator = "==" | "!=" | "<" | ">" | "<=" | ">=" | "+" | "-" | "*" | "/" | "%" | "^" | "&&" | "||";
export class PCCBinaryContainer extends PCCContainer {
    private _ccsTree: CCS.CCSExpression;

    constructor(readonly leftContainer: PCCContainer, readonly rightContainer: PCCContainer, readonly operator: PCCBinaryContainerOperator) {
        let inferredCCSType: CCSType;
        let ccsTree: CCS.CCSExpression;
        if (operator == "+" || operator == "^") {
            if (leftContainer.ccsType.isString() || rightContainer.ccsType.isString() || operator == "^") {
                ccsTree = new CCS.CCSConcatenatingExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
                inferredCCSType = CCSType.STRING;
            } else {
                ccsTree = new CCS.CCSAdditiveExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator)
                inferredCCSType = CCSType.INT;
            }
        } else if (operator == "-") {
            ccsTree = new CCS.CCSAdditiveExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator)
            inferredCCSType = CCSType.INT;
        } else if (operator == "*" || operator == "/" || operator == "%") {
            ccsTree = new CCS.CCSMultiplicativeExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.INT;
        } else if (operator == "<" || operator == "<=" || operator == ">" || operator == ">=") {
            ccsTree = new CCS.CCSRelationalExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.BOOL;
        } else if (operator == "==" || operator == "!=") {
            ccsTree = new CCS.CCSEqualityExpression(leftContainer.ccsTree(), rightContainer.ccsTree(), operator);
            inferredCCSType = CCSType.BOOL;
        } else if (operator == "&&") {
            ccsTree = new CCS.CCSAndExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
            inferredCCSType = CCSType.BOOL;
        } else if (operator == "||") {
            ccsTree = new CCS.CCSOrExpression(leftContainer.ccsTree(), rightContainer.ccsTree());
            inferredCCSType = CCSType.BOOL;
        } else {
            throw Error("Unknown operator");
        }

        super(inferredCCSType);
        this._ccsTree = ccsTree;
    }

    isReadonly(): boolean {
        return true;
    }

    ccsTree(): CCS.CCSExpression {
        return this._ccsTree;
    }

    isEqual(container: PCCContainer): boolean {
        return container instanceof PCCBinaryContainer && container.operator == this.operator && container.leftContainer.isEqual(this.leftContainer) && container.rightContainer.isEqual(this.rightContainer);
    }
}


export class PCCUnaryContainer extends PCCContainer {
    private _ccsTree: CCS.CCSExpression;

    constructor(readonly operator: "!"|"-"|"+", readonly container: PCCContainer) {
        let inferredCCSType: CCSType;
        let ccsTree: CCS.CCSExpression;
        if (operator == "!") {
            ccsTree = new CCS.CCSComplementExpression(container.ccsTree());
            inferredCCSType = CCSType.BOOL;
        } else if (operator == "-") {
            ccsTree = new CCS.CCSAdditiveExpression(new CCS.CCSConstantExpression(0), container.ccsTree(), "-");
            inferredCCSType = CCSType.INT;
        } else if (operator == "+") {
            ccsTree = container.ccsTree();
            inferredCCSType = CCSType.INT;
        } else {
            throw Error("Unknown operator");
        }
        super(inferredCCSType);
        this._ccsTree = ccsTree;
    }

    isReadonly(): boolean {
        return true;
    }

    ccsTree(): CCS.CCSExpression {
        return this._ccsTree;
    }

    isEqual(container: PCCContainer): boolean {
        return container instanceof PCCUnaryContainer && container.operator == this.operator && container.container.isEqual(this.container); 
    }
}


