/*** 
 * Manages multiple versions of process frames. A process frame is the set of all arguments, local variables and temporary containers of a compile unit, also called a process groupable (e.g. a procedure) at a specific time.
 * Process frames may exist in multiple versions: One version for each CCS process. Because of restrictions of CCS it is not always possible to compile into a single process. This class helps to manage the fragmentation of code into multiple processes.
 * Versions are partially ordered.
 * 
 * ***/

import * as PCT from "@pseuco/static-semantics";
import { PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
import { CCSType, PCCClass, PCCLocalVariable, PCCProcedure, PCCVariable, PCCVariableInfo } from "./PCCProgramController.js";
import { PCCCompiler } from "./PCCCompiler.js";
import { PCCApplicationPlaceholderStackElement, PCCProcessDefinitionStackElement } from "./PCCCompilerStack.js";
import { utils } from "./utils.js";

export interface PCCGroupable {
    //frameCount?: number,

    getProcessName(): string,
    //compilerGetProcedure?(compiler: PCCCompiler, identifier: string): any
}

export class PCCSimpleGroupable implements PCCGroupable {
    constructor(private processName: string) {}

    getProcessName(): string {
        return this.processName;
    }
}


export class PCCProcessFrame {
    // Gives a process frame a unique identifier within the group, which is used to determine a unique name for the created process definition. For scopes the value of this variable is not relevant.
    private processID: number;

    // The number of issued auto-named PCCVariableContainer objects (using this.createContainer).
    private containerIndex: number = 0;

    // An array of PCCVariableContainer objects that are already in use in the scope of the represented process definition.
    private usedContainers: Array<PCCVariableContainer> = new Array();

    // A map from PseuCo identifiers (string) to PCCContainer objects which represents the current value of the corresponding identifier.
    private varTable: Map<string, PCCContainer> = new Map();

    // A stack (array) of PCCContainer objects which represent a temporary value.
    private protections: Array<PCCContainer> = new Array();

    // The initial PCCVariableContainers used as arguments for the process definition
    private initialVariablesVariableContainers?: Array<PCCVariableContainer>;
    private initialProtectionsVariableContainers?: Array<PCCVariableContainer>;

    // The number of incomming variables.
    // private get initialVariableCount(): number {
    //     return this.initialVariablesVariableContainers.length;
    // }
    private initialVariableCount: number;

    // If isScope is false, then the process frame represents a CCS process definition. If it's true, the frame is used to drop variables/protected containers later (i.e., for the typical scope use-case)
    private isScope: boolean = false;

    // One groupable may be represented by partially ordered process frames; parentFrame references a (unique) previous process frame
    private parentFrame?: PCCProcessFrame;


    /*** 
     * @param groupable: An object that implements the «Groupable» interface. All process frames with the same groupable object are in the same process group.
    * @param variables: An array of PCCVariableInfo objects, which represent the PseuCo variables currently covered by the process frame. It contains the incoming variables that will get passed on transition to the represented process definition.
    * @param tempTypes: An array of PCTType objects, which represents the values that will get passed as temporary values on transition to the represented process definition.
    * @param autoInit: By default true. If the created process frame will be followup process frame, the creator of the frame might want to perform its own initialization, which it can communicate through this parameter.
    ***/
    constructor(readonly groupable: PCCGroupable, readonly variables: Array<PCCVariableInfo> = [], readonly tempTypes: Array<CCSType> = [], autoInit: boolean = true) {
        this.initialVariableCount = this.variables.length;
        if (autoInit) {
            this.relateGroupableToProcessFrame();

            // this frame gets as index the number of the already created process frames (so it's unique)
            this.processID = PCCProcessFrame.groupables.get(this.groupable)!.length - 1; // (subtract 1, because `this` is already in that list)

            // auto-assign variable containers for the identifiers of the incoming variables; these will be used as the arguments of the represented process definition
            this.initialVariablesVariableContainers = new Array();
            for (const v of this.variables) {
                const container = this.createContainer(v.getCCSType(), v.getSuggestedContainerName());
                this.initialVariablesVariableContainers.push(container);
                this.varTable.set(v.getIdentifier(), container);
            }

            // the incoming temporary values will remain temporary values, so we'll create a new variable container for each of it and add it to protections array
            this.initialProtectionsVariableContainers = this.tempTypes.map((tempType, i) => this.createContainer(tempType, `t${i}`));
            this.protections = this.initialProtectionsVariableContainers.slice();
        } else {
            // Initialise variables with default values
            this.processID = -1;
        }
    }

    private relateGroupableToProcessFrame() {
        // Keep track of the relations between groupables and process frames
        if (!PCCProcessFrame.groupables.has(this.groupable)) {
            PCCProcessFrame.groupables.set(this.groupable, [this]);
        } else {
            PCCProcessFrame.groupables.get(this.groupable)!.push(this);
        }
    }

    private argumentsForProcessDefinition(): Array<PCCVariableContainer> {
        if (!this.initialVariablesVariableContainers || !this.initialProtectionsVariableContainers) {
            throw new Error("argumentsForProcessDefinition not available for scopes.");
        }
        return this.initialVariablesVariableContainers.concat(this.initialProtectionsVariableContainers);
    }

    
    emitProcessDefinition(compiler: PCCCompiler): PCCProcessDefinitionStackElement | null {
        if (this.isScope) {
            // This frame does not represent a process definition, return
            return null;
        }

        const processArgs = this.argumentsForProcessDefinition();
        return compiler.silentlyAddProcessDefinition(this.getProcessName(), processArgs);
    } 

    private getProcessName(): string {
        return `${this.groupable.getProcessName()}${(this.processID > 0) ? `_${this.processID}` : ""}`;
    }

    private getProcessArgCount(): number {
        if (!this.initialVariablesVariableContainers || !this.initialProtectionsVariableContainers) {
            throw new Error("argumentsForProcessDefinition not available for scopes.");
        }
        return this.initialVariablesVariableContainers.length + this.initialProtectionsVariableContainers.length;
    }


    private didAlreadyUseContainerWithIdentifier(id: string): boolean {
        return this.usedContainers.some((c) => c.identifier == id);
    }

    createContainer(ccsType: CCSType, desiredName?: string): PCCVariableContainer {
        let id: string;
        // Check if the desired name is allowed and free; if not, pick a different name
        if (typeof desiredName == "string" && desiredName.length > 0 && !desiredName.match(/^\$[0-9]/) && !this.didAlreadyUseContainerWithIdentifier(desiredName)) {
            id = desiredName;
        } else {
            id = `$${this.containerIndex++}`;
        }

        const container = new PCCVariableContainer(id, ccsType);
        this.usedContainers.push(container);

        return container;
    }

    getContainerForVariable(identifier: string): PCCContainer {
        if (!this.varTable.has(identifier)) {
            throw new Error("Unknown variable!");
        } else {
            return this.varTable.get(identifier)!;
        }
    }

    assignContainerToVariable(identifier: string, container: PCCContainer) {
        if (!this.varTable.has(identifier)) {
            throw new Error("Unknown variable!");
        } else {
            this.varTable.set(identifier, container)!;
        }
    }

    private addLocalVariable(variable: PCCVariableInfo, container: PCCContainer) {
        if (!this.varTable.has(variable.getIdentifier())) {
            this.variables.push(variable);
            // TODO: This looks suspicious. It seems to use an implicitly assumed invariant that variable in this.variables iff variable in this.varTable. Why not check whether variable is in this.variables?
        }
        this.varTable.set(variable.getIdentifier(), container);
    }


    /***
     * Container protection is a core feature of this compiler. To realise control flow it is sometimes necessary to transition from one CCS process definition to another (follow-up) CCS process definition. In such a case it must be known which temporary values are still needed. Such values can be `protected` and protected containers are always preserved during process definition transition.
     * Example:
     * A := a?x. B[x]
     * B[x] := b!x. 0
     * Here, x must be protected before the transition from A to B is realised. If x were not protected, the result would be erratic CCS process (because of the unbound `x` in B):
     * A := a?x. B
     * B := b!x. 0
     */
    protectContainer(container: PCCContainer) {
        this.protections.push(container);
    }

    unprotectContainer(): PCCContainer {
        if (this.protections.length == 0) {
            throw new Error("No protected containers available.");
        }
        return this.protections.pop()!;
    }

    // Get the most recently protected container
    getProtectedContainer(): PCCContainer {
        if (this.protections.length == 0) {
            throw new Error("No protected containers available.");
        }
        return this.protections[this.protections.length - 1];
    }


    // TODO: What is meant by local variable here? Assumption: a pseuCo local variable.
    isContainerLocalVariable(container: PCCContainer): boolean {
        return Array.from(this.varTable.values()).some((c) => container.isEqual(c));
    }

    createFollowupFrame(): PCCProcessFrame {
        return PCCProcessFrame.createFollowupFrameForFrames([this]);
    }

    protected createFollowupFrameAcceptingTempTypes(tempTypes: Array<CCSType>): PCCProcessFrame {
        return new PCCProcessFrame(this.groupable, this.variables.slice(), tempTypes);
    }

    protected getTypesForTemporaryValues(): Array<CCSType> {
        return this.protections.map((c) => c.ccsType);
    }



    createScope(): PCCProcessFrame {
        const tempTypes = this.getTypesForTemporaryValues();
        const res = new PCCProcessFrame(this.groupable, this.variables.slice(), tempTypes, false);

        // Configure the new frame's properties
        this.configureVariablesInScope(res);

        return res;
    }

    protected configureVariablesInScope(frame: PCCProcessFrame) {
        frame.processID = this.processID;
        frame.parentFrame = this;
        frame.isScope = true;
    }



    /***
     * Guarantees that it is allowed to transition from `frame` to `this`.
     */
    private checkTransitionConsistency(frame: PCCProcessFrame) {
        if (frame.protections.length != this.tempTypes.length) {
            throw new Error("Call consistency is violated!");
        }
        if (frame.groupable != this.groupable) {
            throw new Error("Call consistency is violated!");
        }
        if (this.isScope && (this.varTable.size > 0 || this.usedContainers.length > 0)) { //scope that already transitioned
            // This is only a heuristic for scope transition; if the condition is not satisfied is no guarantee that the scope did not yet transition; we decided not to add more precise logic as this is only an assertion to detect inappropriate usage of process frames
            throw new Error("Call consistency is violated!");
        }
        if (frame.variables.length < this.initialVariableCount) {
            throw new Error("Call consistency is violated!");
        }
        for (let i = 0; i < this.initialVariableCount; i++) {
            if (this.variables[i].getIdentifier() != frame.variables[i].getIdentifier()) {
                throw new Error("Call consistency is violated!");
            }
        }
    }

    private variablesForTransitions(): Array<PCCVariableInfo> {
        return this.variables.slice();
    }

    private argumentsToCallProcessFromFrame(frame: PCCProcessFrame): Array<PCCContainer> {
        const args = this.variablesForTransitions().map((v) => frame.varTable.get(v.getIdentifier()));
        if (!args.every((c) => c != undefined)) {
            throw new Error("Found unknown variable during process transition");
        }
        return (args as Array<PCCContainer>).concat(frame.protections.slice(0, this.tempTypes.length));
    }

    emitCallProcessFromFrame(compiler: PCCCompiler, frame: PCCProcessFrame, appPlaceholder?: PCCApplicationPlaceholderStackElement) {
        this.checkTransitionConsistency(frame);
        const args = this.argumentsToCallProcessFromFrame(frame);
        this.emitCallProcessWithArgumentContainers(compiler, args, appPlaceholder);
    }

    emitCallProcessWithArgumentContainers(compiler: PCCCompiler, containers: Array<PCCContainer>, appPlaceholder?: PCCApplicationPlaceholderStackElement) {
        if (this.isScope) {
            throw new Error("Illegal operation! This method cannot be used with scopes!");
        }
        if (containers.length != this.getProcessArgCount()) {
            throw new Error("Number of argument containers does not match number of required arguments");
        }

        if (appPlaceholder) {
            appPlaceholder.set(this.getProcessName(), containers);
        } else {
            compiler.emitProcessApplication(this.getProcessName(), containers);
        }
    }

    private emitScopeTransition(compiler: PCCCompiler, frame: PCCProcessFrame) {
        this.containerIndex = frame.containerIndex;
        this.usedContainers = frame.usedContainers.slice();
        for (const v of this.variablesForTransitions()) {
            const varContainer = frame.varTable.get(v.getIdentifier());
            if (!varContainer) {
                throw new Error("Variable not found");
            }
            this.varTable.set(v.getIdentifier(), varContainer);
        }
        this.protections = frame.protections.slice();
    }

    emitTransitionFromFrame(compiler: PCCCompiler, frame: PCCProcessFrame) {
        if (this.isScope) {
            this.checkTransitionConsistency(frame);
            this.emitScopeTransition(compiler, frame);
        } else {
            this.emitCallProcessFromFrame(compiler, frame);
        }
    }



    // - Compiler delegate methods
    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): void {
        const c = variable.compileDefaultValue(compiler);
        utils.assertNonNull(compiler.getProcessFrame()).addLocalVariable(variable, c);
        // The order is important here, since compileDefaultValue is allowed to begin new process frames. 
    }

    compilerGetVariable(compiler: PCCCompiler, identifier: string): PCCLocalVariable | null {
        // TODO: Why LocalVariable?
        const v = this.variables.find((v) => v.getIdentifier() == identifier);
        if (v) {
            return new PCCLocalVariable(v.node, v.getName(), v.type, v.isInternal);
        } else {
            return null;
        }
    }

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null {
        return null;
    }



    // A global mapping to keep track of all process frames that were created for the individuals groupables.
    static groupables: Map<PCCGroupable, Array<PCCProcessFrame>> = new Map();



    /***
     * The following code is used to create follow up frames from one or more process frames
     * @param frames: The array with the process frames that the new frame should be derived by.
     */
    static createFollowupFrameForFrames(frames: Array<PCCProcessFrame>): PCCProcessFrame {
        PCCProcessFrame.checkFramesForConsistency(frames);
	    const closestAncestor = PCCProcessFrame.findClosesAncestorForFrames(frames);

        const tempTypes = frames[0].getTypesForTemporaryValues();
        const result = closestAncestor.createFollowupFrameAcceptingTempTypes(tempTypes);
	    result.parentFrame = closestAncestor;
	    return result	
    }

    // Checks if prot1 and prot2 hold the same sequence of ccs types
    static checkTempTypesEquality(prot1: Array<PCCContainer>, prot2: Array<PCCContainer>): boolean {
        if (prot1 === prot2) {
            return true;
        }
        if (prot1.length != prot2.length) {
            return false;
        }
        return prot1.every((c1, i) => c1.ccsType.isEqual(prot2[i].ccsType));
    }

    // Checks if all frames represent the same groupable and has equal protected containers. Throws in case of inconsistency.
    static checkFramesForConsistency(frames: Array<PCCProcessFrame>) {
        const groupable = frames[0].groupable;
        const protections = frames[0].protections;
        for (const frame of frames) {
            if (frame.groupable != groupable || !PCCProcessFrame.checkTempTypesEquality(frame.protections, protections)) {
                throw new Error("Inconsistent process frames");
            }
        }
    }


    // Determine the closest ancestor
    static findClosesAncestorForFrames(frames: Array<PCCProcessFrame>): PCCProcessFrame {
        const currentFrames: Array<PCCProcessFrame | undefined> = frames.slice();

        // Every frame can be 'marked' arbitrary many times
        const markings = new Map<PCCProcessFrame, number>();
        const mark = (frame: PCCProcessFrame) => {
            if (!markings.has(frame)) {
                markings.set(frame, 1);
            } else {
                markings.set(frame, 1 + markings.get(frame)!);
            }
        }

        while (true) {
            for (const [i, currentFrame] of currentFrames.entries()) {
                if (currentFrame != undefined) {
                    mark(currentFrame);
                    if (markings.get(currentFrame)! === frames.length) {
                        return currentFrame;
                    }
                    currentFrames[i] = currentFrame.parentFrame;
                }
            }
        }
    }
}




export class PCCProcedureFrame extends PCCProcessFrame {
    readonly procedure: PCCProcedure;
    readonly tempTypes: Array<CCSType>;
    constructor(procedure: PCCProcedure, variables: Array<PCCVariableInfo> | null = null, tempTypes: Array<CCSType> = [], autoInit: boolean = true) {
        if(!variables) {
            variables = procedure.procedureArguments.slice()
			if(procedure.isClassProcedure()) {
                const cls = procedure.parent as PCCClass;
				variables.unshift(new PCCVariableInfo(null, "i", cls.type, true));
            }
			variables.unshift(new PCCVariableInfo(null, "a", new PCT.SystemType(PCT.ConcreteSystemType.AGENT), true));
        }

        super(procedure, variables, tempTypes, autoInit);
        this.procedure = procedure;
        this.tempTypes = tempTypes;
    }

    getProcedure(): PCCProcedure {
        return this.procedure;
    }

    private createFollowupProcedureFrameAcceptingTempTypes(tempTypes: Array<CCSType>): PCCProcedureFrame {
        return new PCCProcedureFrame(this.procedure, this.variables.slice(), tempTypes);
    }

    protected override createFollowupFrameAcceptingTempTypes(tempTypes: Array<CCSType>): PCCProcessFrame {
        return this.createFollowupProcedureFrameAcceptingTempTypes(tempTypes);
    }

    createScope(): PCCProcedureFrame {
        const res = new PCCProcedureFrame(this.procedure, this.variables.slice(), this.getTypesForTemporaryValues(), false);
        this.configureVariablesInScope(res);
        return res;
    }

    override emitProcessDefinition(compiler: PCCCompiler): PCCProcessDefinitionStackElement | null {
        const res = super.emitProcessDefinition(compiler);
        res?.markAsProcedure();
        return res;
    }




    // - Compiler delegate methods

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure | null {
        const p = this.getProcedure();
        if (p.getName() == identifier) {
            return p;
        } else {
            return p.getProcedureWithName(identifier);
        }
    }


}