import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { PCCProgramController, PCCVariableInfo } from "./PCCProgramController.js";

export class PCCProgramAnalysisVisitor extends PC.AbstractVisitor<void> {
    
    constructor(readonly programController: PCCProgramController) {
        super();
    }

    visitProgram(node: PC.Program, context?: unknown): void {
        node.declarations.forEach((decl) => decl.accept(this));
    }

    visitMainAgent(node: PC.MainAgent, context?: unknown): void {
        this.programController.beginMainAgent(node);
        node.body.accept(this);
        this.programController.endMainAgent();
    }

    visitProcedureDeclaration(node: PC.ProcedureDeclaration, context?: unknown): void {
        const args = node.parameters.map((param) => {
            const declType = this.programController.staticSemantics.typeForNode(param) as PCT.DeclarationType;
            return new PCCVariableInfo(param, param.name, declType.declaredType);
        });
        const procType = (this.programController.staticSemantics.typeForNode(node) as PCT.DeclarationType).declaredType as PCT.ProcedureType;
        const returnType = procType.returnType;
        this.programController.beginNewProcedure(node, node.name, returnType, args);

        node.body.accept(this);

        this.programController.endProcedure();
    }


    visitMonitor(node: PC.Monitor, context?: unknown): void {
        this.programController.beginClass(node.name);
        node.declarations.forEach((decl) => decl.accept(this));
        this.programController.endClass();
    }

    visitStruct(node: PC.Struct, context?: unknown): void {
        this.programController.beginClass(node.name);
        node.declarations.forEach((decl) => decl.accept(this));
        this.programController.endClass();
    }

    visitConditionDeclaration(node: PC.ConditionDeclaration, context?: unknown): void {
        const v = new PCCVariableInfo(node, node.name, PCT.SystemTypes.condition);
        this.programController.processNewVariable(v);
    }


    visitVariableDeclarations(node: PC.VariableDeclarations, context?: unknown): void {
        node.declarators.forEach((decl) => decl.accept(this));
    }

    visitVariableDeclarator(node: PC.VariableDeclarator, context?: unknown): void {
        const type = this.programController.staticSemantics.typeForNode(node) as PCT.DeclarationType;
        const v = new PCCVariableInfo(node, node.name, type.declaredType);
        this.programController.processNewVariable(v);
    }

    



    visitStatementWrapper(node: PC.StatementWrapper, context?: unknown): void {
        node.statement?.accept(this);
    }

    visitBreakStatement(node: PC.BreakStatement, context?: unknown): void {
        // nothing to do
    }

    visitContinueStatement(node: PC.ContinueStatement, context?: unknown): void {
        // nothing to do
    }

    visitStatementBlock(node: PC.StatementBlock, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.blockStatements.forEach((stmt) => stmt.accept(this));
        this.programController.closeEnvironment();
    }

    visitStatementExpression(node: PC.StatementExpression, context?: unknown): void {
        // nothing to do
    }

    visitSelectStatement(node: PC.SelectStatement, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.cases.forEach((cse) => cse.accept(this));
        this.programController.closeEnvironment();
    }

    visitCase(node: PC.Case, context?: unknown): void {
        node.statement.accept(this);
    }

    visitIfStatement(node: PC.IfStatement, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.consequence.accept(this);
        node.alternative?.accept(this);
        this.programController.closeEnvironment();
    }

    visitWhileStatement(node: PC.WhileStatement, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }

    visitDoStatement(node: PC.DoStatement, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }

    visitForStatement(node: PC.ForStatement, context?: unknown): void {
        this.programController.openEnvironment(node);
        node.loopInit?.accept(this);
        node.loopBody.accept(this);
        this.programController.closeEnvironment();
    }

    visitForInit(node: PC.ForInit, context?: unknown): void {
        node.variableInitialisations.forEach((init) => init.accept(this));
    }

    visitReturnStatement(node: PC.ReturnStatement, context?: unknown): void {
        // nothing to do
    }

    visitPrimitiveStatement(node: PC.PrimitiveStatement, context?: unknown): void {
        // nothing to do
    }

    visitPrintStatement(node: PC.PrintStatement, context?: unknown): void {
        // nothing to do
    }

    visitConditionStatement(node: PC.ConditionStatement, context?: unknown): void {
        // nothing to do
    }





    // The remaining visit methods will not be called
    visitProcedureParameter(node: PC.ProcedureParameter, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitVariableInitializer(node: PC.VariableInitializer, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitArrayTypeNode(node: PC.ArrayTypeNode, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitSimpleTypeNode(node: PC.SimpleTypeNode, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitChannelTypeNode(node: PC.ChannelTypeNode, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitClassTypeNode(node: PC.ClassTypeNode, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitStartExpression(node: PC.StartExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitAssignExpression(node: PC.AssignExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitAssignDestination(node: PC.AssignDestination, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitSendExpression(node: PC.SendExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitConditionalExpression(node: PC.ConditionalExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitOrExpression(node: PC.OrExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitAndExpression(node: PC.AndExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitEqualityExpression(node: PC.EqualityExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitRelationalExpression(node: PC.RelationalExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitAdditiveExpression(node: PC.AdditiveExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitMultiplicativeExpression(node: PC.MultiplicativeExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitUnaryExpression(node: PC.UnaryExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitPostfixExpression(node: PC.PostfixExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitReceiveExpression(node: PC.ReceiveExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitProcedureCall(node: PC.ProcedureCall, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitClassCall(node: PC.ClassCall, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitArrayExpression(node: PC.ArrayExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitLiteralExpression(node: PC.LiteralExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }
    visitIdentifierExpression(node: PC.IdentifierExpression, context?: unknown): void {
        throw new Error("Method not implemented.");
    }

}