import * as PC from "@pseuco/lang";
import * as PCT from "@pseuco/static-semantics";
import { PCCBinaryContainer, PCCConstantContainer, PCCContainer, PCCVariableContainer } from "./PCCContainer.js";
import { PCCCompiler } from "./PCCCompiler.js";
import { PCCGroupable, PCCSimpleGroupable } from "./PCCProcessFrame.js";
import { utils } from "./utils.js";
import { PCCVariableInitialisation } from "./PCCCompileProgramVisitor.js";
import { PCCProgramAnalysisVisitor } from "./PCCProgramAnalysisVisitor.js";

export class PCCEnvironmentNode {
    protected _parent?: PCCEnvironmentNode;
    get parent(): PCCEnvironmentNode {
        return utils.assertNonNull(this._parent);
    }
    private children: Array<PCCEnvironmentNode> = new Array();
    protected _variables: Map<string, PCCVariable> = new Map();
    get variables(): ReadonlyMap<string, PCCVariable> {
        return this._variables;
    }
    protected _procedures: Map<string, PCCProcedure> = new Map();
    get procedures(): ReadonlyMap<string, PCCProcedure> {
        return this._procedures;
    }
    protected _blocks: Array<PCCEnvironmentNode> = new Array();
    get blocks(): ReadonlyArray<PCCEnvironmentNode> {
        return this._blocks;
    }

    __id?: number;  // numeric id that is assigned elsewhere

    constructor(public node: PC.Node, public label: string) {}

    addChild(child: PCCEnvironmentNode | PCCVariable) {
        if (child instanceof PCCProcedure) {
            this._procedures.set(child.getName(), child);
            child._parent = this;
        } else if (child instanceof PCCVariable) {
			this._variables.set(child.getIdentifier(), child);
        } else if (child.__id /*child instanceof PCCEnvironmentNode*/) {
			this._blocks[child.__id] = child;
            child._parent = this;
        } else {
            this.children.push(child);
            child._parent = this;
        }
    }

    getVariableWithName(name: string): PCCVariable {
        if (!this.variables.has(name)) {
            if (!this._parent) {
                throw new Error(`Variable '${name}' wasn't declared.`);
            }
            return this._parent.getVariableWithName(name);
        } else {
            return this.variables.get(name)!;
        }
    }

    getProcedureWithName(name: string): PCCProcedure {
        if (!this.procedures.has(name)) {
            if (!this._parent) {
                throw new Error(`Procedure '${name}' wasn't declared.`);
            } else {
                return this._parent.getProcedureWithName(name);
            }
        } else {
            return this.procedures.get(name)!;
        }
    }

    constructVariableObject(variableNode: PC.Node | null, variableName: string, variableType: PCT.Type): PCCVariable {
        return this.parent.constructVariableObject(variableNode, variableName, variableType);
    }

    getBlockWithId(id: number) {
        return this.blocks[id];
    }

    getComposedLabel(): string {
        const parentLabel = this._parent?.getComposedLabel();
        return `${(parentLabel && parentLabel.length > 0) ? `${parentLabel}_$` : ""}${this.label}`;
    }

    getAllClasses(): Array<PCCClass> {
        var result: Array<PCCClass> = [];
        for (const c of this.children) {
            if (c instanceof PCCClass) {
                result.push(c);
            }
        }
        return result;
    }

    isDescendantOfProcedure(): boolean {
        return this._parent?.isDescendantOfProcedure() === true;
    }

    // getVariableClass() {
    //     return this.parent?.getVariableClass();
    // }

    // getUsedTypes(res) {
    //     this.children.forEach((c) => c.getUsedTypes(res));
    // }
}

export class PCCGlobal extends PCCEnvironmentNode {
    constructor(program: PC.Program) {
        super(program, "");
    }

    override constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCGlobalVariable {
        return new PCCGlobalVariable(variableNode, variableName, variableType);
    }

    compilerGetVariable(compiler: PCCCompiler, identifier: string) {
        return this.getVariableWithName(identifier);
    }

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure {
        return this.getProcedureWithName(identifier);
    }

    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo): PCCVariable {
        const variable2 = this.getVariableWithName(variable.getName());
        if (!variable2) {
            throw new Error("Unexpected new identifier found!")
        }
        if (!(variable2 instanceof PCCGlobalVariable)) {
            throw new Error("Illegal state.");
        }
        variable2.emitConstructor(compiler);
        return variable2;
    }
}

export class PCCClass extends PCCEnvironmentNode {
    protected usedClassTypes: Array<PCT.ClassType> = new Array();
    // protected _hasVariables?: boolean;
    constructor(node: PC.Node, public type: PCT.ClassType) {
        super(node, type.className);
        if (this.isMonitor()) {
            this.addChild(new PCCInternalReadOnlyField(null, "guard", PCT.SystemTypes.lock, this));
        }
    }

    override addChild(child: PCCEnvironmentNode | PCCVariable) {
        if (child instanceof PCCVariable) {
            if (!(child instanceof PCCField)) {
                throw new Error("Expected variables of class to be of type PCCField");
            }
        }
        super.addChild(child);
    }

    get fields(): ReadonlyMap<string, PCCField> {
        // This type conversion is protected by the `addChild` override in this class
        return this.variables as ReadonlyMap<string, PCCField>;
    }

    getName(): string {
        return this.label;
    }

    isMonitor(): boolean {
        return this.type.concreteClassType == PCT.ConcreteClassType.MONITOR;
    }

    addUseOfClassType(type: PCT.ClassType) {
        this.usedClassTypes.push(type);
    }

    getUsedClassTypes(): ReadonlyArray<PCT.Type> {
        // ToDo: Remove the .slice() once the code types correctly with the ReadonlyArray return type
        return this.usedClassTypes.slice();
    }

    hasEnvironment(): boolean {
        // if (this._hasVariables == undefined) {
        //     this._hasVariables = false;
        //     for (const n in this.variables) {
        //         this._hasVariables = true;
        //         break;
        //     }
        // }
        // return this._hasVariables;
        return this.variables.size > 0;
    }

    getAllConditions(): Array<PCCVariable> {
        const result: Array<PCCVariable> = [];
        for (const [n, v] of this.variables) {
            if (v.type?.isConditionType) {
                result.push(v);
            }
        }
        return result;
    }

    override constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCField {
        return new PCCField(variableNode, variableName, variableType, this, false);
    }

    compilerGetVariable(compiler: PCCCompiler, identifier: string) {
        return this.getVariableWithName(identifier);
    }

    compilerGetProcedure(compiler: PCCCompiler, identifier: string): PCCProcedure {
        return this.getProcedureWithName(identifier);
    }

    compilerHandleNewVariable(compiler: PCCCompiler, variable: PCCVariableInfo) {
        this.getVariableWithName(variable.getName());
    }

    emitEnvironment(compiler: PCCCompiler) {
        const variableNames = Array.from(this.fields.keys());
        const variables: Array<PCCVariableInfo> = Array.from(this.fields.values());
        // internal variable for class instance
        const instanceInfo = new PCCVariableInfo(null, "i", this.type, true);
        variableNames.unshift(instanceInfo.getIdentifier()); 
        variables.unshift(instanceInfo);
        compiler.beginProcessGroup(new PCCSimpleGroupable(this.getEnvProcessName()), variables);
        const localVariables = variableNames.map((name) => {
            const variable = compiler.getVariableWithName(name);
            if (!variable) {
                throw new Error("Unknown variable.");
            } else {
                return variable;
            }
        });
        const instance = localVariables[0];
        const variablesEntries = Array.from(this.fields.entries());
        for (let i = 0; i < variablesEntries.length-1; i++) {
            const [n, v] = variablesEntries[i];
            const control = compiler.emitChoice();
            // Ask the variable to emit the accessor actions; note that we must provide the full list of field variables, because the variable's implementation must know about all variables the environment process stores
            v.emitAccessors(compiler, localVariables, instance.getContainer(compiler));
            control.setBranchFinished();
        }
        const [n, v] = variablesEntries[variablesEntries.length-1];
        v.emitAccessors(compiler, localVariables, instance.getContainer(compiler));
        compiler.endProcessGroup();
    }

    emitConstructor(compiler: PCCCompiler) {
        if (this.hasEnvironment()) {
            this.emitEnvironment(compiler);
        }
        
        const envName = this.getEnvProcessName();
        const nextIVar = new PCCVariableInfo(null, "next_i", this.type, true);
        compiler.beginProcessGroup(new PCCSimpleGroupable(envName+"_cons"), [nextIVar]);
        const entry = utils.assertNonNull(compiler.getProcessFrame());
        const nextIVarLocal = utils.assertNonNull(compiler.getVariableWithNameOfClass(nextIVar.getName(), null, nextIVar.isInternal));

        compiler.emitOutput(`class_${this.getName()}_create`, null, nextIVarLocal.getContainer(compiler));
        compiler.protectContainer(nextIVarLocal.getContainer(compiler));
        let numberOfProtectedVariables = 1;
        for(const v of this.variables.values()) {
            const c = v.compileDefaultValue(compiler);
            compiler.protectContainer(c);
            numberOfProtectedVariables++;
        }

        const vars = new Array<PCCContainer>();
        for (let i = 0; i < numberOfProtectedVariables; i++) {
            vars.unshift(compiler.unprotectContainer());
        }
        const nextIVarRecovered = vars[0];
		compiler.emitOutput(`class_${this.getName()}_initialized`, nextIVarRecovered, null)  // vars[0] is the instance number of the object being created
        // `nextIVarLocal.getContainer(compiler)` should always be equal to `nextIVarRecovered`; it would be more elegant to use the former; ToDo
        
        // If this class has an environment, we must in parallel initiate the environment process and must recursively return to the constructor process (with `next_i + 1`)
        const control = (this.hasEnvironment()) ? compiler.emitParallel() : null;
        if(this.hasEnvironment()) {
			compiler.emitProcessApplication(envName, vars);
            control!.setBranchFinished()
        }
        
        // Update nextIVarLocal for recursive call
        const nextIValNew = new PCCBinaryContainer(nextIVarRecovered, new PCCConstantContainer(1), "+");
        nextIVarLocal.setContainer(compiler, nextIValNew);

        entry.emitCallProcessFromFrame(compiler, utils.assertNonNull(compiler.getProcessFrame()));
        control?.setBranchFinished()

        compiler.endProcessGroup()
        
        // Add the constructor process to the system processes
        compiler.emitSystemProcessApplication(this.getProcessName(), [new PCCConstantContainer(1)], PCCCompiler.SysEnvironment);
    }

    // constructorGetName(): string {
    //     return this.getEnvProcessName();
    // }

    // constructorGetArguments(): Array<PCCVariableInfo> {
    //     return [new PCCVariableInfo(null, "next_i", this.type, true)];
    // }

    // constructorProtectEnvironmentArguments(/*cons: PCCConstructor, */compiler: PCCCompiler, variables: Array<PCCVariable>): number {
    //     const instance = variables[0];
    //     compiler.emitOutput(`class_${this.getName()}_create`, null, instance.getContainer(compiler));
    //     compiler.protectContainer(instance.getContainer(compiler));
    //     let numberOfProtectedVariables = 1;
    //     for(const v of this.variables.values()) {
    //         const c = v.compileDefaultValue(compiler);
    //         compiler.protectContainer(c);
    //         numberOfProtectedVariables++;
    //     }
    //     return numberOfProtectedVariables;
    // }

    // constructorShouldCallRecursively() {
    //     return true;
    // }

    // constructorUpdateVariablesForRecursiveCall(cons: PCCConstructor, compiler: PCCCompiler, entry: any, variables:Array<PCCVariable>) {
    //     const instance = variables[0];
    //     const new_i = new PCCBinaryContainer(instance.getContainer(compiler), new PCCConstantContainer(1), "+");
    //     instance.setContainer(compiler, new_i);
    // }

    getProcessName(): string {
        return `Env_class_${this.getName()}_cons`;
    }

    getEnvProcessName(): string {
        return `Env_class_${this.getName()}`;
    }

}

export class PCCProcedure extends PCCEnvironmentNode implements PCCGroupable {
   constructor(node: PC.Node, name: string, readonly returnType: PCT.Type, readonly procedureArguments: Array<PCCVariableInfo>) {
        super(node, name);
        for (const arg of this.procedureArguments) {
            const v = this.constructVariableObject(utils.assertNonNull(arg.node), arg.getName(), arg.type);
            // this._variables.set(arg.getName(), arg);
            this.addChild(v);
        }
    }

    getName(): string {
        return this.label;
    }

    isDescendantOfProcedure(): boolean {
        return true;
    }

    // Removed isNestedProcedure()


    isClassProcedure(): boolean {
        return this._parent instanceof PCCClass;
    }

    isMonitorProcedure(): boolean {
        return this._parent instanceof PCCClass && this._parent.isMonitor();
    }


    getProcessName(): string {
        return `Proc_${this.getComposedLabel()}`;
    }

    getAgentStarterChannel(): string {
        return `start_${this.getComposedLabel()}`;
    }

    getAgentProcessName(): string {
        return `Agent_${this.getComposedLabel()}`;
    }

    override constructVariableObject(variableNode: PC.Node, variableName: string, variableType: PCT.Type): PCCLocalVariable {
        return new PCCLocalVariable(variableNode, variableName, variableType);
    }

    getAllArgumentContainers(compiler: PCCCompiler, explicitArgumentContainers: Array<PCCContainer>, instanceContainer: PCCContainer | null): Array<PCCContainer> {
        const argumentContainers = explicitArgumentContainers.slice();
        if (this.isClassProcedure()) {
            const i = instanceContainer ?? utils.assertNonNull(compiler.getVariableWithNameOfClass("i", null, true)).getContainer(compiler);
            argumentContainers.unshift(i);
        }
        const a = utils.assertNonNull(compiler.getVariableWithNameOfClass("a", null, true)).getContainer(compiler);
        argumentContainers.unshift(a);
        return argumentContainers;
    }

    getImplicitAndExplicitArgumentCount(): number {
        if (this.isClassProcedure()) {
            return this.procedureArguments.length + 2;
        } else {
            return this.procedureArguments.length + 1;
        }
    }

    emitAgentConstructor(compiler: PCCCompiler, addJoiner: boolean = true) {
        const definitionName = this.getAgentProcessName();
        compiler.beginProcessDefinition(definitionName, []);
        const i = new PCCVariableContainer("i", CCSType.INT);
        compiler.emitInput("agent_new", null, i);
		compiler.emitOutput(this.getAgentStarterChannel(), null, i);
		const starter = new PCCVariableContainer("starter", CCSType.INT)
		compiler.emitInput("start_set_arg", i, starter)	// We ignore this first (implicit) argument `starter`, because we currently do not use the agent information of the agent which started the new agent
        const args: Array<PCCVariableContainer> = [];
        for (let j = 1; j < this.getImplicitAndExplicitArgumentCount(); j++) {
            const a = new PCCVariableContainer(`a${j}`, CCSType.VOID); // Type is unimportant here
            compiler.emitInput("start_set_arg", i, a);
			args.push(a);
        }

        const control1 = compiler.emitParallel();
        compiler.emitProcessApplication(definitionName, []);
		control1.setBranchFinished();
		const control2 = (addJoiner) ? compiler.emitParallel() : null;
		const control3 = compiler.emitSequence()
		args.unshift(i);		// Basically, here we replace 'starter' with the instance number of the new agent
		compiler.emitProcessApplication(this.getProcessName(), args);
		control3.setBranchFinished();

        if (!this.returnType.isVoidType) {
            // Eat the return value
            compiler.emitInput("rreturn", null, new PCCVariableContainer("$0", CCSType.fromPseucoType(this.returnType)));
        }
        compiler.emitOutput("agent_terminate", i, null);
		compiler.emitStop();
		control3.setBranchFinished();

        if (control2) {
            control2.setBranchFinished();
			compiler.emitProcessApplication("AgentJoiner", [i, new PCCConstantContainer(0)]);
			control2.setBranchFinished();
        }
        control1.setBranchFinished();
		compiler.endProcessDefinition();
    }

    emitExit(compiler: PCCCompiler) {
        if (this.isMonitorProcedure()) {
            const guard = utils.assertNonNull(compiler.getVariableWithNameOfClass("guard", null, true));
			const agent = utils.assertNonNull(compiler.getVariableWithNameOfClass("a", null, true));
			compiler.emitOutput("unlock", guard.getContainer(compiler), agent.getContainer(compiler))
        }
        compiler.emitExit();
    }

}

export class PCCVariableInfo {
    constructor(readonly node: PC.Node | null, readonly label: string, readonly type: PCT.Type, readonly isInternal: boolean = false) { }

    /**
     * Get the name of the variable
     * @returns The name of the variable
     */
    getName(): string {
        return this.label;
    }

    /**
     * Variables can be internal or regular. The identifier provides a unique name to distinguish internal and regular variables in the same name space.
     * @returns Returns an identifier for the variable
     */
    getIdentifier(): string {
        if (this.isInternal) {
            return `#${this.getName()}`;
        } else {
            return this.getName();
        }
    }

    /**
     * 
     * @deprecated 
     */
    getComposedLabel(): string {
        // if (this.parent) {
        //     return `${this.parent.getComposedLabel()}_$${this.label}`;
        // } else {
        //     return this.label;
        // }
        return this.label;
    }

    /**
     * Provides a common way to transfer pseuCo variable identifiers into CCS variable identifiers.
     * @returns An identifier for a CCS variable
     */
    getSuggestedContainerName(): string {
        if (this.isInternal) {
            return PCCVariableInfo.suggestCCSVariableNameForInternalVariable(this.getName());
        } else {
            return PCCVariableInfo.suggestCCSVariableNameForPseucoVariable(this.getName());
        }
    }

    /**
     * Translates pseuCo value types into CCS value types.
     * @returns The type of CCS values encoding the pseuCo values that this variable represents.
     */
    getCCSType() {
        return CCSType.fromPseucoType(this.type);
    }

    /**
     * Compiles a variable's default value.
     * @param compiler The shared compiler object
     * @returns A container that holds the variable's default value
     */
    compileDefaultValue(compiler: PCCCompiler): PCCContainer {
        return PCCVariableInitialisation.compileDefaultValue(compiler, this);
    }


    private static suggestCCSVariableNameForInternalVariable(name: string) {
        return name;
    }

    private static suggestCCSVariableNameForPseucoVariable(name: string) {
        return `$${name}`;
    }
}


export abstract class PCCVariable extends PCCVariableInfo {
    /**
     * Retrieves the current value of this variable and returns it as Container object.
     * @param compiler The shared compiler object
     * @returns A container that holds the variable's current value.
     */
    abstract getContainer(compiler: PCCCompiler): PCCContainer;

    /**
     * Stores the value of `container` as the variable's current value.
     * @param compiler The shared compiler object
     * @param container The new value for this variable
     */
    abstract setContainer(compiler: PCCCompiler, container: PCCContainer): void;
}

export class PCCGlobalVariable extends PCCVariable {
    accessorChannel(set: boolean): string {
        const t = (set) ? "set" : "get";
        return `env_global_${t}_${this.getName()}`;
    }

    override getContainer(compiler: PCCCompiler): PCCContainer {
        const result = compiler.getFreshContainer(CCSType.fromPseucoType(this.type));
		compiler.emitInput(this.accessorChannel(false), null, result);
		return result;
    }

    override setContainer(compiler: PCCCompiler, container: PCCContainer) {
        compiler.emitOutput(this.accessorChannel(true), null, container);
    }


    /**
     * Emits actions to the environment process of a global variable, monitor variable, or struct variable to get or set the value of the variable.
     * @param compiler The shared compiler coordination object
     * @param variables A list of variables that is stored in the CCS environment process to which the accessors are added; always contains at least `this`
     * @param instance For accessors of monitor (or struct) variables, this parameter is a container that contains a reference to the monitor (struct) object.
     */
    emitAccessors(compiler: PCCCompiler, variables: Array<PCCVariable>, instance: PCCContainer | null = null) {
        const local = utils.assertNonNull(compiler.getVariableWithName(this.getIdentifier()));
        const containers = variables.map((v) => v.getContainer(compiler));
        const control = compiler.emitChoice();
		const c = local.getContainer(compiler);
		compiler.emitOutput(this.accessorChannel(false), instance, c);
		compiler.emitProcessApplication(this.getEnvProcessName(), containers);
		control.setBranchFinished();
        // It is not clear from the type system at this point, but we may assume that c is a PCCVariableContainer object (because this method is called by `emitConstructor`, which creates a ProcessFrame, which creates `c` in its constructor and remains unchanged after that.)
        if (!(c instanceof PCCVariableContainer)) {
            throw new Error("Internal inconsistency.");
        }
		compiler.emitInput(this.accessorChannel(true), instance, c);
        compiler.emitProcessApplication(this.getEnvProcessName(), containers);
		control.setBranchFinished();
    }

    emitConstructor(compiler: PCCCompiler) {
        compiler.beginProcessGroup(new PCCSimpleGroupable(this.getEnvProcessName()), [this]);
        const local = utils.assertNonNull(compiler.getVariableWithName(this.getIdentifier()));
        this.emitAccessors(compiler, [local]);
        compiler.endProcessGroup();

        // Since initialisations happen later, we do not need constructors anymore and can spawn the environment directly
        const container = this.getCCSType().getDefaultContainer();
        compiler.emitSystemProcessApplication(this.getEnvProcessName(), [container], PCCCompiler.SysEnvironment);

        // compiler.beginProcessGroup(this, []);
        // // const container = PCCVariableInitialisation.compileDefaultValue(compiler, this);
        // // Use the default CCS value now and initialise later immediately before the main agent starts
        // const container = this.getCCSType().getDefaultContainer();
        // compiler.emitProcessApplication(this.getEnvProcessName(), [container]);
		// compiler.endProcessGroup();
		// compiler.emitSystemProcessApplication(this.getProcessName(), [], PCCCompiler.SysEnvironment);
    }

    getProcessName(): string {
        return `Env_global_${this.getName()}_cons`;
    }

    getEnvProcessName(): string {
        return `Env_global_${this.getName()}`;
    }
}

export class PCCField extends PCCGlobalVariable {
    constructor(node: PC.Node | null, label: string, type: PCT.Type, readonly cls: PCCClass, isInternal: boolean = false) {
        super(node, label, type, isInternal);
    }

    accessorChannel(set: boolean): string {
        return `env_class_${this.cls.getName()}_${(set) ? "set" : "get"}_${this.getName()}`
    }

    override getContainer(compiler: PCCCompiler): PCCContainer {
        // We know that this.getIdentifier() != "#guard"
        const result = compiler.getFreshContainer(CCSType.fromPseucoType(this.type));
        compiler.emitInput(this.accessorChannel(false), compiler.getVariableWithNameOfClass("i", null, true)!.getContainer(compiler), result); // ToDo: Understand why className = null in getVariableWithNameOfClass call
        return result;
    }

    override setContainer(compiler: PCCCompiler, container: PCCContainer) {
        compiler.emitOutput(this.accessorChannel(true), compiler.getVariableWithNameOfClass("i", null, true)!.getContainer(compiler), container);
    }

    getEnvProcessName(): string {
        return `Env_class_${this.cls.getName()}`;
    }
}

export class PCCInternalReadOnlyField extends PCCField {
    constructor(node: PC.Node | null, label: string, type: PCT.Type, cls: PCCClass) {
        super(node, label, type, cls, true);
    }

    override emitAccessors(compiler: PCCCompiler, variables: Array<PCCVariable>, instance: PCCContainer): void {
        const local = compiler.getVariableWithName(this.getIdentifier());
        if (!local) {
            throw new Error("Illegal state.");
        }
        const containers = variables.map((v) => v.getContainer(compiler));
        const c = local.getContainer(compiler);
        compiler.emitOutput(`env_class_${this.cls.getName()}_guard`, instance, c);  // too much hard coded? ToDo: Try to replace guard in string by `this.getIdentifier()`
        compiler.emitProcessApplication(this.getEnvProcessName(), containers);
    }

    override getContainer(compiler: PCCCompiler): PCCContainer {
        const result = compiler.getFreshContainer(CCSType.INT);
        compiler.emitInput(`env_class_${this.cls.getName()}_guard`, compiler.getVariableWithNameOfClass("i", null, true)!.getContainer(compiler), result);
        return result;
    }

    override setContainer(compiler: PCCCompiler, container: PCCContainer): void {
        throw new Error("Setting container for read only variable!");
    }
}

export class PCCCondition extends PCCField {
    constructor(node: PC.Node | null, label: string, cls: PCCClass, readonly expressionNode: PC.Node) {
        super(node, label, PCT.SystemTypes.agent, cls, false);
    }
}

// LocalVariables are variables for which a value exists as a CCS expression, stored in a process frame
export class PCCLocalVariable extends PCCVariable {
    override getContainer(compiler: PCCCompiler): PCCContainer {
        const result = compiler.getProcessFrame()?.getContainerForVariable(this.getIdentifier());
        if (!result) {
            throw new Error("Cannot get container.");
        }
        return result;
    }

    override setContainer(compiler: PCCCompiler, container: PCCContainer) {
        const frame = compiler.getProcessFrame();
        if (!frame) {
            throw new Error("Cannot set container");
        }
        frame.assignContainerToVariable(this.getIdentifier(), container);
    }
}



export class PCCProgramController {
    root: PCCGlobal; // = new PCCGlobal(null);
    classes: Map<string, PCCClass> = new Map();
    agents: Map<string, PCCProcedure> = new Map();
    protected _envStack: PCCEnvironmentNode;
    protected blockCounter = 1;
    pseucoToEnvironmentNode: Map<PC.Node, PCCEnvironmentNode> = new Map();

    constructor(readonly program: PC.Program, readonly staticSemantics: PCT.StaticSemantics) {
        this.root = new PCCGlobal(program);
        this._envStack = this.root;
    }

    analyseProgram() {
        // Prepare class info
        const classDecls = this.program.declarations.filter((decl) => decl instanceof PC.Monitor || decl instanceof PC.Struct) as Array<PC.Monitor | PC.Struct>;
        classDecls.forEach((decl) => {
            const declType = this.staticSemantics.typeForNode(decl) as PCT.DeclarationType;
            const classType = declType.declaredType as PCT.ClassType;
            this.processNewClass(decl, classType);
        });
        
        // Collect the hierarchy of environments
        const visitor = new PCCProgramAnalysisVisitor(this);
        this.program.accept(visitor);
        // ToDo: Try to use information from staticSemantics instead of building another env tree
    }

    static PCTEnvironmentControllerInternalMainAgentName = "$mainAgent";

    getGlobal(): PCCGlobal {
        return this.root;
    }

    getStackState(): PCCEnvironmentNode {
        return this._envStack;
    }
    private restoreStackState(stackState: PCCEnvironmentNode) {
        this._envStack = stackState;
    }

    openEnvironment(node: PC.Node): PCCEnvironmentNode {
        // node.__id = this.blockCounter++;
        const blockID = this.blockCounter++;
        const child = new PCCEnvironmentNode(node, `#${blockID}`);
        child.__id = blockID;
        this.pseucoToEnvironmentNode.set(node, child);
        this._envStack.addChild(child);
        this._envStack = child;
        return child;
    }

    reopenEnvironment(node: PC.Node) {
        // Look up corresponding environment node in mapping
        const envNode = this.pseucoToEnvironmentNode.get(node);
        // Throw if it cannot be found
        if (envNode == undefined) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : true, "name" : "BlockNotFound", "message" : "Block not found in pseuco to environment node mapping!"});
        }
        // Throw if the env node cannot be found in parent's blocks arrays
        if (!envNode.__id || envNode != this._envStack.getBlockWithId(envNode.__id)) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : true, "name" : "BlockNotFound", "message" : "Block not found in list of list of environment blocks!"});
        }
        this._envStack = envNode;
        return envNode;
    }

    // The following method was removed, because apparently it is not needed
    // getEnvironment(node: PC.Node, id: number) {
    //     const node2 = this._envStack.getBlockWithId(id);
    //     if (!node2) {
    //         throw ({"line" : 0, "column" : 0, "wholeFile" : true, "name" : "BlockNotFound", "message" : "Block not found!"});
    //     }
    //     this._envStack = node2;
    //     return node2;
    // }

    closeEnvironment() {
        if (!(this._envStack instanceof PCCEnvironmentNode)) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : true, "name" : "MissingEnvironment", "message" : "No open environment!"});
        }
        this._envStack = this._envStack.parent!;
    }

    getClassWithName(name: string): PCCClass {
        const result = this.classes.get(name);
        if (result == undefined) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : false, "name" : "UnknownClass", "message" : `Unknown class '${name}'`})
        }
        return result;
    }

    getAllClasses() {
        return this.root.getAllClasses();
    }

    
    getVariableWithName(name: string) {
        return this._envStack.getVariableWithName(name);
    }

    getProcedureWithName(name: string) {
        return this._envStack.getProcedureWithName(name);
    }

    getMainAgent() {
        return this.getProcedureWithName(PCCProgramController.PCTEnvironmentControllerInternalMainAgentName);
    }

    processNewClass(node: PC.Monitor | PC.Struct, classType: PCT.ClassType) {
        const tnode = new PCCClass(node, classType);
        this._processNewClass(tnode);
    }

    _processNewClass(node: PCCClass) {
        if (this.classes.has(node.getName())) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : false, "name" : "Redeclaration", "message" : `Class ${node.getName()} already declared!`});
        }
        this._envStack.addChild(node);
        this.classes.set(node.getName(), node);
    }

    beginClass(className: string) {
        const node = this.getClassWithName(className);
        this._envStack = node;
        return node;
    }

    endClass() {
        if (!(this._envStack instanceof PCCClass)) {
            throw ({"line" : 0, "column" : 0, "wholeFile" : true, "name" : "ClassError", "message" : "No class did begin!"});
        }
        this._envStack = this._envStack.parent!;
    }

    beginNewProcedure(node: PC.ProcedureDeclaration | PC.MainAgent, procedureName: string, returnType: PCT.Type, args: PCCVariableInfo[]) {
        const tnode = new PCCProcedure(node, procedureName, returnType, args);
        this._beginNewProcedure(tnode);
    }

    protected _beginNewProcedure(node: PCCProcedure) {
        try {
            // We want the following to fail, because the procedure is supposed to be "new"
            this._envStack.getProcedureWithName(node.getName());
        } catch (error) {
            // If it fails, continue beginning the new procedure
            this._envStack.addChild(node);
            return this.beginProcedure(node.getName());
        }
        // If we reach this point, the procedure is not new
        throw new Error(`Procedure ${node.getName()} already declared!`);
    }

    beginProcedure(procedureName: string) {
        const node = this._envStack.getProcedureWithName(procedureName);
        this._envStack = node;
        return node;
    }

    endProcedure() {
        if (!(this._envStack instanceof PCCProcedure)) {
            throw new Error("No procedure did begin!");
        }
        this._envStack = this._envStack.parent!;
    }

    beginMainAgent(node: PC.MainAgent) {
        // mainAgent is treated like a procedure
        // Either create the procedure here or enter the already existing one
        try {
            return this.beginNewProcedure(node, PCCProgramController.PCTEnvironmentControllerInternalMainAgentName, PCT.PrimitiveTypes.void, []);
        } catch (error) {
            return this.beginProcedure(PCCProgramController.PCTEnvironmentControllerInternalMainAgentName);
        }
    }

    endMainAgent() {
        return this.endProcedure();
    }

    processNewVariable(variable: PCCVariableInfo) {
        const tnode = this._envStack.constructVariableObject(variable.node, variable.getName(), variable.type);
        return this._processNewVariable(tnode);
    }

    _processNewVariable(variable: PCCVariable) {
        // Needs check if variable is indeed "new"
        try {
            // This is supposed to fail for truly new variables
            this._envStack.getVariableWithName(variable.getName());
        } catch (error) {
            // Continue here ro process the indeed new variable
            this._envStack.addChild(variable)
            return;
        }
        // If variable is already known, throw an error
        throw new Error(`Variable ${variable.getName()} already declared!`);
    }

    processProcedureAsAgent(procedure: PCCProcedure) {
        this.agents.set(procedure.getName(), procedure);
    }

    getAgents(): Array<PCCProcedure> {
        return Array.from(this.agents.values()); //.filter((proc) => proc instanceof PCCProcedure);
    }


    // Internal remark: Remove this method and get the used types directly from the type mapping of the static semantics
    // getUsedTypes() {
    //     const res: any = {
    //         arrays: {},
    //         channels: {}
    //     };
    //     this.root.getUsedTypes(res);
    //     return res;
    // }

    getUsedTypes(): {arrays: Set<number>, channels: Set<number>} {
        const res = {
            arrays: new Set<number>(),
            channels: new Set<number>()
        };
        this.staticSemantics.allNodesAndTypes.forEach(([node, type]) => {
            if (node instanceof PC.VariableDeclarator) {
                this.collectTypes(type, res)
            }
        });
        return res;
    }

    private collectTypes(nextType: PCT.Type, result: {arrays: Set<number>, channels: Set<number>}) {
        if (nextType instanceof PCT.ArrayType) {
            if (nextType.size == "unknown") {
                throw new Error("Unexpected array size.");
            }
            result.arrays.add(nextType.size);
            this.collectTypes(nextType.elementsType, result);
        } else if (nextType instanceof PCT.ChannelType) {
            if (nextType.capacity == "unknown") {
                throw new Error("Unexpected channel capacity.");
            }
            result.channels.add(nextType.capacity);
        } else if (nextType instanceof PCT.DeclarationType) {
            this.collectTypes(nextType.declaredType, result);
        }
    }
}





export enum CCSTypeDescriptor {
    VOID, BOOL, INT, STRING
}
export class CCSType {
    constructor(readonly descriptor: CCSTypeDescriptor) { }

    isVoid(): boolean {
        return this.descriptor == CCSTypeDescriptor.VOID;
    }

    isInt(): boolean {
        return this.descriptor == CCSTypeDescriptor.INT;
    }

    isString(): boolean {
        return this.descriptor == CCSTypeDescriptor.STRING;
    }

    isBool(): boolean {
        return this.descriptor == CCSTypeDescriptor.BOOL;
    }

    isArrayLike(): boolean {
        return false;
    }

    isClass(): boolean {
        return false;
    }

    isEqual(other: CCSType): boolean {
        return this.descriptor == other.descriptor && !other.isArrayLike() && !other.isClass();
    }

    getDefaultContainer(): PCCContainer {
        // Computes a PCCContainer with a default value for the CCSType's native type (i.e., the type descriptor)
        // From the perspective of pseuCo types (from which a CCSType is inferred), this method computes an over-approximation. Some pseuCo types require custom default container creation (e.g., the default value for an array is a non-zero reference to a newly created array instance)
        // Still, there are intermediate steps in the CCS interpretation of a pseuCo program where "null references" are used and necessary, e.g., during the initialisation of an array of locks
        if (this.isArrayLike() || this.isClass()) {
            return new PCCConstantContainer(0);
        }
        switch (this.descriptor) {
            case CCSTypeDescriptor.INT:
                return new PCCConstantContainer(0);
            
            case CCSTypeDescriptor.BOOL:
                return new PCCConstantContainer(false);
            
            case CCSTypeDescriptor.STRING:
                return new PCCConstantContainer("");
        
            default:
                throw new Error("Void does not have a default value");
        }
    }


    static readonly VOID = new CCSType(CCSTypeDescriptor.VOID);
    static readonly BOOL = new CCSType(CCSTypeDescriptor.BOOL);
    static readonly INT = new CCSType(CCSTypeDescriptor.INT);
    static readonly STRING = new CCSType(CCSTypeDescriptor.STRING);

    static fromPseucoType(pType: PCT.Type): CCSType {
        if (pType instanceof PCT.PrimitiveType) {
            switch (pType.concretePrimitiveType) {
                case PCT.ConcretePrimitiveType.BOOL:
                    return CCSType.BOOL;
                case PCT.ConcretePrimitiveType.INT:
                    return CCSType.INT;
                case PCT.ConcretePrimitiveType.STRING:
                    return CCSType.STRING;
                case PCT.ConcretePrimitiveType.VOID:
                    return CCSType.VOID;
            }
        } else if(pType instanceof PCT.SystemType) {
            switch (pType.concreteSystemType) {
                case PCT.ConcreteSystemType.LOCK:
                    return CCSType.INT;
                case PCT.ConcreteSystemType.CONDITION:
                    return CCSType.INT;
                case PCT.ConcreteSystemType.AGENT:
                    return CCSType.INT; 
            }
        } else if (pType instanceof PCT.ChannelType) {
            return new CCSArrayLikeType(CCSType.fromPseucoType(pType.elementsType));
        } else if (pType instanceof PCT.ArrayType) {
            return new CCSArrayLikeType(CCSType.fromPseucoType(pType.elementsType));
        } else if (pType instanceof PCT.ClassType) {
            return new CCSClassType(pType.className);
        } else if (pType instanceof PCT.ProcedureType) {
            return CCSType.fromPseucoType(pType.returnType);
        } else {
            throw new Error(`Unexpected pseuCo type: ${pType.toString()}`);
        }
    }
}

export class CCSArrayLikeType extends CCSType {
    // Class for types `array` and `channel`
    constructor(readonly elementsType: CCSType) {
        super(CCSTypeDescriptor.INT);
    }

    override isArrayLike(): boolean {
        return true;
    }

    override isEqual(other: CCSType): boolean {
        return other instanceof CCSArrayLikeType && this.elementsType.isEqual(other.elementsType);
    }
}

export class CCSClassType extends CCSType {
    // Class for types `monitor` and `struct`
    constructor(readonly className: string) {
        super(CCSTypeDescriptor.INT);
    }

    override isClass(): boolean {
        return true;
    }

    override isEqual(other: CCSType): boolean {
        return other instanceof CCSClassType && this.className == other.className;
    }
}


